Package: DataAudit
Type: Package
Title: Comprehensive Data Quality Auditing and Validation
Version: 0.1.0
Authors@R: 
    c(
        person(
            given = "Vinodh Kumar",
            family = "Obli Rajendran",
            role = c("aut", "cre"),
            email = "vinodhkumar.rajendran@gmail.com",
            comment = c(
                ORCID = "0000-0002-7232-4122"
            )
        ),
        person(
            given = "Keerthi",
            family = "Aaradhana",
            role = "aut",
            email = "vkeerthi1817@gmail.com"
        )
    )
Description: Provides tools for systematic assessment, validation, and
    reporting of data quality. The package detects common data-quality
    problems including missing or blank values, duplicate observations,
    infinite values, constant and near-zero variance variables, outliers,
    invalid ranges and categories, type and pattern violations, problematic
    dates and identifiers, sequence errors, dependency violations, and
    cross-variable inconsistencies. It also supports reusable validation
    rules, integrated audit reports, and standardized data-quality scoring
    for reproducible data-quality assessment workflows.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 8.0.0
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), tibble
Config/testthat/edition: 3
VignetteBuilder: knitr
URL: https://github.com/vinodhpmd/DataAudit
BugReports: https://github.com/vinodhpmd/DataAudit/issues
NeedsCompilation: no
Packaged: 2026-07-28 12:21:45 UTC; m
Author: Vinodh Kumar Obli Rajendran [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7232-4122>),
  Keerthi Aaradhana [aut]
Maintainer: Vinodh Kumar Obli Rajendran <vinodhkumar.rajendran@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-06 12:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,DataAuditReport)
S3method(print,DataAuditRuleResult)
S3method(print,DataAuditScore)
S3method(summary,DataAuditReport)
S3method(summary,DataAuditRuleResult)
export(apply_audit_rules)
export(audit_data)
export(audit_rules)
export(audit_score)
export(blank_check)
export(case_check)
export(category_check)
export(consistency_check)
export(constant_check)
export(data_overview)
export(date_check)
export(dependency_check)
export(duplicate_check)
export(group_sequence_check)
export(id_check)
export(infinite_check)
export(length_check)
export(missing_check)
export(nzv_check)
export(outlier_check)
export(pattern_check)
export(range_check)
export(sequence_check)
export(type_check)
export(unique_check)
export(whitespace_check)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# DataAudit

<!-- badges: start -->

[![R-CMD-check](https://github.com/vinodhpmd/DataAudit/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/vinodhpmd/DataAudit/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/vinodhpmd/DataAudit/graph/badge.svg)](https://app.codecov.io/gh/vinodhpmd/DataAudit)

<!-- badges: end -->

## Overview

`DataAudit` is an R package for systematic data-quality auditing and
validation.

It provides tools for detecting common data-quality problems, defining
reusable validation rules, generating structured audit reports, and
calculating an overall data-quality score.

DataAudit supports checks for missing values, blank values, duplicates,
infinite values, constant and near-zero variance variables, outliers,
ranges, categories, data types, patterns, dates, identifiers, lengths,
character case, whitespace, sequences, dependencies, uniqueness, and
cross-variable consistency.

## Installation

You can install DataAudit from GitHub with:

``` r
# install.packages("pak")
pak::pak("vinodhpmd/DataAudit")
```

Then load the package:

``` r
library(DataAudit)
```

## Quick example

Create a dataset containing several data-quality problems:

``` r
dat <- data.frame(
  ID = c(
    "A001",
    "A002",
    "A002",
    "BAD"
  ),
  Species = c(
    "Dog",
    "Horse",
    "",
    "Cat"
  ),
  Age = c(
    5,
    40,
    -2,
    NA
  ),
  stringsAsFactors = FALSE
)

dat
```

    ##     ID Species Age
    ## 1 A001     Dog   5
    ## 2 A002   Horse  40
    ## 3 A002          -2
    ## 4  BAD     Cat  NA

## General data audit

Use `audit_data()` for a general first-pass audit:

``` r
report <- audit_data(
  dat,
  include_nzv = FALSE
)

report
```

    ## DataAudit Report
    ## ================
    ## 
    ## Rows:                  4
    ## Variables:             3
    ## Variables with issues: 2
    ## Duplicated rows:       0
    ## Total issues:          2
    ## Rule violations:       0
    ## Overall issues:        2
    ## 
    ## Variables with issues
    ## ---------------------
    ##  Variable     Class Missing MissingPercent Blank Infinite Unique Constant
    ##   Species character       0              0     1        0      4    FALSE
    ##       Age   numeric       1             25     0        0      3    FALSE
    ##  NearZeroVariance Issue
    ##             FALSE  TRUE
    ##             FALSE  TRUE
    ## 
    ## No duplicated rows detected.

Detailed results are available through:

``` r
report$overview
```

    ##                         Metric Value
    ## 1                         Rows     4
    ## 2                    Variables     3
    ## 3               Missing values     1
    ## 4                 Blank values     1
    ## 5              Infinite values     0
    ## 6              Duplicated rows     0
    ## 7           Constant variables     0
    ## 8 Near-zero variance variables     0
    ## 9        Variables with issues     2

``` r
report$variables
```

    ##         Variable     Class Missing MissingPercent Blank Infinite Unique
    ## ID            ID character       0              0     0        0      3
    ## Species  Species character       0              0     1        0      4
    ## Age          Age   numeric       1             25     0        0      3
    ##         Constant NearZeroVariance Issue
    ## ID         FALSE            FALSE FALSE
    ## Species    FALSE            FALSE  TRUE
    ## Age        FALSE            FALSE  TRUE

``` r
report$duplicates
```

    ##   Row Duplicate
    ## 1   1     FALSE
    ## 2   2     FALSE
    ## 3   3     FALSE
    ## 4   4     FALSE

``` r
report$summary
```

    ##   Rows Variables VariablesWithIssues DuplicateRows TotalIssues RuleViolations
    ## 1    4         3                   2             0           2              0
    ##   OverallIssues
    ## 1             2

## Individual checks

Specific checks can also be performed independently.

``` r
missing_check(dat)
```

    ##   Variable Total Missing Complete MissingPercent CompletePercent
    ## 1       ID     4       0        4              0             100
    ## 2  Species     4       0        4              0             100
    ## 3      Age     4       1        3             25              75

``` r
blank_check(dat)
```

    ##         Variable TextVariable Missing Blank BlankPercent HasBlank
    ## ID            ID         TRUE       0     0            0    FALSE
    ## Species  Species         TRUE       0     1           25     TRUE
    ## Age          Age        FALSE       1     0            0    FALSE

``` r
duplicate_check(dat)
```

    ##   Row Duplicate
    ## 1   1     FALSE
    ## 2   2     FALSE
    ## 3   3     FALSE
    ## 4   4     FALSE

``` r
range_check(
  dat,
  variable = "Age",
  min = 0,
  max = 30
)
```

    ##   Row Value BelowMinimum AboveMaximum OutOfRange
    ## 1   1     5        FALSE        FALSE      FALSE
    ## 2   2    40        FALSE         TRUE       TRUE
    ## 3   3    -2         TRUE        FALSE       TRUE
    ## 4   4    NA        FALSE        FALSE      FALSE

``` r
category_check(
  dat,
  variable = "Species",
  allowed = c(
    "Dog",
    "Cat"
  )
)
```

    ##   Row Value Missing ValidCategory InvalidCategory
    ## 1   1   Dog   FALSE          TRUE           FALSE
    ## 2   2 Horse   FALSE         FALSE            TRUE
    ## 3   3         FALSE         FALSE            TRUE
    ## 4   4   Cat   FALSE          TRUE           FALSE

## Validation rules

Reusable validation rules can be defined with `audit_rules()`:

``` r
rules <- audit_rules(
  range = list(
    Age = c(0, 30)
  ),
  category = list(
    Species = c(
      "Dog",
      "Cat"
    )
  ),
  unique = "ID",
  required = "Species",
  pattern = list(
    ID = "^A[0-9]{3}$"
  )
)
```

Apply the rules directly:

``` r
rule_results <- apply_audit_rules(
  dat,
  rules
)

rule_results
```

    ## DataAudit Rule Validation
    ## =========================
    ## 
    ## Status:           Violations detected
    ## Total violations: 8
    ## 
    ## Rule summary
    ## ------------
    ##      Rule Violations
    ##     Range          2
    ##  Category          2
    ##    Unique          2
    ##  Required          1
    ##   Pattern          1
    ## 
    ## Variables with rule violations
    ## ------------------------------
    ## - Range: Age
    ## - Category: Species
    ## - Unique: ID
    ## - Required: Species
    ## - Pattern: ID

## Integrated audit

Custom validation rules can be incorporated directly into the general
audit:

``` r
report <- audit_data(
  dat,
  include_nzv = FALSE,
  rules = rules
)

report
```

    ## DataAudit Report
    ## ================
    ## 
    ## Rows:                  4
    ## Variables:             3
    ## Variables with issues: 2
    ## Duplicated rows:       0
    ## Total issues:          2
    ## Rule violations:       8
    ## Overall issues:        10
    ## 
    ## Variables with issues
    ## ---------------------
    ##  Variable     Class Missing MissingPercent Blank Infinite Unique Constant
    ##   Species character       0              0     1        0      4    FALSE
    ##       Age   numeric       1             25     0        0      3    FALSE
    ##  NearZeroVariance Issue
    ##             FALSE  TRUE
    ##             FALSE  TRUE
    ## 
    ## No duplicated rows detected.
    ## 
    ## Custom rule validation
    ## ----------------------
    ##      Rule Violations
    ##     Range          2
    ##  Category          2
    ##    Unique          2
    ##  Required          1
    ##   Pattern          1
    ## 
    ## Total rule violations: 8

The report now combines automatic auditing and custom validation:

``` r
report$summary
```

    ##   Rows Variables VariablesWithIssues DuplicateRows TotalIssues RuleViolations
    ## 1    4         3                   2             0           2              8
    ##   OverallIssues
    ## 1            10

``` r
report$rules
```

    ## DataAudit Rule Validation
    ## =========================
    ## 
    ## Status:           Violations detected
    ## Total violations: 8
    ## 
    ## Rule summary
    ## ------------
    ##      Rule Violations
    ##     Range          2
    ##  Category          2
    ##    Unique          2
    ##  Required          1
    ##   Pattern          1
    ## 
    ## Variables with rule violations
    ## ------------------------------
    ## - Range: Age
    ## - Category: Species
    ## - Unique: ID
    ## - Required: Species
    ## - Pattern: ID

A compact summary can be obtained with:

``` r
summary(report)
```

    ## $dataset
    ##   Rows Variables VariablesWithIssues DuplicateRows TotalIssues RuleViolations
    ## 1    4         3                   2             0           2              8
    ##   OverallIssues
    ## 1            10
    ## 
    ## $variables_with_issues
    ##         Variable     Class Missing MissingPercent Blank Infinite Unique
    ## Species  Species character       0              0     1        0      4
    ## Age          Age   numeric       1             25     0        0      3
    ##         Constant NearZeroVariance Issue
    ## Species    FALSE            FALSE  TRUE
    ## Age        FALSE            FALSE  TRUE
    ## 
    ## $duplicate_rows
    ## [1] Row       Duplicate
    ## <0 rows> (or 0-length row.names)
    ## 
    ## $rules
    ##       Rule Violations
    ## 1    Range          2
    ## 2 Category          2
    ## 3   Unique          2
    ## 4 Required          1
    ## 5  Pattern          1
    ## 
    ## attr(,"class")
    ## [1] "summary.DataAuditReport" "list"

## Data-quality score

Calculate an overall data-quality score with:

``` r
score <- audit_score(report)

score
```

    ## DataAudit Quality Score
    ## =======================
    ## 
    ## Score:               54.55 / 100
    ## Quality:             Poor
    ## 
    ## Dataset
    ## -------
    ## Rows:                4
    ## Variables:           3
    ## Data cells:          12
    ## 
    ## Issues
    ## ------
    ## Automatic issues:    2
    ## Rule violations:     8
    ## Overall issues:      10

The score ranges from **0 to 100**, with higher scores indicating fewer
detected data-quality issues relative to dataset size.

DataAudit classifies scores as:

- **Excellent**
- **Good**
- **Moderate**
- **Poor**
- **Critical**

The score is intended as a compact audit summary and should be
interpreted alongside the underlying diagnostics and the intended use of
the dataset.

## Main functionality

DataAudit currently provides functions for:

- general data auditing;
- missing and blank values;
- whitespace problems;
- infinite values;
- duplicate and unique values;
- constant and near-zero variance variables;
- outliers;
- ranges and categories;
- data types;
- regular-expression patterns;
- dates and identifiers;
- character lengths and case;
- sequences and grouped sequences;
- dependencies;
- cross-variable consistency;
- reusable validation rules; and
- data-quality scoring.

## Typical workflow

``` text
Raw data
   |
   v
audit_data()
   |
   +----> Automatic checks
   |
   +----> audit_rules()
              |
              v
        Rule validation
   |
   v
DataAuditReport
   |
   +----> summary()
   |
   +----> audit_score()
```

## Authors

**Vinodh Kumar Obli Rajendran**  
ORCID: 0000-0002-7232-4122

**Keerthi Aaradhana**

## License

DataAudit is released under the MIT License.

#==========================================================
# Pattern Check
#==========================================================

#' Check Character Patterns
#'
#' Checks whether values in a character or factor variable
#' match a user-specified regular expression.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   variable to check.
#' @param pattern A single non-empty character string
#'   containing a regular expression.
#' @param ignore_case Logical. If TRUE, matching ignores
#'   letter case. Default is FALSE.
#'
#' @return A data.frame of class \code{PatternCheck}
#'   containing row-level pattern diagnostics.
#'
#' @details
#' Missing values are reported separately and are not
#' classified as matching or invalid.
#'
#' Factor variables are converted to character internally
#' for pattern matching. The original values are retained
#' in the returned data frame.
#'
#' Pattern matching is performed using \code{grepl()}.
#'
#' @examples
#' dat <- data.frame(
#'   Sample_ID = c(
#'     "SMP-001",
#'     "SMP-002",
#'     "BAD-003",
#'     NA
#'   ),
#'   stringsAsFactors = FALSE
#' )
#'
#' pattern_check(
#'   dat,
#'   variable = "Sample_ID",
#'   pattern = "^SMP-[0-9]{3}$"
#' )
#'
#' @export

pattern_check <- function(
    data,
    variable,
    pattern,
    ignore_case = FALSE
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (missing(variable) ||
      !is.character(variable) ||
      length(variable) != 1L ||
      is.na(variable) ||
      !nzchar(variable)) {

    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {

    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate pattern
  #---------------------------------------------
  if (missing(pattern) ||
      !is.character(pattern) ||
      length(pattern) != 1L ||
      is.na(pattern) ||
      !nzchar(pattern)) {

    stop(
      "'pattern' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate ignore_case
  #---------------------------------------------
  if (!is.logical(ignore_case) ||
      length(ignore_case) != 1L ||
      is.na(ignore_case)) {

    stop(
      "'ignore_case' must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Check supported type
  #---------------------------------------------
  if (!is.character(x) &&
      !is.factor(x)) {

    stop(
      "'variable' must refer to a character or factor variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Convert for matching
  #---------------------------------------------
  values <- as.character(x)

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_value <- is.na(values)

  #---------------------------------------------
  # Initialize matching vector
  #---------------------------------------------
  matches <- rep(
    FALSE,
    length(values)
  )

  #---------------------------------------------
  # Match non-missing observations
  #---------------------------------------------
  observed <- !missing_value

  if (any(observed)) {

    matches[observed] <- grepl(
      pattern = pattern,
      x = values[observed],
      ignore.case = ignore_case
    )
  }

  #---------------------------------------------
  # Invalid pattern values
  #---------------------------------------------
  invalid_pattern <-
    !missing_value &
    !matches

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = x,
    Missing = missing_value,
    Match = matches,
    InvalidPattern = invalid_pattern,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "PatternCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("pattern_check returns PatternCheck", {

  dat <- data.frame(
    ID = c("SMP-001", "SMP-002")
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_s3_class(
    result,
    "PatternCheck"
  )

})

test_that("pattern_check returns data.frame", {

  dat <- data.frame(
    ID = c("SMP-001", "SMP-002")
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("pattern_check returns expected columns", {

  dat <- data.frame(
    ID = "SMP-001"
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "Match",
      "InvalidPattern"
    )
  )

})


#==========================================================
# Matching
#==========================================================

test_that("pattern_check identifies valid patterns", {

  dat <- data.frame(
    ID = c(
      "SMP-001",
      "SMP-002",
      "SMP-999"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_true(
    all(result$Match)
  )

  expect_false(
    any(result$InvalidPattern)
  )

})

test_that("pattern_check identifies invalid patterns", {

  dat <- data.frame(
    ID = c(
      "SMP-001",
      "SMP001",
      "ABC-001",
      "SMP-12"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_equal(
    result$Match,
    c(
      TRUE,
      FALSE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$InvalidPattern,
    c(
      FALSE,
      TRUE,
      TRUE,
      TRUE
    )
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("pattern_check handles missing values", {

  dat <- data.frame(
    ID = c(
      "SMP-001",
      NA,
      "BAD"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$Match,
    c(
      TRUE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$InvalidPattern,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

})

test_that("pattern_check handles all missing values", {

  dat <- data.frame(
    ID = rep(
      NA_character_,
      5
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-"
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$Match)
  )

  expect_false(
    any(result$InvalidPattern)
  )

})


#==========================================================
# Factor variables
#==========================================================

test_that("pattern_check handles factor variables", {

  dat <- data.frame(
    ID = factor(
      c(
        "SMP-001",
        "BAD-001",
        "SMP-002"
      )
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_equal(
    result$Match,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

  expect_true(
    is.factor(result$Value)
  )

})


#==========================================================
# Case sensitivity
#==========================================================

test_that("pattern_check is case sensitive by default", {

  dat <- data.frame(
    ID = c(
      "SMP-001",
      "smp-002"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_equal(
    result$Match,
    c(
      TRUE,
      FALSE
    )
  )

})

test_that("pattern_check can ignore case", {

  dat <- data.frame(
    ID = c(
      "SMP-001",
      "smp-002",
      "Smp-003"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$",
    ignore_case = TRUE
  )

  expect_true(
    all(result$Match)
  )

})


#==========================================================
# General patterns
#==========================================================

test_that("pattern_check handles alphabetic pattern", {

  dat <- data.frame(
    Code = c(
      "ABC",
      "XYZ",
      "A12"
    )
  )

  result <- pattern_check(
    dat,
    "Code",
    "^[A-Za-z]+$"
  )

  expect_equal(
    result$Match,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )

})

test_that("pattern_check handles numeric text pattern", {

  dat <- data.frame(
    Code = c(
      "123",
      "456",
      "12A"
    )
  )

  result <- pattern_check(
    dat,
    "Code",
    "^[0-9]+$"
  )

  expect_equal(
    result$Match,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )

})

test_that("pattern_check handles blank strings", {

  dat <- data.frame(
    Code = c(
      "ABC",
      "",
      "XYZ"
    )
  )

  result <- pattern_check(
    dat,
    "Code",
    "^[A-Z]+$"
  )

  expect_equal(
    result$InvalidPattern,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Row numbers
#==========================================================

test_that("pattern_check returns row numbers", {

  dat <- data.frame(
    ID = c(
      "A",
      "B",
      "C"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^[A-Z]$"
  )

  expect_equal(
    result$Row,
    1:3
  )

})


#==========================================================
# Zero rows
#==========================================================

test_that("pattern_check handles zero-row data", {

  dat <- data.frame(
    ID = character(0)
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-"
  )

  expect_s3_class(
    result,
    "PatternCheck"
  )

  expect_equal(
    nrow(result),
    0L
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("pattern_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    ID = c(
      "SMP-001",
      "BAD",
      "SMP-002"
    )
  )

  result <- pattern_check(
    dat,
    "ID",
    "^SMP-[0-9]{3}$"
  )

  expect_s3_class(
    result,
    "PatternCheck"
  )

  expect_equal(
    sum(result$InvalidPattern),
    1L
  )

})


#==========================================================
# Variable validation
#==========================================================

test_that("pattern_check rejects missing variable", {

  dat <- data.frame(
    ID = "SMP-001"
  )

  expect_error(
    pattern_check(
      dat,
      pattern = "^SMP-"
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects NULL variable", {

  expect_error(
    pattern_check(
      iris,
      NULL,
      "^A"
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects numeric variable argument", {

  expect_error(
    pattern_check(
      iris,
      1,
      "^A"
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects multiple variable names", {

  expect_error(
    pattern_check(
      iris,
      c("Species", "Sepal.Length"),
      "^A"
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects NA variable", {

  expect_error(
    pattern_check(
      iris,
      NA_character_,
      "^A"
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects empty variable", {

  expect_error(
    pattern_check(
      iris,
      "",
      "^A"
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects unknown variable", {

  expect_error(
    pattern_check(
      iris,
      "Unknown",
      "^A"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )

})


#==========================================================
# Pattern validation
#==========================================================

test_that("pattern_check rejects missing pattern", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID"
    ),
    "'pattern' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects NULL pattern", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      NULL
    ),
    "'pattern' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects numeric pattern", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      123
    ),
    "'pattern' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects multiple patterns", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      c("^A", "^B")
    ),
    "'pattern' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects NA pattern", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      NA_character_
    ),
    "'pattern' must be a single non-empty character string"
  )

})

test_that("pattern_check rejects empty pattern", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      ""
    ),
    "'pattern' must be a single non-empty character string"
  )

})


#==========================================================
# ignore_case validation
#==========================================================

test_that("pattern_check rejects character ignore_case", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^A$",
      ignore_case = "TRUE"
    ),
    "'ignore_case' must be TRUE or FALSE"
  )

})

test_that("pattern_check rejects numeric ignore_case", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^A$",
      ignore_case = 1
    ),
    "'ignore_case' must be TRUE or FALSE"
  )

})

test_that("pattern_check rejects vector ignore_case", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^A$",
      ignore_case = c(TRUE, FALSE)
    ),
    "'ignore_case' must be TRUE or FALSE"
  )

})

test_that("pattern_check rejects NA ignore_case", {

  dat <- data.frame(
    ID = "A"
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^A$",
      ignore_case = NA
    ),
    "'ignore_case' must be TRUE or FALSE"
  )

})


#==========================================================
# Unsupported variable types
#==========================================================

test_that("pattern_check rejects numeric data variable", {

  dat <- data.frame(
    ID = 1:5
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^[0-9]+$"
    ),
    "'variable' must refer to a character or factor variable"
  )

})

test_that("pattern_check rejects logical data variable", {

  dat <- data.frame(
    ID = c(
      TRUE,
      FALSE
    )
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^TRUE$"
    ),
    "'variable' must refer to a character or factor variable"
  )

})

test_that("pattern_check rejects Date data variable", {

  dat <- data.frame(
    ID = as.Date("2026-01-01") + 0:2
  )

  expect_error(
    pattern_check(
      dat,
      "ID",
      "^2026"
    ),
    "'variable' must refer to a character or factor variable"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("pattern_check rejects NULL data", {

  expect_error(
    pattern_check(
      NULL,
      "ID",
      "^A"
    ),
    "'data' cannot be NULL"
  )

})

test_that("pattern_check rejects vector data", {

  expect_error(
    pattern_check(
      letters,
      "ID",
      "^A"
    ),
    "'data' must be a data.frame"
  )

})

test_that("pattern_check rejects list data", {

  expect_error(
    pattern_check(
      list(ID = letters),
      "ID",
      "^A"
    ),
    "'data' must be a data.frame"
  )

})

test_that("pattern_check rejects empty data.frame", {

  expect_error(
    pattern_check(
      data.frame(),
      "ID",
      "^A"
    ),
    "at least one variable"
  )

})

#==========================================================
# Duplicate Row Check
#==========================================================

#' Check Duplicate Rows
#'
#' Identifies duplicate rows in a dataset and reports their
#' row positions.
#'
#' @param data A data.frame.
#'
#' @return An object of class \code{DuplicateCheck}. The
#'   returned data.frame contains the row number and a logical
#'   indicator showing whether each row is a duplicate of an
#'   earlier row.
#'
#' @details
#' Duplicate rows are identified using \code{duplicated()}.
#' The first occurrence of a repeated row is not classified
#' as a duplicate. Only subsequent occurrences are marked as
#' duplicates.
#'
#' @examples
#' duplicate_check(iris)
#'
#' dat <- data.frame(
#'   x = c(1, 2, 2, 3),
#'   y = c("A", "B", "B", "C")
#' )
#'
#' duplicate_check(dat)
#'
#' @export

duplicate_check <- function(data) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Duplicate detection
  #---------------------------------------------
  duplicate_status <- duplicated(data)

  #---------------------------------------------
  # Result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_len(nrow(data)),
    Duplicate = duplicate_status,
    stringsAsFactors = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "DuplicateCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("duplicate_check returns DuplicateCheck object", {

  result <- duplicate_check(iris)

  expect_s3_class(
    result,
    "DuplicateCheck"
  )

})

test_that("duplicate_check returns data.frame", {

  result <- duplicate_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("duplicate_check contains expected columns", {

  result <- duplicate_check(iris)

  expect_equal(
    names(result),
    c(
      "Row",
      "Duplicate"
    )
  )

})

test_that("duplicate_check returns one row per observation", {

  result <- duplicate_check(iris)

  expect_equal(
    nrow(result),
    nrow(iris)
  )

})


#==========================================================
# No duplicates
#==========================================================

test_that("duplicate_check reports no duplicates when absent", {

  dat <- data.frame(
    x = 1:5,
    y = letters[1:5]
  )

  result <- duplicate_check(dat)

  expect_false(
    any(result$Duplicate)
  )

})

test_that("duplicate_check reports FALSE for every unique row", {

  dat <- data.frame(
    x = 1:4
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    rep(FALSE, 4)
  )

})


#==========================================================
# Duplicate rows
#==========================================================

test_that("duplicate_check detects one duplicate", {

  dat <- data.frame(
    x = c(1, 2, 2, 3),
    y = c("A", "B", "B", "C")
  )

  result <- duplicate_check(dat)

  expect_equal(
    sum(result$Duplicate),
    1L
  )

})

test_that("duplicate_check identifies correct duplicate row", {

  dat <- data.frame(
    x = c(1, 2, 2, 3),
    y = c("A", "B", "B", "C")
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Row[result$Duplicate],
    3L
  )

})

test_that("duplicate_check does not mark first occurrence", {

  dat <- data.frame(
    x = c(1, 1)
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(FALSE, TRUE)
  )

})


#==========================================================
# Multiple duplicates
#==========================================================

test_that("duplicate_check detects multiple duplicate rows", {

  dat <- data.frame(
    x = c(
      1,
      1,
      1,
      2,
      2,
      3
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    sum(result$Duplicate),
    3L
  )

})


#==========================================================
# Completely duplicated data
#==========================================================

test_that("duplicate_check handles repeated identical rows", {

  dat <- data.frame(
    x = rep(1, 5),
    y = rep("A", 5)
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      TRUE,
      TRUE
    )
  )

  expect_equal(
    sum(result$Duplicate),
    4L
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("duplicate_check handles missing values", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      NA,
      2
    ),
    y = c(
      "A",
      "B",
      "B",
      "C"
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Character data
#==========================================================

test_that("duplicate_check handles character variables", {

  dat <- data.frame(
    x = c(
      "A",
      "B",
      "B",
      "C"
    ),
    stringsAsFactors = FALSE
  )

  result <- duplicate_check(dat)

  expect_equal(
    sum(result$Duplicate),
    1L
  )

})


#==========================================================
# Factor data
#==========================================================

test_that("duplicate_check handles factor variables", {

  dat <- data.frame(
    x = factor(
      c(
        "A",
        "B",
        "B",
        "C"
      )
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    sum(result$Duplicate),
    1L
  )

})


#==========================================================
# Logical data
#==========================================================

test_that("duplicate_check handles logical variables", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE,
      FALSE,
      TRUE
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      FALSE,
      TRUE,
      TRUE
    )
  )

})


#==========================================================
# Mixed data
#==========================================================

test_that("duplicate_check handles mixed variable types", {

  dat <- data.frame(
    x = c(1, 2, 2),
    y = c("A", "B", "B"),
    z = factor(
      c("Low", "High", "High")
    ),
    flag = c(
      TRUE,
      FALSE,
      FALSE
    ),
    stringsAsFactors = FALSE
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Date
#==========================================================

test_that("duplicate_check handles Date variables", {

  dat <- data.frame(
    date = as.Date(
      c(
        "2026-01-01",
        "2026-01-02",
        "2026-01-02"
      )
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# POSIXct
#==========================================================

test_that("duplicate_check handles POSIXct variables", {

  dat <- data.frame(
    time = as.POSIXct(
      c(
        "2026-01-01 10:00:00",
        "2026-01-02 10:00:00",
        "2026-01-02 10:00:00"
      ),
      tz = "UTC"
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# One-column data
#==========================================================

test_that("duplicate_check works with one-column data", {

  dat <- data.frame(
    x = c(
      1,
      2,
      2,
      3
    )
  )

  result <- duplicate_check(dat)

  expect_equal(
    nrow(result),
    4L
  )

  expect_equal(
    sum(result$Duplicate),
    1L
  )

})


#==========================================================
# Row numbers
#==========================================================

test_that("duplicate_check generates correct row numbers", {

  dat <- data.frame(
    x = 11:15
  )

  result <- duplicate_check(dat)

  expect_equal(
    result$Row,
    1:5
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("duplicate_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- duplicate_check(dat)

  expect_s3_class(
    result,
    "DuplicateCheck"
  )

  expect_equal(
    nrow(result),
    0L
  )

  expect_length(
    result$Row,
    0L
  )

  expect_length(
    result$Duplicate,
    0L
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("duplicate_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::as_tibble(
    data.frame(
      x = c(1, 2, 2),
      y = c("A", "B", "B")
    )
  )

  result <- duplicate_check(dat)

  expect_s3_class(
    result,
    "DuplicateCheck"
  )

  expect_equal(
    sum(result$Duplicate),
    1L
  )

})


#==========================================================
# Invalid input
#==========================================================

test_that("duplicate_check rejects NULL", {

  expect_error(
    duplicate_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("duplicate_check rejects numeric vector", {

  expect_error(
    duplicate_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("duplicate_check rejects character vector", {

  expect_error(
    duplicate_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("duplicate_check rejects list", {

  expect_error(
    duplicate_check(
      list(
        x = 1:5,
        y = 6:10
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("duplicate_check rejects data.frame without variables", {

  expect_error(
    duplicate_check(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Sequence Check
#==========================================================

#' Check Numeric Sequence Integrity
#'
#' Checks a numeric or integer variable for missing values,
#' duplicated values, ordering problems, and gaps in an
#' expected sequence.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   numeric or integer variable to check.
#' @param step A positive numeric value specifying the
#'   expected increment between consecutive values.
#'   Default is 1.
#'
#' @return A data.frame of class \code{SequenceCheck}
#'   containing row-level sequence diagnostics.
#'
#' @details
#' The sequence is evaluated in the original row order.
#'
#' A duplicate is a non-missing value appearing more than
#' once in the variable. All occurrences of duplicated
#' values are flagged.
#'
#' For each non-missing observation after the first available
#' observation, the difference from the previous non-missing
#' value is calculated.
#'
#' An ordering problem occurs when this difference is less
#' than or equal to zero. Duplicate values are therefore also
#' ordering problems.
#'
#' A gap occurs when the difference from the previous
#' non-missing value is greater than \code{step}.
#'
#' Differences smaller than \code{step} but greater than zero
#' are classified as step mismatches.
#'
#' Missing observations are reported separately.
#'
#' @examples
#' dat <- data.frame(
#'   Visit = c(1, 2, 4, 4, 3, NA, 5)
#' )
#'
#' sequence_check(
#'   dat,
#'   variable = "Visit"
#' )
#'
#' sequence_check(
#'   data.frame(Time = c(0, 5, 10, 20)),
#'   variable = "Time",
#'   step = 5
#' )
#'
#' @export

sequence_check <- function(
    data,
    variable,
    step = 1
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (
    missing(variable) ||
    !is.character(variable) ||
    length(variable) != 1L ||
    is.na(variable) ||
    !nzchar(variable)
  ) {
    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {
    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate step
  #---------------------------------------------
  if (
    !is.numeric(step) ||
    length(step) != 1L ||
    is.na(step) ||
    !is.finite(step) ||
    step <= 0
  ) {
    stop(
      "'step' must be a single positive finite numeric value.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Validate variable type
  #---------------------------------------------
  if (!is.numeric(x)) {
    stop(
      "'variable' must refer to a numeric or integer variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_value <- is.na(x)

  #---------------------------------------------
  # Duplicate values
  #---------------------------------------------
  duplicate <- rep(
    FALSE,
    length(x)
  )

  available <- which(!missing_value)

  if (length(available) > 0L) {

    duplicate[available] <-
      duplicated(x[available]) |
      duplicated(
        x[available],
        fromLast = TRUE
      )
  }

  #---------------------------------------------
  # Previous non-missing value
  #---------------------------------------------
  previous_value <- rep(
    NA_real_,
    length(x)
  )

  difference <- rep(
    NA_real_,
    length(x)
  )

  if (length(available) >= 2L) {

    current_index <-
      available[-1L]

    previous_index <-
      available[-length(available)]

    previous_value[current_index] <-
      x[previous_index]

    difference[current_index] <-
      x[current_index] -
      x[previous_index]
  }

  #---------------------------------------------
  # Ordering problems
  #---------------------------------------------
  order_issue <- rep(
    FALSE,
    length(x)
  )

  comparable <-
    !is.na(difference)

  order_issue[comparable] <-
    difference[comparable] <= 0

  #---------------------------------------------
  # Gaps
  #---------------------------------------------
  gap <- rep(
    FALSE,
    length(x)
  )

  gap[comparable] <-
    difference[comparable] > step

  #---------------------------------------------
  # Step mismatch
  #---------------------------------------------
  step_mismatch <- rep(
    FALSE,
    length(x)
  )

  step_mismatch[comparable] <-
    difference[comparable] > 0 &
    difference[comparable] != step

  #---------------------------------------------
  # Overall sequence issue
  #---------------------------------------------
  sequence_issue <-
    missing_value |
    duplicate |
    order_issue |
    step_mismatch

  #---------------------------------------------
  # Valid sequence
  #---------------------------------------------
  valid_sequence <-
    !sequence_issue

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = x,
    Missing = missing_value,
    Duplicate = duplicate,
    PreviousValue = previous_value,
    Difference = difference,
    OrderIssue = order_issue,
    Gap = gap,
    StepMismatch = step_mismatch,
    SequenceIssue = sequence_issue,
    ValidSequence = valid_sequence,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "SequenceCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("sequence_check returns correct class", {

  dat <- data.frame(
    x = 1:5
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_s3_class(
    result,
    "SequenceCheck"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("sequence_check returns expected columns", {

  dat <- data.frame(
    x = 1:3
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "Duplicate",
      "PreviousValue",
      "Difference",
      "OrderIssue",
      "Gap",
      "StepMismatch",
      "SequenceIssue",
      "ValidSequence"
    )
  )
})

#==========================================================
# Correct sequence
#==========================================================

test_that("correct sequence is valid", {

  dat <- data.frame(
    x = 1:5
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_false(
    any(result$SequenceIssue)
  )

  expect_true(
    all(result$ValidSequence)
  )

  expect_false(
    any(result$Gap)
  )

  expect_false(
    any(result$OrderIssue)
  )
})

#==========================================================
# Previous values and differences
#==========================================================

test_that("previous values and differences are calculated", {

  dat <- data.frame(
    x = c(1, 2, 3, 4)
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$PreviousValue,
    c(
      NA,
      1,
      2,
      3
    )
  )

  expect_equal(
    result$Difference,
    c(
      NA,
      1,
      1,
      1
    )
  )
})

#==========================================================
# Gaps
#==========================================================

test_that("sequence_check detects gaps", {

  dat <- data.frame(
    x = c(
      1,
      2,
      4,
      5
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$Gap,
    c(
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_true(
    result$StepMismatch[3]
  )

  expect_true(
    result$SequenceIssue[3]
  )
})

test_that("large sequence gaps are detected", {

  dat <- data.frame(
    x = c(
      1,
      10,
      11
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$Difference,
    c(
      NA,
      9,
      1
    )
  )

  expect_equal(
    result$Gap,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )
})

#==========================================================
# Duplicates
#==========================================================

test_that("sequence_check flags all duplicate occurrences", {

  dat <- data.frame(
    x = c(
      1,
      2,
      2,
      3
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )
})

test_that("duplicate produces ordering problem", {

  dat <- data.frame(
    x = c(
      1,
      2,
      2,
      3
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_true(
    result$OrderIssue[3]
  )

  expect_equal(
    result$Difference[3],
    0
  )
})

#==========================================================
# Ordering
#==========================================================

test_that("sequence_check detects decreasing values", {

  dat <- data.frame(
    x = c(
      1,
      3,
      2,
      4
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$OrderIssue,
    c(
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$Difference[3],
    -1
  )
})

#==========================================================
# Missing values
#==========================================================

test_that("sequence_check detects missing values", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      2,
      3
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE,
      FALSE
    )
  )

  expect_true(
    result$SequenceIssue[2]
  )

  expect_false(
    result$ValidSequence[2]
  )
})

test_that("sequence continues across missing observations", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      2,
      NA,
      3
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_equal(
    result$PreviousValue,
    c(
      NA,
      NA,
      1,
      NA,
      2
    )
  )

  expect_equal(
    result$Difference,
    c(
      NA,
      NA,
      1,
      NA,
      1
    )
  )
})

test_that("all missing values are handled", {

  dat <- data.frame(
    x = c(
      NA_real_,
      NA_real_,
      NA_real_
    )
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$Duplicate)
  )

  expect_true(
    all(result$SequenceIssue)
  )
})

#==========================================================
# Custom step
#==========================================================

test_that("sequence_check supports custom step", {

  dat <- data.frame(
    x = c(
      0,
      5,
      10,
      15
    )
  )

  result <- sequence_check(
    dat,
    "x",
    step = 5
  )

  expect_false(
    any(result$StepMismatch)
  )

  expect_false(
    any(result$Gap)
  )

  expect_true(
    all(result$ValidSequence)
  )
})

test_that("custom step detects gap", {

  dat <- data.frame(
    x = c(
      0,
      5,
      15,
      20
    )
  )

  result <- sequence_check(
    dat,
    "x",
    step = 5
  )

  expect_equal(
    result$Gap,
    c(
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )
})

test_that("difference below expected step is mismatch but not gap", {

  dat <- data.frame(
    x = c(
      0,
      5,
      7
    )
  )

  result <- sequence_check(
    dat,
    "x",
    step = 5
  )

  expect_true(
    result$StepMismatch[3]
  )

  expect_false(
    result$Gap[3]
  )
})

#==========================================================
# Numeric sequences
#==========================================================

test_that("sequence_check supports numeric sequences", {

  dat <- data.frame(
    x = c(
      0,
      0.5,
      1,
      1.5
    )
  )

  result <- sequence_check(
    dat,
    "x",
    step = 0.5
  )

  expect_true(
    all(result$ValidSequence)
  )
})

#==========================================================
# Single observation
#==========================================================

test_that("sequence_check handles single observation", {

  dat <- data.frame(
    x = 1
  )

  result <- sequence_check(
    dat,
    "x"
  )

  expect_true(
    is.na(result$PreviousValue)
  )

  expect_true(
    is.na(result$Difference)
  )

  expect_true(
    result$ValidSequence
  )
})

#==========================================================
# Variable validation
#==========================================================

test_that("sequence_check rejects invalid variable argument", {

  expect_error(
    sequence_check(iris),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    sequence_check(
      iris,
      NULL
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    sequence_check(
      iris,
      ""
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    sequence_check(
      iris,
      1
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    sequence_check(
      iris,
      NA_character_
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    sequence_check(
      iris,
      c(
        "Sepal.Length",
        "Sepal.Width"
      )
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("sequence_check rejects unknown variable", {

  expect_error(
    sequence_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

#==========================================================
# Variable type
#==========================================================

test_that("sequence_check rejects character variable", {

  dat <- data.frame(
    x = c(
      "1",
      "2",
      "3"
    )
  )

  expect_error(
    sequence_check(
      dat,
      "x"
    ),
    "'variable' must refer to a numeric or integer variable"
  )
})

test_that("sequence_check rejects factor variable", {

  dat <- data.frame(
    x = factor(
      c(
        "1",
        "2",
        "3"
      )
    )
  )

  expect_error(
    sequence_check(
      dat,
      "x"
    ),
    "'variable' must refer to a numeric or integer variable"
  )
})

test_that("sequence_check rejects logical variable", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE
    )
  )

  expect_error(
    sequence_check(
      dat,
      "x"
    ),
    "'variable' must refer to a numeric or integer variable"
  )
})

#==========================================================
# Step validation
#==========================================================

test_that("sequence_check rejects invalid step", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    sequence_check(
      dat,
      "x",
      step = 0
    ),
    "'step' must be a single positive finite numeric value"
  )

  expect_error(
    sequence_check(
      dat,
      "x",
      step = -1
    ),
    "'step' must be a single positive finite numeric value"
  )

  expect_error(
    sequence_check(
      dat,
      "x",
      step = NA_real_
    ),
    "'step' must be a single positive finite numeric value"
  )

  expect_error(
    sequence_check(
      dat,
      "x",
      step = Inf
    ),
    "'step' must be a single positive finite numeric value"
  )

  expect_error(
    sequence_check(
      dat,
      "x",
      step = "1"
    ),
    "'step' must be a single positive finite numeric value"
  )

  expect_error(
    sequence_check(
      dat,
      "x",
      step = c(1, 2)
    ),
    "'step' must be a single positive finite numeric value"
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("sequence_check rejects invalid data", {

  expect_error(
    sequence_check(
      NULL,
      "x"
    )
  )

  expect_error(
    sequence_check(
      1:5,
      "x"
    )
  )

  expect_error(
    sequence_check(
      list(x = 1:5),
      "x"
    )
  )

  expect_error(
    sequence_check(
      data.frame(),
      "x"
    )
  )
})

#==========================================================
# Date Check
#==========================================================

#' Check Date Values
#'
#' Checks a Date or POSIXct variable for missing values,
#' dates outside user-defined limits, and optionally future
#' dates.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   Date or POSIXct variable to check.
#' @param min_date Optional minimum acceptable date.
#' @param max_date Optional maximum acceptable date.
#' @param allow_future Logical. If FALSE, dates later than
#'   the current date are flagged. Default is TRUE.
#'
#' @return A data.frame of class \code{DateCheck} containing
#'   row-level date diagnostics.
#'
#' @details
#' The selected variable must inherit from \code{Date} or
#' \code{POSIXct}.
#'
#' Missing values are reported separately and are not
#' classified as out-of-range dates.
#'
#' When \code{allow_future = FALSE}, observations later than
#' the current date are flagged in \code{Future}.
#'
#' The \code{min_date} and \code{max_date} arguments may be
#' supplied as Date, POSIXct, or single character strings
#' in ISO format (\code{YYYY-MM-DD}).
#'
#' @examples
#' dat <- data.frame(
#'   Sample_Date = as.Date(c(
#'     "2025-01-01",
#'     "2025-06-01",
#'     "2026-01-01",
#'     NA
#'   ))
#' )
#'
#' date_check(
#'   dat,
#'   variable = "Sample_Date",
#'   min_date = "2025-01-01",
#'   max_date = "2025-12-31"
#' )
#'
#' @export

date_check <- function(
    data,
    variable,
    min_date = NULL,
    max_date = NULL,
    allow_future = TRUE
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (
    missing(variable) ||
    !is.character(variable) ||
    length(variable) != 1L ||
    is.na(variable) ||
    !nzchar(variable)
  ) {

    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {

    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate allow_future
  #---------------------------------------------
  if (
    !is.logical(allow_future) ||
    length(allow_future) != 1L ||
    is.na(allow_future)
  ) {

    stop(
      "'allow_future' must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Validate date class
  #---------------------------------------------
  if (
    !inherits(x, "Date") &&
    !inherits(x, "POSIXct")
  ) {

    stop(
      "'variable' must refer to a Date or POSIXct variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Helper: convert date limits
  #---------------------------------------------
  convert_limit <- function(value, argument) {

    # NULL is allowed
    if (is.null(value)) {
      return(NULL)
    }

    # Must contain exactly one non-missing value
    if (
      length(value) != 1L ||
      is.na(value)
    ) {

      stop(
        sprintf(
          "'%s' must contain exactly one non-missing date.",
          argument
        ),
        call. = FALSE
      )
    }

    # POSIXct
    if (inherits(value, "POSIXct")) {

      return(
        as.Date(value)
      )
    }

    # Date
    if (inherits(value, "Date")) {

      return(value)
    }

    # Character date
    if (is.character(value)) {

      converted <- suppressWarnings(
        as.Date(
          value,
          format = "%Y-%m-%d"
        )
      )

      if (is.na(converted)) {

        stop(
          sprintf(
            "'%s' could not be converted to a valid date.",
            argument
          ),
          call. = FALSE
        )
      }

      return(converted)
    }

    # Unsupported input
    stop(
      sprintf(
        "'%s' must be a Date, POSIXct, character date, or NULL.",
        argument
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Convert date limits
  #---------------------------------------------
  min_date <- convert_limit(
    min_date,
    "min_date"
  )

  max_date <- convert_limit(
    max_date,
    "max_date"
  )

  #---------------------------------------------
  # Validate relationship between limits
  #---------------------------------------------
  if (
    !is.null(min_date) &&
    !is.null(max_date) &&
    min_date > max_date
  ) {

    stop(
      "'min_date' cannot be later than 'max_date'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Convert observations to Date
  #---------------------------------------------
  date_value <- as.Date(x)

  #---------------------------------------------
  # Missing dates
  #---------------------------------------------
  missing_value <- is.na(date_value)

  #---------------------------------------------
  # Below minimum
  #---------------------------------------------
  below_min <- rep(
    FALSE,
    length(date_value)
  )

  if (!is.null(min_date)) {

    below_min <-
      !missing_value &
      date_value < min_date
  }

  #---------------------------------------------
  # Above maximum
  #---------------------------------------------
  above_max <- rep(
    FALSE,
    length(date_value)
  )

  if (!is.null(max_date)) {

    above_max <-
      !missing_value &
      date_value > max_date
  }

  #---------------------------------------------
  # Future dates
  #---------------------------------------------
  future <- rep(
    FALSE,
    length(date_value)
  )

  if (!allow_future) {

    future <-
      !missing_value &
      date_value > Sys.Date()
  }

  #---------------------------------------------
  # Overall validity
  #---------------------------------------------
  valid_date <-
    !missing_value &
    !below_min &
    !above_max &
    !future

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = x,
    Missing = missing_value,
    BelowMin = below_min,
    AboveMax = above_max,
    Future = future,
    ValidDate = valid_date,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "DateCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("date_check returns DateCheck", {

  dat <- data.frame(
    Date = as.Date("2025-01-01") + 0:4
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_s3_class(
    result,
    "DateCheck"
  )
})

test_that("date_check returns data.frame", {

  dat <- data.frame(
    Date = as.Date("2025-01-01") + 0:4
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("date_check returns expected columns", {

  dat <- data.frame(
    Date = as.Date("2025-01-01")
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "BelowMin",
      "AboveMax",
      "Future",
      "ValidDate"
    )
  )
})


#==========================================================
# Valid dates
#==========================================================

test_that("date_check identifies valid dates", {

  dat <- data.frame(
    Date = as.Date(c(
      "2025-01-01",
      "2025-06-01",
      "2025-12-31"
    ))
  )

  result <- date_check(
    dat,
    "Date",
    min_date = "2025-01-01",
    max_date = "2025-12-31"
  )

  expect_true(
    all(result$ValidDate)
  )

  expect_false(
    any(result$BelowMin)
  )

  expect_false(
    any(result$AboveMax)
  )
})


#==========================================================
# Minimum date
#==========================================================

test_that("date_check detects dates below minimum", {

  dat <- data.frame(
    Date = as.Date(c(
      "2024-12-31",
      "2025-01-01",
      "2025-06-01"
    ))
  )

  result <- date_check(
    dat,
    "Date",
    min_date = "2025-01-01"
  )

  expect_equal(
    result$BelowMin,
    c(
      TRUE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$ValidDate,
    c(
      FALSE,
      TRUE,
      TRUE
    )
  )
})


#==========================================================
# Maximum date
#==========================================================

test_that("date_check detects dates above maximum", {

  dat <- data.frame(
    Date = as.Date(c(
      "2025-01-01",
      "2025-12-31",
      "2026-01-01"
    ))
  )

  result <- date_check(
    dat,
    "Date",
    max_date = "2025-12-31"
  )

  expect_equal(
    result$AboveMax,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )
})


#==========================================================
# Minimum and maximum
#==========================================================

test_that("date_check handles both date limits", {

  dat <- data.frame(
    Date = as.Date(c(
      "2024-12-31",
      "2025-06-01",
      "2026-01-01"
    ))
  )

  result <- date_check(
    dat,
    "Date",
    min_date = "2025-01-01",
    max_date = "2025-12-31"
  )

  expect_equal(
    result$BelowMin,
    c(
      TRUE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$AboveMax,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

  expect_equal(
    result$ValidDate,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )
})


#==========================================================
# Boundary values
#==========================================================

test_that("date_check includes boundary dates", {

  dat <- data.frame(
    Date = as.Date(c(
      "2025-01-01",
      "2025-12-31"
    ))
  )

  result <- date_check(
    dat,
    "Date",
    min_date = "2025-01-01",
    max_date = "2025-12-31"
  )

  expect_true(
    all(result$ValidDate)
  )
})


#==========================================================
# Missing values
#==========================================================

test_that("date_check detects missing dates", {

  dat <- data.frame(
    Date = as.Date(c(
      "2025-01-01",
      NA,
      "2025-03-01"
    ))
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )
})

test_that("missing dates are not valid", {

  dat <- data.frame(
    Date = as.Date(c(
      "2025-01-01",
      NA
    ))
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_equal(
    result$ValidDate,
    c(
      TRUE,
      FALSE
    )
  )
})

test_that("missing dates are not out of range", {

  dat <- data.frame(
    Date = as.Date(c(
      NA,
      "2025-06-01"
    ))
  )

  result <- date_check(
    dat,
    "Date",
    min_date = "2025-01-01",
    max_date = "2025-12-31"
  )

  expect_false(
    result$BelowMin[1]
  )

  expect_false(
    result$AboveMax[1]
  )
})


#==========================================================
# Future dates
#==========================================================

test_that("date_check allows future dates by default", {

  dat <- data.frame(
    Date = Sys.Date() + c(
      0,
      10
    )
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_false(
    any(result$Future)
  )

  expect_true(
    all(result$ValidDate)
  )
})

test_that("date_check detects future dates", {

  dat <- data.frame(
    Date = Sys.Date() + c(
      -1,
      0,
      10
    )
  )

  result <- date_check(
    dat,
    "Date",
    allow_future = FALSE
  )

  expect_equal(
    result$Future,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

  expect_equal(
    result$ValidDate,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )
})


#==========================================================
# Date limits
#==========================================================

test_that("date_check accepts Date limits", {

  dat <- data.frame(
    Date = as.Date("2025-01-01") + 0:2
  )

  result <- date_check(
    dat,
    "Date",
    min_date = as.Date("2025-01-01"),
    max_date = as.Date("2025-01-03")
  )

  expect_true(
    all(result$ValidDate)
  )
})

test_that("date_check accepts POSIXct limits", {

  dat <- data.frame(
    Date = as.Date("2025-01-01") + 0:2
  )

  result <- date_check(
    dat,
    "Date",
    min_date = as.POSIXct(
      "2025-01-01",
      tz = "UTC"
    ),
    max_date = as.POSIXct(
      "2025-01-03",
      tz = "UTC"
    )
  )

  expect_true(
    all(result$ValidDate)
  )
})

test_that("date_check accepts character limits", {

  dat <- data.frame(
    Date = as.Date("2025-01-01") + 0:2
  )

  expect_no_error(
    date_check(
      dat,
      "Date",
      min_date = "2025-01-01",
      max_date = "2025-01-03"
    )
  )
})


#==========================================================
# POSIXct variable
#==========================================================

test_that("date_check handles POSIXct", {

  dat <- data.frame(
    Date = as.POSIXct(
      c(
        "2025-01-01 10:00:00",
        "2025-01-02 11:00:00"
      ),
      tz = "UTC"
    )
  )

  result <- date_check(
    dat,
    "Date",
    min_date = "2025-01-01",
    max_date = "2025-01-02"
  )

  expect_s3_class(
    result,
    "DateCheck"
  )

  expect_s3_class(
    result$Value,
    "POSIXct"
  )

  expect_true(
    all(result$ValidDate)
  )
})


#==========================================================
# All missing
#==========================================================

test_that("date_check handles all missing Date values", {

  dat <- data.frame(
    Date = as.Date(
      rep(
        NA_character_,
        5
      )
    )
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$ValidDate)
  )
})


#==========================================================
# Zero rows
#==========================================================

test_that("date_check handles zero-row Date data", {

  dat <- data.frame(
    Date = as.Date(
      character(0)
    )
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_s3_class(
    result,
    "DateCheck"
  )

  expect_equal(
    nrow(result),
    0L
  )
})


#==========================================================
# Tibble
#==========================================================

test_that("date_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    Date = as.Date(
      c(
        "2025-01-01",
        "2025-06-01"
      )
    )
  )

  result <- date_check(
    dat,
    "Date"
  )

  expect_s3_class(
    result,
    "DateCheck"
  )
})


#==========================================================
# Variable validation
#==========================================================

test_that("date_check rejects missing variable", {

  expect_error(
    date_check(iris),
    "'variable' must be a single non-empty character string"
  )
})

test_that("date_check rejects NULL variable", {

  expect_error(
    date_check(
      iris,
      NULL
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("date_check rejects numeric variable argument", {

  expect_error(
    date_check(
      iris,
      1
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("date_check rejects multiple variable names", {

  expect_error(
    date_check(
      iris,
      c(
        "Sepal.Length",
        "Sepal.Width"
      )
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("date_check rejects NA variable", {

  expect_error(
    date_check(
      iris,
      NA_character_
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("date_check rejects empty variable", {

  expect_error(
    date_check(
      iris,
      ""
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("date_check rejects unknown variable", {

  expect_error(
    date_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})


#==========================================================
# Variable type validation
#==========================================================

test_that("date_check rejects numeric variable", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    date_check(
      dat,
      "x"
    ),
    "'variable' must refer to a Date or POSIXct variable"
  )
})

test_that("date_check rejects character variable", {

  dat <- data.frame(
    x = c(
      "2025-01-01",
      "2025-01-02"
    )
  )

  expect_error(
    date_check(
      dat,
      "x"
    ),
    "'variable' must refer to a Date or POSIXct variable"
  )
})

test_that("date_check rejects factor variable", {

  dat <- data.frame(
    x = factor(
      c(
        "2025-01-01",
        "2025-01-02"
      )
    )
  )

  expect_error(
    date_check(
      dat,
      "x"
    ),
    "'variable' must refer to a Date or POSIXct variable"
  )
})


#==========================================================
# allow_future validation
#==========================================================

test_that("date_check rejects character allow_future", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      allow_future = "FALSE"
    ),
    "'allow_future' must be TRUE or FALSE"
  )
})

test_that("date_check rejects numeric allow_future", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      allow_future = 0
    ),
    "'allow_future' must be TRUE or FALSE"
  )
})

test_that("date_check rejects NA allow_future", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      allow_future = NA
    ),
    "'allow_future' must be TRUE or FALSE"
  )
})

test_that("date_check rejects vector allow_future", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      allow_future = c(
        TRUE,
        FALSE
      )
    ),
    "'allow_future' must be TRUE or FALSE"
  )
})


#==========================================================
# min_date validation
#==========================================================

test_that("date_check rejects invalid min_date", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      min_date = "not-a-date"
    ),
    "'min_date' could not be converted to a valid date"
  )
})

test_that("date_check rejects numeric min_date", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      min_date = 1
    ),
    "'min_date' must be a Date, POSIXct, character date, or NULL"
  )
})

test_that("date_check rejects multiple min_date values", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      min_date = c(
        "2025-01-01",
        "2025-02-01"
      )
    ),
    "'min_date' must contain exactly one non-missing date"
  )
})

test_that("date_check rejects NA min_date", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      min_date = NA_character_
    ),
    "'min_date' must contain exactly one non-missing date"
  )
})


#==========================================================
# max_date validation
#==========================================================

test_that("date_check rejects invalid max_date", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      max_date = "invalid"
    ),
    "'max_date' could not be converted to a valid date"
  )
})

test_that("date_check rejects numeric max_date", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      max_date = 1
    ),
    "'max_date' must be a Date, POSIXct, character date, or NULL"
  )
})

test_that("date_check rejects multiple max_date values", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      max_date = c(
        "2025-01-01",
        "2025-02-01"
      )
    ),
    "'max_date' must contain exactly one non-missing date"
  )
})


#==========================================================
# Limit relationship
#==========================================================

test_that("date_check rejects min_date after max_date", {

  dat <- data.frame(
    Date = Sys.Date()
  )

  expect_error(
    date_check(
      dat,
      "Date",
      min_date = "2025-12-31",
      max_date = "2025-01-01"
    ),
    "'min_date' cannot be later than 'max_date'"
  )
})


#==========================================================
# Invalid data
#==========================================================

test_that("date_check rejects NULL data", {

  expect_error(
    date_check(
      NULL,
      "Date"
    ),
    "'data' cannot be NULL"
  )
})

test_that("date_check rejects vector data", {

  expect_error(
    date_check(
      1:10,
      "Date"
    ),
    "'data' must be a data.frame"
  )
})

test_that("date_check rejects list data", {

  expect_error(
    date_check(
      list(
        Date = Sys.Date()
      ),
      "Date"
    ),
    "'data' must be a data.frame"
  )
})

test_that("date_check rejects empty data.frame", {

  expect_error(
    date_check(
      data.frame(),
      "Date"
    ),
    "at least one variable"
  )
})

#==========================================================
# Near-Zero Variance Check
#==========================================================

#' Check Near-Zero Variance Variables
#'
#' Identifies variables with zero or near-zero variance
#' based on the percentage of unique values and the
#' frequency ratio of the two most common values.
#'
#' @param data A data.frame.
#' @param unique_cut Numeric. Maximum percentage of unique
#'   values used to identify near-zero variance variables.
#'   Default is 10.
#' @param freq_cut Numeric. Minimum frequency ratio used to
#'   identify near-zero variance variables. Default is 19.
#'
#' @return A data.frame of class \code{NZVCheck} containing
#'   variable-level near-zero variance diagnostics.
#'
#' @details
#' Missing values are excluded when calculating the number
#' of unique values, percentage of unique values, and
#' frequency ratio.
#'
#' A variable is classified as having zero variance when it
#' contains one or fewer unique non-missing values.
#'
#' A non-constant variable is classified as near-zero
#' variance when both:
#'
#' \itemize{
#'   \item PercentUnique is less than or equal to
#'   \code{unique_cut}.
#'   \item FrequencyRatio is greater than \code{freq_cut}.
#' }
#'
#' Variables with zero variance are also flagged in the
#' \code{NearZeroVariance} column.
#'
#' @examples
#' dat <- data.frame(
#'   x = 1:100,
#'   y = c(rep(0, 99), 1),
#'   z = rep(5, 100)
#' )
#'
#' nzv_check(dat)
#'
#' nzv_check(
#'   dat,
#'   unique_cut = 20,
#'   freq_cut = 10
#' )
#'
#' @export

nzv_check <- function(
    data,
    unique_cut = 10,
    freq_cut = 19
) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  if (!is.numeric(unique_cut) ||
      length(unique_cut) != 1L ||
      is.na(unique_cut) ||
      !is.finite(unique_cut)) {

    stop(
      "'unique_cut' must be a single finite numeric value.",
      call. = FALSE
    )
  }

  if (unique_cut < 0 || unique_cut > 100) {

    stop(
      "'unique_cut' must be between 0 and 100.",
      call. = FALSE
    )
  }

  if (!is.numeric(freq_cut) ||
      length(freq_cut) != 1L ||
      is.na(freq_cut) ||
      !is.finite(freq_cut)) {

    stop(
      "'freq_cut' must be a single finite numeric value.",
      call. = FALSE
    )
  }

  if (freq_cut < 0) {

    stop(
      "'freq_cut' must be non-negative.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Analyze individual variable
  #---------------------------------------------
  analyse_variable <- function(x) {

    missing <- sum(is.na(x))

    observed <- x[!is.na(x)]

    n_observed <- length(observed)

    unique_values <- length(
      unique(observed)
    )

    #-------------------------------------------
    # Percentage unique
    #-------------------------------------------
    if (n_observed == 0L) {

      percent_unique <- NA_real_

    } else {

      percent_unique <-
        100 * unique_values / n_observed
    }

    #-------------------------------------------
    # Frequency ratio
    #-------------------------------------------
    if (unique_values <= 1L) {

      frequency_ratio <- Inf

    } else {

      frequencies <- sort(
        table(observed),
        decreasing = TRUE
      )

      frequency_ratio <-
        as.numeric(frequencies[1L]) /
        as.numeric(frequencies[2L])
    }

    #-------------------------------------------
    # Zero variance
    #-------------------------------------------
    zero_variance <- unique_values <= 1L

    #-------------------------------------------
    # Near-zero variance
    #-------------------------------------------
    if (zero_variance) {

      near_zero_variance <- TRUE

    } else {

      near_zero_variance <-
        percent_unique <= unique_cut &&
        frequency_ratio > freq_cut
    }

    list(
      Observed = n_observed,
      Missing = missing,
      UniqueValues = unique_values,
      PercentUnique = percent_unique,
      FrequencyRatio = frequency_ratio,
      ZeroVariance = zero_variance,
      NearZeroVariance = near_zero_variance
    )
  }

  #---------------------------------------------
  # Apply to variables
  #---------------------------------------------
  diagnostics <- lapply(
    data,
    analyse_variable
  )

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(

    Variable = names(data),

    Observed = vapply(
      diagnostics,
      function(x) x$Observed,
      integer(1)
    ),

    Missing = vapply(
      diagnostics,
      function(x) x$Missing,
      integer(1)
    ),

    UniqueValues = vapply(
      diagnostics,
      function(x) x$UniqueValues,
      integer(1)
    ),

    PercentUnique = vapply(
      diagnostics,
      function(x) x$PercentUnique,
      numeric(1)
    ),

    FrequencyRatio = vapply(
      diagnostics,
      function(x) x$FrequencyRatio,
      numeric(1)
    ),

    ZeroVariance = vapply(
      diagnostics,
      function(x) x$ZeroVariance,
      logical(1)
    ),

    NearZeroVariance = vapply(
      diagnostics,
      function(x) x$NearZeroVariance,
      logical(1)
    ),

    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Round percentages
  #---------------------------------------------
  result$PercentUnique <- round(
    result$PercentUnique,
    2
  )

  result$FrequencyRatio <- round(
    result$FrequencyRatio,
    2
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "NZVCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("nzv_check returns NZVCheck object", {

  result <- nzv_check(iris)

  expect_s3_class(
    result,
    "NZVCheck"
  )

})

test_that("nzv_check returns data.frame", {

  result <- nzv_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("nzv_check contains expected columns", {

  result <- nzv_check(iris)

  expect_equal(
    names(result),
    c(
      "Variable",
      "Observed",
      "Missing",
      "UniqueValues",
      "PercentUnique",
      "FrequencyRatio",
      "ZeroVariance",
      "NearZeroVariance"
    )
  )

})

test_that("nzv_check returns one row per variable", {

  result <- nzv_check(iris)

  expect_equal(
    nrow(result),
    ncol(iris)
  )

})


#==========================================================
# Normal variation
#==========================================================

test_that("nzv_check identifies variable with adequate variation", {

  dat <- data.frame(
    x = 1:100
  )

  result <- nzv_check(dat)

  expect_false(
    result$ZeroVariance
  )

  expect_false(
    result$NearZeroVariance
  )

  expect_equal(
    result$UniqueValues,
    100L
  )

  expect_equal(
    result$PercentUnique,
    100
  )

  expect_equal(
    result$FrequencyRatio,
    1
  )

})


#==========================================================
# Zero variance
#==========================================================

test_that("nzv_check identifies constant numeric variable", {

  dat <- data.frame(
    x = rep(10, 100)
  )

  result <- nzv_check(dat)

  expect_true(
    result$ZeroVariance
  )

  expect_true(
    result$NearZeroVariance
  )

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    is.infinite(result$FrequencyRatio)
  )

})

test_that("nzv_check identifies constant character variable", {

  dat <- data.frame(
    x = rep("A", 50),
    stringsAsFactors = FALSE
  )

  result <- nzv_check(dat)

  expect_true(
    result$ZeroVariance
  )

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# Near-zero variance
#==========================================================

test_that("nzv_check detects near-zero variance variable", {

  dat <- data.frame(
    x = c(
      rep(0, 99),
      1
    )
  )

  result <- nzv_check(dat)

  expect_false(
    result$ZeroVariance
  )

  expect_true(
    result$NearZeroVariance
  )

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_equal(
    result$PercentUnique,
    2
  )

  expect_equal(
    result$FrequencyRatio,
    99
  )

})

test_that("nzv_check does not flag balanced binary variable", {

  dat <- data.frame(
    x = rep(
      c(0, 1),
      each = 50
    )
  )

  result <- nzv_check(dat)

  expect_false(
    result$ZeroVariance
  )

  expect_false(
    result$NearZeroVariance
  )

  expect_equal(
    result$FrequencyRatio,
    1
  )

})


#==========================================================
# Threshold boundaries
#==========================================================

test_that("nzv_check uses greater than frequency cutoff", {

  dat <- data.frame(
    x = c(
      rep(0, 19),
      1
    )
  )

  result <- nzv_check(
    dat,
    unique_cut = 10,
    freq_cut = 19
  )

  expect_false(
    result$NearZeroVariance
  )

})

test_that("nzv_check responds to custom frequency cutoff", {

  dat <- data.frame(
    x = c(
      rep(0, 19),
      1
    )
  )

  result <- nzv_check(
    dat,
    unique_cut = 10,
    freq_cut = 10
  )

  expect_true(
    result$NearZeroVariance
  )

})

test_that("nzv_check responds to custom unique cutoff", {

  dat <- data.frame(
    x = c(
      rep(0, 39),
      1
    )
  )

  result1 <- nzv_check(
    dat,
    unique_cut = 4,
    freq_cut = 19
  )

  result2 <- nzv_check(
    dat,
    unique_cut = 5,
    freq_cut = 19
  )

  expect_false(
    result1$NearZeroVariance
  )

  expect_true(
    result2$NearZeroVariance
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("nzv_check excludes missing values from calculations", {

  dat <- data.frame(
    x = c(
      rep(0, 99),
      1,
      rep(NA_real_, 10)
    )
  )

  result <- nzv_check(dat)

  expect_equal(
    result$Observed,
    100L
  )

  expect_equal(
    result$Missing,
    10L
  )

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_equal(
    result$PercentUnique,
    2
  )

  expect_equal(
    result$FrequencyRatio,
    99
  )

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# All missing
#==========================================================

test_that("nzv_check handles all-NA variable", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      10
    )
  )

  result <- nzv_check(dat)

  expect_equal(
    result$Observed,
    0L
  )

  expect_equal(
    result$Missing,
    10L
  )

  expect_equal(
    result$UniqueValues,
    0L
  )

  expect_true(
    is.na(result$PercentUnique)
  )

  expect_true(
    is.infinite(result$FrequencyRatio)
  )

  expect_true(
    result$ZeroVariance
  )

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# Factor
#==========================================================

test_that("nzv_check handles factors", {

  dat <- data.frame(
    x = factor(
      c(
        rep("A", 99),
        "B"
      )
    )
  )

  result <- nzv_check(dat)

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_equal(
    result$FrequencyRatio,
    99
  )

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# Character
#==========================================================

test_that("nzv_check handles character variables", {

  dat <- data.frame(
    x = c(
      rep("Dog", 99),
      "Cat"
    ),
    stringsAsFactors = FALSE
  )

  result <- nzv_check(dat)

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# Logical
#==========================================================

test_that("nzv_check handles logical variables", {

  dat <- data.frame(
    x = c(
      rep(TRUE, 99),
      FALSE
    )
  )

  result <- nzv_check(dat)

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# Date
#==========================================================

test_that("nzv_check handles Date variables", {

  dat <- data.frame(
    x = c(
      rep(
        as.Date("2026-01-01"),
        99
      ),
      as.Date("2026-01-02")
    )
  )

  result <- nzv_check(dat)

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# POSIXct
#==========================================================

test_that("nzv_check handles POSIXct variables", {

  a <- as.POSIXct(
    "2026-01-01 10:00:00",
    tz = "UTC"
  )

  b <- as.POSIXct(
    "2026-01-02 10:00:00",
    tz = "UTC"
  )

  dat <- data.frame(
    x = c(
      rep(a, 99),
      b
    )
  )

  result <- nzv_check(dat)

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_true(
    result$NearZeroVariance
  )

})


#==========================================================
# Multiple variables
#==========================================================

test_that("nzv_check correctly classifies multiple variables", {

  dat <- data.frame(
    ID = 1:100,
    Constant = rep(5, 100),
    Rare = c(
      rep(0, 99),
      1
    ),
    Balanced = rep(
      c(0, 1),
      each = 50
    )
  )

  result <- nzv_check(dat)

  expect_equal(
    result$Variable,
    c(
      "ID",
      "Constant",
      "Rare",
      "Balanced"
    )
  )

  expect_equal(
    result$ZeroVariance,
    c(
      FALSE,
      TRUE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$NearZeroVariance,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# One-row data
#==========================================================

test_that("nzv_check handles one-row data", {

  dat <- data.frame(
    x = 1,
    y = "A",
    stringsAsFactors = FALSE
  )

  result <- nzv_check(dat)

  expect_true(
    all(result$ZeroVariance)
  )

  expect_true(
    all(result$NearZeroVariance)
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("nzv_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- nzv_check(dat)

  expect_equal(
    result$Observed,
    c(0L, 0L)
  )

  expect_equal(
    result$Missing,
    c(0L, 0L)
  )

  expect_equal(
    result$UniqueValues,
    c(0L, 0L)
  )

  expect_true(
    all(is.na(result$PercentUnique))
  )

  expect_true(
    all(is.infinite(result$FrequencyRatio))
  )

  expect_true(
    all(result$ZeroVariance)
  )

  expect_true(
    all(result$NearZeroVariance)
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("nzv_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = 1:100,
    y = c(
      rep(0, 99),
      1
    )
  )

  result <- nzv_check(dat)

  expect_s3_class(
    result,
    "NZVCheck"
  )

  expect_equal(
    result$NearZeroVariance,
    c(FALSE, TRUE)
  )

})


#==========================================================
# unique_cut validation
#==========================================================

test_that("nzv_check rejects non-numeric unique_cut", {

  expect_error(
    nzv_check(
      iris,
      unique_cut = "10"
    ),
    "'unique_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects vector unique_cut", {

  expect_error(
    nzv_check(
      iris,
      unique_cut = c(5, 10)
    ),
    "'unique_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects NA unique_cut", {

  expect_error(
    nzv_check(
      iris,
      unique_cut = NA_real_
    ),
    "'unique_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects infinite unique_cut", {

  expect_error(
    nzv_check(
      iris,
      unique_cut = Inf
    ),
    "'unique_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects negative unique_cut", {

  expect_error(
    nzv_check(
      iris,
      unique_cut = -1
    ),
    "'unique_cut' must be between 0 and 100"
  )

})

test_that("nzv_check rejects unique_cut above 100", {

  expect_error(
    nzv_check(
      iris,
      unique_cut = 101
    ),
    "'unique_cut' must be between 0 and 100"
  )

})


#==========================================================
# freq_cut validation
#==========================================================

test_that("nzv_check rejects non-numeric freq_cut", {

  expect_error(
    nzv_check(
      iris,
      freq_cut = "19"
    ),
    "'freq_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects vector freq_cut", {

  expect_error(
    nzv_check(
      iris,
      freq_cut = c(10, 19)
    ),
    "'freq_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects NA freq_cut", {

  expect_error(
    nzv_check(
      iris,
      freq_cut = NA_real_
    ),
    "'freq_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects infinite freq_cut", {

  expect_error(
    nzv_check(
      iris,
      freq_cut = Inf
    ),
    "'freq_cut' must be a single finite numeric value"
  )

})

test_that("nzv_check rejects negative freq_cut", {

  expect_error(
    nzv_check(
      iris,
      freq_cut = -1
    ),
    "'freq_cut' must be non-negative"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("nzv_check rejects NULL", {

  expect_error(
    nzv_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("nzv_check rejects numeric vector", {

  expect_error(
    nzv_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("nzv_check rejects character vector", {

  expect_error(
    nzv_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("nzv_check rejects list", {

  expect_error(
    nzv_check(
      list(
        x = 1:10
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("nzv_check rejects data.frame without variables", {

  expect_error(
    nzv_check(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Apply Audit Rules
#==========================================================

#' Apply Data Audit Rules
#'
#' Applies a set of validation rules created by
#' \code{\link{audit_rules}} to a data frame.
#'
#' @param data A data.frame.
#' @param rules An object of class \code{DataAuditRules}
#'   created by \code{\link{audit_rules}}.
#'
#' @return An object of class \code{DataAuditRuleResult}.
#'
#' @details
#' The function evaluates range, category, uniqueness,
#' required-value, and regular-expression rules.
#'
#' Rules referring to variables that are not present in
#' \code{data} produce an error.
#'
#' @examples
#' dat <- data.frame(
#'   Animal_ID = c("A001", "A002", "A002", NA),
#'   Species = c("Dog", "Cat", "Horse", "Dog"),
#'   Age = c(5, 3, 40, 2),
#'   stringsAsFactors = FALSE
#' )
#'
#' rules <- audit_rules(
#'   range = list(
#'     Age = c(0, 30)
#'   ),
#'   category = list(
#'     Species = c("Dog", "Cat", "Cattle")
#'   ),
#'   unique = "Animal_ID",
#'   required = "Animal_ID",
#'   pattern = list(
#'     Animal_ID = "^A[0-9]+$"
#'   )
#' )
#'
#' apply_audit_rules(dat, rules)
#'
#' @export

apply_audit_rules <- function(
    data,
    rules
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate rules
  #---------------------------------------------
  if (!inherits(rules, "DataAuditRules")) {
    stop(
      "'rules' must be an object created by audit_rules().",
      call. = FALSE
    )
  }

  required_components <- c(
    "range",
    "category",
    "unique",
    "required",
    "pattern"
  )

  if (
    !all(
      required_components %in% names(rules)
    )
  ) {
    stop(
      "Invalid 'DataAuditRules' object.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Determine variables used by rules
  #---------------------------------------------
  rule_variables <- unique(
    c(
      names(rules$range),
      names(rules$category),
      rules$unique,
      rules$required,
      names(rules$pattern)
    )
  )

  rule_variables <- rule_variables[
    !is.na(rule_variables) &
      nzchar(rule_variables)
  ]

  #---------------------------------------------
  # Check variables exist
  #---------------------------------------------
  missing_variables <- setdiff(
    rule_variables,
    names(data)
  )

  if (length(missing_variables) > 0L) {

    stop(
      sprintf(
        "Variable(s) referenced by rules not found in 'data': %s.",
        paste(
          missing_variables,
          collapse = ", "
        )
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Storage
  #---------------------------------------------
  range_results <- list()
  category_results <- list()
  unique_results <- list()
  required_results <- list()
  pattern_results <- list()

  #---------------------------------------------
  # Range rules
  #---------------------------------------------
  if (
    !is.null(rules$range) &&
    length(rules$range) > 0L
  ) {

    for (variable in names(rules$range)) {

      limits <- rules$range[[variable]]

      x <- data[[variable]]

      if (!is.numeric(x)) {
        stop(
          sprintf(
            "Range rule requires '%s' to be numeric.",
            variable
          ),
          call. = FALSE
        )
      }

      invalid <-
        !is.na(x) &
        (
          x < limits[1L] |
            x > limits[2L]
        )

      range_results[[variable]] <- data.frame(
        Row = seq_len(nrow(data)),
        Variable = variable,
        Value = x,
        Minimum = limits[1L],
        Maximum = limits[2L],
        Violation = invalid,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  #---------------------------------------------
  # Category rules
  #---------------------------------------------
  if (
    !is.null(rules$category) &&
    length(rules$category) > 0L
  ) {

    for (variable in names(rules$category)) {

      allowed <- rules$category[[variable]]

      x <- data[[variable]]

      invalid <-
        !is.na(x) &
        !(x %in% allowed)

      category_results[[variable]] <- data.frame(
        Row = seq_len(nrow(data)),
        Variable = variable,
        Value = as.character(x),
        Violation = invalid,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  #---------------------------------------------
  # Unique rules
  #---------------------------------------------
  if (
    !is.null(rules$unique) &&
    length(rules$unique) > 0L
  ) {

    for (variable in rules$unique) {

      x <- data[[variable]]

      invalid <-
        !is.na(x) &
        (
          duplicated(x) |
            duplicated(
              x,
              fromLast = TRUE
            )
        )

      unique_results[[variable]] <- data.frame(
        Row = seq_len(nrow(data)),
        Variable = variable,
        Value = x,
        Violation = invalid,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  #---------------------------------------------
  # Required rules
  #---------------------------------------------
  if (
    !is.null(rules$required) &&
    length(rules$required) > 0L
  ) {

    for (variable in rules$required) {

      x <- data[[variable]]

      invalid <- is.na(x)

      # Treat blank character/factor values as missing
      if (
        is.character(x) ||
        is.factor(x)
      ) {

        y <- as.character(x)

        invalid <-
          invalid |
          (
            !is.na(y) &
              grepl(
                "^[[:space:]]*$",
                y
              )
          )
      }

      required_results[[variable]] <- data.frame(
        Row = seq_len(nrow(data)),
        Variable = variable,
        Value = as.character(x),
        Violation = invalid,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  #---------------------------------------------
  # Pattern rules
  #---------------------------------------------
  if (
    !is.null(rules$pattern) &&
    length(rules$pattern) > 0L
  ) {

    for (variable in names(rules$pattern)) {

      regex <- rules$pattern[[variable]]

      x <- data[[variable]]

      y <- as.character(x)

      invalid <- rep(
        FALSE,
        length(y)
      )

      non_missing <- !is.na(y)

      invalid[non_missing] <-
        !grepl(
          regex,
          y[non_missing],
          perl = TRUE
        )

      pattern_results[[variable]] <- data.frame(
        Row = seq_len(nrow(data)),
        Variable = variable,
        Value = y,
        Pattern = regex,
        Violation = invalid,
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    }
  }

  #---------------------------------------------
  # Helper: count violations
  #---------------------------------------------
  count_violations <- function(x) {

    if (length(x) == 0L) {
      return(0L)
    }

    as.integer(
      sum(
        vapply(
          x,
          function(z) {
            sum(z$Violation)
          },
          numeric(1)
        )
      )
    )
  }

  #---------------------------------------------
  # Counts
  #---------------------------------------------
  range_n <-
    count_violations(range_results)

  category_n <-
    count_violations(category_results)

  unique_n <-
    count_violations(unique_results)

  required_n <-
    count_violations(required_results)

  pattern_n <-
    count_violations(pattern_results)

  total_violations <-
    range_n +
    category_n +
    unique_n +
    required_n +
    pattern_n

  #---------------------------------------------
  # Summary
  #---------------------------------------------
  summary <- data.frame(
    Rule = c(
      "Range",
      "Category",
      "Unique",
      "Required",
      "Pattern"
    ),
    Violations = c(
      range_n,
      category_n,
      unique_n,
      required_n,
      pattern_n
    ),
    stringsAsFactors = FALSE
  )

  #---------------------------------------------
  # Overall status
  #---------------------------------------------
  status <- if (
    total_violations == 0L
  ) {
    "Passed"
  } else {
    "Violations detected"
  }

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- list(
    status = status,
    total_violations =
      total_violations,
    summary = summary,
    range = range_results,
    category = category_results,
    unique = unique_results,
    required = required_results,
    pattern = pattern_results
  )

  class(result) <- c(
    "DataAuditRuleResult",
    "list"
  )

  result
}

library(testthat)
library(DataAudit)


make_rule_data <- function() {

  data.frame(
    Animal_ID = c(
      "A001",
      "A002",
      "A002",
      NA,
      "BAD"
    ),
    Species = c(
      "Dog",
      "Cat",
      "Horse",
      " ",
      "Cattle"
    ),
    Age = c(
      5,
      3,
      40,
      NA,
      -2
    ),
    stringsAsFactors = FALSE
  )
}


test_that("apply_audit_rules returns correct class", {

  dat <- make_rule_data()

  result <- apply_audit_rules(
    dat,
    audit_rules()
  )

  expect_s3_class(
    result,
    "DataAuditRuleResult"
  )

  expect_type(
    result,
    "list"
  )
})


test_that("result contains expected components", {

  result <- apply_audit_rules(
    make_rule_data(),
    audit_rules()
  )

  expect_equal(
    names(result),
    c(
      "status",
      "total_violations",
      "summary",
      "range",
      "category",
      "unique",
      "required",
      "pattern"
    )
  )
})


test_that("empty rules pass", {

  result <- apply_audit_rules(
    make_rule_data(),
    audit_rules()
  )

  expect_equal(
    result$status,
    "Passed"
  )

  expect_equal(
    result$total_violations,
    0L
  )

  expect_equal(
    sum(result$summary$Violations),
    0
  )
})


test_that("range rules detect violations", {

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    )
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$range$Age$Violation,
    c(
      FALSE,
      FALSE,
      TRUE,
      FALSE,
      TRUE
    )
  )

  expect_equal(
    sum(result$range$Age$Violation),
    2
  )
})


test_that("range rule rejects nonnumeric variable", {

  rules <- audit_rules(
    range = list(
      Species = c(0, 10)
    )
  )

  expect_error(
    apply_audit_rules(
      make_rule_data(),
      rules
    ),
    "Range rule requires 'Species' to be numeric"
  )
})


test_that("category rules detect violations", {

  rules <- audit_rules(
    category = list(
      Species = c(
        "Dog",
        "Cat",
        "Cattle"
      )
    )
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$category$Species$Violation,
    c(
      FALSE,
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )
})


test_that("unique rules detect all duplicated values", {

  rules <- audit_rules(
    unique = "Animal_ID"
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$unique$Animal_ID$Violation,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE,
      FALSE
    )
  )
})


test_that("missing values do not count as uniqueness violations", {

  dat <- data.frame(
    ID = c(
      NA,
      NA,
      "A",
      "B"
    )
  )

  result <- apply_audit_rules(
    dat,
    audit_rules(
      unique = "ID"
    )
  )

  expect_false(
    any(
      result$unique$ID$Violation[
        1:2
      ]
    )
  )
})


test_that("required rule detects missing values", {

  rules <- audit_rules(
    required = "Animal_ID"
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$required$Animal_ID$Violation,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )
})


test_that("required rule detects blank strings", {

  rules <- audit_rules(
    required = "Species"
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$required$Species$Violation,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )
})


test_that("required rule works with numeric variables", {

  rules <- audit_rules(
    required = "Age"
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$required$Age$Violation,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )
})


test_that("pattern rules detect violations", {

  rules <- audit_rules(
    pattern = list(
      Animal_ID =
        "^A[0-9]{3}$"
    )
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$pattern$Animal_ID$Violation,
    c(
      FALSE,
      FALSE,
      FALSE,
      FALSE,
      TRUE
    )
  )
})


test_that("missing values are not pattern violations", {

  dat <- data.frame(
    ID = c(
      "A001",
      NA,
      "BAD"
    )
  )

  result <- apply_audit_rules(
    dat,
    audit_rules(
      pattern = list(
        ID = "^A[0-9]{3}$"
      )
    )
  )

  expect_equal(
    result$pattern$ID$Violation,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )
})


test_that("multiple rule types are combined", {

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    ),
    category = list(
      Species = c(
        "Dog",
        "Cat",
        "Cattle"
      )
    ),
    unique = "Animal_ID",
    required = c(
      "Animal_ID",
      "Species"
    ),
    pattern = list(
      Animal_ID =
        "^A[0-9]{3}$"
    )
  )

  result <- apply_audit_rules(
    make_rule_data(),
    rules
  )

  expect_equal(
    result$status,
    "Violations detected"
  )

  expect_true(
    result$total_violations > 0L
  )

  expect_equal(
    result$total_violations,
    sum(result$summary$Violations)
  )
})


test_that("summary contains all rule types", {

  result <- apply_audit_rules(
    make_rule_data(),
    audit_rules()
  )

  expect_equal(
    result$summary$Rule,
    c(
      "Range",
      "Category",
      "Unique",
      "Required",
      "Pattern"
    )
  )
})


test_that("rule referencing absent variable produces error", {

  rules <- audit_rules(
    range = list(
      Weight = c(0, 100)
    )
  )

  expect_error(
    apply_audit_rules(
      make_rule_data(),
      rules
    ),
    "Variable\\(s\\) referenced by rules not found in 'data': Weight"
  )
})


test_that("multiple missing rule variables are reported", {

  rules <- audit_rules(
    range = list(
      Weight = c(0, 100)
    ),
    required = "Breed"
  )

  expect_error(
    apply_audit_rules(
      make_rule_data(),
      rules
    ),
    "Weight, Breed"
  )
})


test_that("non DataAuditRules object is rejected", {

  expect_error(
    apply_audit_rules(
      make_rule_data(),
      list()
    ),
    "'rules' must be an object created by audit_rules"
  )
})


test_that("malformed DataAuditRules object is rejected", {

  rules <- list(
    range = NULL
  )

  class(rules) <- c(
    "DataAuditRules",
    "list"
  )

  expect_error(
    apply_audit_rules(
      make_rule_data(),
      rules
    ),
    "Invalid 'DataAuditRules' object"
  )
})


test_that("invalid data is rejected", {

  rules <- audit_rules()

  expect_error(
    apply_audit_rules(
      1:5,
      rules
    )
  )

  expect_error(
    apply_audit_rules(
      list(x = 1:5),
      rules
    )
  )
})


test_that("range output contains limits", {

  result <- apply_audit_rules(
    make_rule_data(),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  expect_true(
    all(
      result$range$Age$Minimum == 0
    )
  )

  expect_true(
    all(
      result$range$Age$Maximum == 30
    )
  )
})


test_that("pattern output contains pattern", {

  result <- apply_audit_rules(
    make_rule_data(),
    audit_rules(
      pattern = list(
        Animal_ID =
          "^A[0-9]{3}$"
      )
    )
  )

  expect_true(
    all(
      result$pattern$Animal_ID$Pattern ==
        "^A[0-9]{3}$"
    )
  )
})

library(testthat)
library(DataAudit)


test_that("print method returns object invisibly", {

  dat <- data.frame(
    Age = c(
      5,
      40
    )
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    )
  )

  result <- apply_audit_rules(
    dat,
    rules
  )

  returned <- withVisible(
    print(result)
  )

  expect_false(
    returned$visible
  )

  expect_identical(
    returned$value,
    result
  )
})


test_that("print method displays header", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:3
    ),
    audit_rules()
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(
      grepl(
        "DataAudit Rule Validation",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method displays status", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:3
    ),
    audit_rules()
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(
      grepl(
        "Status:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Passed",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method displays total violations", {

  dat <- data.frame(
    Age = c(
      5,
      40,
      -2
    )
  )

  result <- apply_audit_rules(
    dat,
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(
      grepl(
        "Total violations: 2",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method displays rule summary", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:3
    ),
    audit_rules()
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(
      grepl(
        "Rule summary",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Range",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method reports no violations", {

  result <- apply_audit_rules(
    data.frame(
      Age = c(
        5,
        10
      )
    ),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(
      grepl(
        "No rule violations detected.",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method reports range violation variable", {

  result <- apply_audit_rules(
    data.frame(
      Age = c(
        5,
        40
      )
    ),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(
      grepl(
        "Range: Age",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method reports all rule types", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A001",
      "BAD"
    ),
    Species = c(
      "Dog",
      "Horse",
      ""
    ),
    Age = c(
      5,
      40,
      10
    ),
    stringsAsFactors = FALSE
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    ),
    category = list(
      Species = c(
        "Dog",
        "Cat"
      )
    ),
    unique = "ID",
    required = "Species",
    pattern = list(
      ID = "^A[0-9]{3}$"
    )
  )

  result <- apply_audit_rules(
    dat,
    rules
  )

  output <- capture.output(
    print(result)
  )

  expect_true(
    any(grepl(
      "Range: Age",
      output,
      fixed = TRUE
    ))
  )

  expect_true(
    any(grepl(
      "Category: Species",
      output,
      fixed = TRUE
    ))
  )

  expect_true(
    any(grepl(
      "Unique: ID",
      output,
      fixed = TRUE
    ))
  )

  expect_true(
    any(grepl(
      "Required: Species",
      output,
      fixed = TRUE
    ))
  )

  expect_true(
    any(grepl(
      "Pattern: ID",
      output,
      fixed = TRUE
    ))
  )
})


test_that("print rejects malformed result object", {

  x <- list()

  class(x) <- c(
    "DataAuditRuleResult",
    "list"
  )

  expect_error(
    print(x),
    "Invalid 'DataAuditRuleResult' object"
  )
})


test_that("print rejects malformed summary", {

  x <- list(
    status = "Passed",
    total_violations = 0L,
    summary = "invalid",
    range = list(),
    category = list(),
    unique = list(),
    required = list(),
    pattern = list()
  )

  class(x) <- c(
    "DataAuditRuleResult",
    "list"
  )

  expect_error(
    print(x),
    "Invalid rule summary"
  )
})


test_that("print rejects summary missing required columns", {

  x <- list(
    status = "Passed",
    total_violations = 0L,
    summary = data.frame(
      Rule = "Range"
    ),
    range = list(),
    category = list(),
    unique = list(),
    required = list(),
    pattern = list()
  )

  class(x) <- c(
    "DataAuditRuleResult",
    "list"
  )

  expect_error(
    print(x),
    "Invalid rule summary"
  )
})

library(testthat)
library(DataAudit)


test_that("summary returns correct class", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:5
    ),
    audit_rules()
  )

  s <- summary(result)

  expect_s3_class(
    s,
    "summary.DataAuditRuleResult"
  )

  expect_type(
    s,
    "list"
  )
})


test_that("summary contains expected components", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:5
    ),
    audit_rules()
  )

  s <- summary(result)

  expect_equal(
    names(s),
    c(
      "status",
      "total_violations",
      "by_rule",
      "violations"
    )
  )
})


test_that("summary preserves status", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:5
    ),
    audit_rules()
  )

  s <- summary(result)

  expect_equal(
    s$status,
    "Passed"
  )

  expect_equal(
    s$total_violations,
    0L
  )
})


test_that("summary preserves rule counts", {

  dat <- data.frame(
    Age = c(
      5,
      40,
      -2
    )
  )

  result <- apply_audit_rules(
    dat,
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  s <- summary(result)

  expect_equal(
    s$by_rule,
    result$summary
  )

  expect_equal(
    s$total_violations,
    2L
  )
})


test_that("summary returns empty violation table when clean", {

  result <- apply_audit_rules(
    data.frame(
      Age = c(
        5,
        10,
        20
      )
    ),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  s <- summary(result)

  expect_true(
    is.data.frame(
      s$violations
    )
  )

  expect_equal(
    nrow(s$violations),
    0L
  )

  expect_equal(
    names(s$violations),
    c(
      "Rule",
      "Variable",
      "Row",
      "Value"
    )
  )
})


test_that("summary collects range violations", {

  result <- apply_audit_rules(
    data.frame(
      Age = c(
        5,
        40,
        -2
      )
    ),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  s <- summary(result)

  expect_equal(
    s$violations$Rule,
    c(
      "Range",
      "Range"
    )
  )

  expect_equal(
    s$violations$Variable,
    c(
      "Age",
      "Age"
    )
  )

  expect_equal(
    s$violations$Row,
    c(
      2L,
      3L
    )
  )

  expect_equal(
    s$violations$Value,
    c(
      "40",
      "-2"
    )
  )
})


test_that("summary collects all rule types", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002",
      "BAD"
    ),
    Species = c(
      "Dog",
      "Horse",
      "",
      "Cat"
    ),
    Age = c(
      5,
      40,
      -2,
      10
    ),
    stringsAsFactors = FALSE
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    ),
    category = list(
      Species = c(
        "Dog",
        "Cat"
      )
    ),
    unique = "ID",
    required = "Species",
    pattern = list(
      ID = "^A[0-9]{3}$"
    )
  )

  result <- apply_audit_rules(
    dat,
    rules
  )

  s <- summary(result)

  expect_true(
    all(
      c(
        "Range",
        "Category",
        "Unique",
        "Required",
        "Pattern"
      ) %in%
        s$violations$Rule
    )
  )
})


test_that("summary violation count matches result", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002"
    ),
    Age = c(
      5,
      40,
      -2
    )
  )

  result <- apply_audit_rules(
    dat,
    audit_rules(
      range = list(
        Age = c(0, 30)
      ),
      unique = "ID"
    )
  )

  s <- summary(result)

  expect_equal(
    nrow(s$violations),
    result$total_violations
  )
})


#==========================================================
# Defensive validation
#==========================================================

test_that("summary rejects object without correct class", {

  x <- list()

  expect_error(
    summary.DataAuditRuleResult(x),
    "Invalid 'DataAuditRuleResult' object"
  )
})


test_that("summary rejects malformed object", {

  x <- list(
    status = "Passed"
  )

  class(x) <- c(
    "DataAuditRuleResult",
    "list"
  )

  expect_error(
    summary(x),
    "Invalid 'DataAuditRuleResult' object"
  )
})


test_that("summary rejects non-data-frame rule summary", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:5
    ),
    audit_rules()
  )

  result$summary <- "invalid"

  expect_error(
    summary(result),
    "Invalid rule summary in 'DataAuditRuleResult'"
  )
})


test_that("summary rejects malformed rule summary", {

  result <- apply_audit_rules(
    data.frame(
      x = 1:5
    ),
    audit_rules()
  )

  result$summary <- data.frame(
    Wrong = 1
  )

  expect_error(
    summary(result),
    "Invalid rule summary in 'DataAuditRuleResult'"
  )
})
test_that("summary skips malformed internal rule table", {

  result <- apply_audit_rules(
    data.frame(
      Age = c(
        5,
        40
      )
    ),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  result$range$Age <-
    "malformed"

  expect_no_error(
    s <- summary(result)
  )

  expect_equal(
    nrow(s$violations),
    0L
  )
})


test_that("summary handles rule table without Row or Value", {

  result <- apply_audit_rules(
    data.frame(
      Age = c(
        5,
        40
      )
    ),
    audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  result$range$Age <- data.frame(
    Violation = c(
      FALSE,
      TRUE
    )
  )

  s <- summary(result)

  expect_equal(
    s$violations$Row,
    2L
  )

  expect_true(
    is.na(
      s$violations$Value
    )
  )
})

#==========================================================
# Length Check
#==========================================================

#' Check Character Length
#'
#' Checks the number of characters in a character or factor
#' variable and identifies values outside user-defined
#' length limits.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   variable to check.
#' @param min_length Optional minimum acceptable number of
#'   characters.
#' @param max_length Optional maximum acceptable number of
#'   characters.
#' @param exact_length Optional exact required number of
#'   characters.
#'
#' @return A data.frame of class \code{LengthCheck}
#'   containing row-level length diagnostics.
#'
#' @details
#' The selected variable must be character or factor.
#'
#' Missing values are reported separately and are not
#' classified as length violations.
#'
#' If \code{exact_length} is supplied, \code{min_length}
#' and \code{max_length} must both be NULL.
#'
#' @examples
#' dat <- data.frame(
#'   Sample_ID = c("A01", "A002", "B12", NA)
#' )
#'
#' length_check(
#'   dat,
#'   variable = "Sample_ID",
#'   exact_length = 3
#' )
#'
#' @export

length_check <- function(
    data,
    variable,
    min_length = NULL,
    max_length = NULL,
    exact_length = NULL
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (
    missing(variable) ||
    !is.character(variable) ||
    length(variable) != 1L ||
    is.na(variable) ||
    !nzchar(variable)
  ) {
    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {
    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable type
  #---------------------------------------------
  x <- data[[variable]]

  if (!is.character(x) && !is.factor(x)) {
    stop(
      "'variable' must refer to a character or factor variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate length arguments
  #---------------------------------------------
  validate_length <- function(value, argument) {

    if (is.null(value)) {
      return(NULL)
    }

    if (
      !is.numeric(value) ||
      length(value) != 1L ||
      is.na(value) ||
      !is.finite(value) ||
      value < 0 ||
      value != floor(value)
    ) {
      stop(
        sprintf(
          "'%s' must be a single non-negative integer.",
          argument
        ),
        call. = FALSE
      )
    }

    as.integer(value)
  }

  min_length <- validate_length(
    min_length,
    "min_length"
  )

  max_length <- validate_length(
    max_length,
    "max_length"
  )

  exact_length <- validate_length(
    exact_length,
    "exact_length"
  )

  #---------------------------------------------
  # Validate combinations
  #---------------------------------------------
  if (
    !is.null(exact_length) &&
    (!is.null(min_length) || !is.null(max_length))
  ) {
    stop(
      "'exact_length' cannot be combined with 'min_length' or 'max_length'.",
      call. = FALSE
    )
  }

  if (
    !is.null(min_length) &&
    !is.null(max_length) &&
    min_length > max_length
  ) {
    stop(
      "'min_length' cannot be greater than 'max_length'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Convert factor to character
  #---------------------------------------------
  value <- as.character(x)

  #---------------------------------------------
  # Missing values and string lengths
  #---------------------------------------------
  missing_value <- is.na(value)

  string_length <- nchar(
    value,
    type = "chars",
    allowNA = TRUE
  )

  #---------------------------------------------
  # Below minimum
  #---------------------------------------------
  below_min <- rep(
    FALSE,
    length(value)
  )

  if (!is.null(min_length)) {
    below_min <-
      !missing_value &
      string_length < min_length
  }

  #---------------------------------------------
  # Above maximum
  #---------------------------------------------
  above_max <- rep(
    FALSE,
    length(value)
  )

  if (!is.null(max_length)) {
    above_max <-
      !missing_value &
      string_length > max_length
  }

  #---------------------------------------------
  # Exact-length mismatch
  #---------------------------------------------
  exact_mismatch <- rep(
    FALSE,
    length(value)
  )

  if (!is.null(exact_length)) {
    exact_mismatch <-
      !missing_value &
      string_length != exact_length
  }

  #---------------------------------------------
  # Overall validity
  #---------------------------------------------
  valid_length <-
    !missing_value &
    !below_min &
    !above_max &
    !exact_mismatch

  #---------------------------------------------
  # Result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = value,
    Length = string_length,
    Missing = missing_value,
    BelowMin = below_min,
    AboveMax = above_max,
    ExactMismatch = exact_mismatch,
    ValidLength = valid_length,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  class(result) <- c(
    "LengthCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("length_check returns LengthCheck", {

  dat <- data.frame(
    ID = c("A01", "B02", "C03")
  )

  result <- length_check(
    dat,
    "ID"
  )

  expect_s3_class(result, "LengthCheck")
  expect_s3_class(result, "data.frame")
})

test_that("length_check returns expected columns", {

  dat <- data.frame(
    ID = c("A", "BB")
  )

  result <- length_check(
    dat,
    "ID"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Length",
      "Missing",
      "BelowMin",
      "AboveMax",
      "ExactMismatch",
      "ValidLength"
    )
  )
})

#==========================================================
# Length calculation
#==========================================================

test_that("length_check calculates character lengths", {

  dat <- data.frame(
    ID = c("A", "AB", "ABC", "")
  )

  result <- length_check(dat, "ID")

  expect_equal(
    result$Length,
    c(1L, 2L, 3L, 0L)
  )
})

#==========================================================
# Exact length
#==========================================================

test_that("length_check detects exact length mismatch", {

  dat <- data.frame(
    ID = c("A01", "A002", "B12")
  )

  result <- length_check(
    dat,
    "ID",
    exact_length = 3
  )

  expect_equal(
    result$ExactMismatch,
    c(FALSE, TRUE, FALSE)
  )

  expect_equal(
    result$ValidLength,
    c(TRUE, FALSE, TRUE)
  )
})

test_that("exact length accepts boundary exactly", {

  dat <- data.frame(
    ID = "ABC"
  )

  result <- length_check(
    dat,
    "ID",
    exact_length = 3
  )

  expect_true(result$ValidLength)
})

#==========================================================
# Minimum length
#==========================================================

test_that("length_check detects values below minimum", {

  dat <- data.frame(
    ID = c("A", "AB", "ABC")
  )

  result <- length_check(
    dat,
    "ID",
    min_length = 2
  )

  expect_equal(
    result$BelowMin,
    c(TRUE, FALSE, FALSE)
  )
})

#==========================================================
# Maximum length
#==========================================================

test_that("length_check detects values above maximum", {

  dat <- data.frame(
    ID = c("A", "ABC", "ABCDE")
  )

  result <- length_check(
    dat,
    "ID",
    max_length = 3
  )

  expect_equal(
    result$AboveMax,
    c(FALSE, FALSE, TRUE)
  )
})

#==========================================================
# Range
#==========================================================

test_that("length_check handles minimum and maximum", {

  dat <- data.frame(
    ID = c("A", "ABC", "ABCDE")
  )

  result <- length_check(
    dat,
    "ID",
    min_length = 2,
    max_length = 4
  )

  expect_equal(
    result$ValidLength,
    c(FALSE, TRUE, FALSE)
  )
})

#==========================================================
# Missing values
#==========================================================

test_that("length_check handles missing values", {

  dat <- data.frame(
    ID = c("ABC", NA, "DEF")
  )

  result <- length_check(
    dat,
    "ID",
    exact_length = 3
  )

  expect_equal(
    result$Missing,
    c(FALSE, TRUE, FALSE)
  )

  expect_true(
    is.na(result$Length[2])
  )

  expect_false(
    result$ValidLength[2]
  )

  expect_false(
    result$ExactMismatch[2]
  )
})

#==========================================================
# Factor
#==========================================================

test_that("length_check handles factors", {

  dat <- data.frame(
    ID = factor(
      c("A", "BB", "CCC")
    )
  )

  result <- length_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Length,
    c(1L, 2L, 3L)
  )
})

#==========================================================
# Unicode
#==========================================================

test_that("length_check counts Unicode characters", {

  dat <- data.frame(
    ID = c("ABC", "café")
  )

  result <- length_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Length,
    c(3L, 4L)
  )
})

#==========================================================
# Invalid variable
#==========================================================

test_that("length_check rejects missing variable", {

  expect_error(
    length_check(iris),
    "'variable' must be a single non-empty character string"
  )
})

test_that("length_check rejects invalid variable argument", {

  expect_error(
    length_check(iris, NULL),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    length_check(iris, ""),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    length_check(iris, 1),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    length_check(
      iris,
      c("Species", "Sepal.Length")
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    length_check(iris, NA_character_),
    "'variable' must be a single non-empty character string"
  )
})

test_that("length_check rejects unknown variable", {

  expect_error(
    length_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

#==========================================================
# Invalid variable type
#==========================================================

test_that("length_check rejects numeric variable", {

  expect_error(
    length_check(
      iris,
      "Sepal.Length"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

test_that("length_check rejects logical variable", {

  dat <- data.frame(
    x = c(TRUE, FALSE)
  )

  expect_error(
    length_check(dat, "x"),
    "'variable' must refer to a character or factor variable"
  )
})

#==========================================================
# Invalid limits
#==========================================================

test_that("length_check rejects invalid min_length", {

  dat <- data.frame(x = "ABC")

  expect_error(
    length_check(dat, "x", min_length = -1),
    "'min_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", min_length = 1.5),
    "'min_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", min_length = NA_real_),
    "'min_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", min_length = Inf),
    "'min_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", min_length = "2"),
    "'min_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", min_length = c(1, 2)),
    "'min_length' must be a single non-negative integer"
  )
})

test_that("length_check rejects invalid max_length", {

  dat <- data.frame(x = "ABC")

  expect_error(
    length_check(dat, "x", max_length = -1),
    "'max_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", max_length = 2.5),
    "'max_length' must be a single non-negative integer"
  )
})

test_that("length_check rejects invalid exact_length", {

  dat <- data.frame(x = "ABC")

  expect_error(
    length_check(dat, "x", exact_length = -1),
    "'exact_length' must be a single non-negative integer"
  )

  expect_error(
    length_check(dat, "x", exact_length = NA_real_),
    "'exact_length' must be a single non-negative integer"
  )
})

#==========================================================
# Invalid combinations
#==========================================================

test_that("exact_length cannot be combined with min_length", {

  dat <- data.frame(x = "ABC")

  expect_error(
    length_check(
      dat,
      "x",
      min_length = 2,
      exact_length = 3
    ),
    "'exact_length' cannot be combined"
  )
})

test_that("exact_length cannot be combined with max_length", {

  dat <- data.frame(x = "ABC")

  expect_error(
    length_check(
      dat,
      "x",
      max_length = 4,
      exact_length = 3
    ),
    "'exact_length' cannot be combined"
  )
})

test_that("minimum cannot exceed maximum", {

  dat <- data.frame(x = "ABC")

  expect_error(
    length_check(
      dat,
      "x",
      min_length = 5,
      max_length = 2
    ),
    "'min_length' cannot be greater than 'max_length'"
  )
})

#==========================================================
# Zero length
#==========================================================

test_that("zero is a valid length limit", {

  dat <- data.frame(
    x = c("", "A")
  )

  result <- length_check(
    dat,
    "x",
    exact_length = 0
  )

  expect_equal(
    result$ValidLength,
    c(TRUE, FALSE)
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("length_check rejects invalid data", {

  expect_error(
    length_check(NULL, "x")
  )

  expect_error(
    length_check(1:5, "x")
  )

  expect_error(
    length_check(list(x = "A"), "x")
  )

  expect_error(
    length_check(data.frame(), "x")
  )
})

#==========================================================
# Blank Value Check
#==========================================================

#' Check Blank Values
#'
#' Identifies blank values in character and factor variables.
#'
#' @param data A data.frame.
#' @param trim Logical. If TRUE, strings containing only
#'   whitespace are treated as blank. Default is TRUE.
#'
#' @return A data.frame of class \code{BlankCheck}
#'   containing variable-level blank-value diagnostics.
#'
#' @details
#' Blank values are evaluated only for character and factor
#' variables. Missing values (\code{NA}) are reported
#' separately and are not counted as blank values.
#'
#' When \code{trim = TRUE}, leading and trailing whitespace
#' is removed before checking whether a value is blank.
#' Therefore values such as \code{" "}, \code{"   "}, and
#' \code{"\\t"} are classified as blank.
#'
#' @examples
#' dat <- data.frame(
#'   name = c("Dog", "", "Cat", "   ", NA),
#'   value = 1:5,
#'   stringsAsFactors = FALSE
#' )
#'
#' blank_check(dat)
#'
#' blank_check(dat, trim = FALSE)
#'
#' @export

blank_check <- function(data, trim = TRUE) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  if (!is.logical(trim) ||
      length(trim) != 1L ||
      is.na(trim)) {

    stop(
      "'trim' must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Analyze one variable
  #---------------------------------------------
  analyse_variable <- function(x) {

    text_variable <-
      is.character(x) ||
      is.factor(x)

    missing <- sum(is.na(x))

    if (!text_variable) {

      return(
        list(
          TextVariable = FALSE,
          Missing = missing,
          Blank = 0L
        )
      )
    }

    #-------------------------------------------
    # Convert factor safely to character
    #-------------------------------------------
    values <- as.character(x)

    #-------------------------------------------
    # Exclude missing values
    #-------------------------------------------
    observed <- values[!is.na(values)]

    #-------------------------------------------
    # Optional whitespace trimming
    #-------------------------------------------
    if (trim) {

      observed <- trimws(observed)

    }

    #-------------------------------------------
    # Count blanks
    #-------------------------------------------
    blank <- sum(
      observed == ""
    )

    list(
      TextVariable = TRUE,
      Missing = missing,
      Blank = as.integer(blank)
    )
  }

  #---------------------------------------------
  # Apply to variables
  #---------------------------------------------
  diagnostics <- lapply(
    data,
    analyse_variable
  )

  #---------------------------------------------
  # Number of rows
  #---------------------------------------------
  n <- nrow(data)

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(

    Variable = names(data),

    TextVariable = vapply(
      diagnostics,
      function(x) x$TextVariable,
      logical(1)
    ),

    Missing = vapply(
      diagnostics,
      function(x) x$Missing,
      integer(1)
    ),

    Blank = vapply(
      diagnostics,
      function(x) x$Blank,
      integer(1)
    ),

    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Blank percentage
  #---------------------------------------------
  if (n == 0L) {

    result$BlankPercent <- rep(
      NA_real_,
      ncol(data)
    )

  } else {

    result$BlankPercent <- round(
      100 * result$Blank / n,
      2
    )

  }

  #---------------------------------------------
  # Flag variables
  #---------------------------------------------
  result$HasBlank <- result$Blank > 0L

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "BlankCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("blank_check returns BlankCheck object", {

  result <- blank_check(iris)

  expect_s3_class(
    result,
    "BlankCheck"
  )

})

test_that("blank_check returns data.frame", {

  result <- blank_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("blank_check contains expected columns", {

  result <- blank_check(iris)

  expect_equal(
    names(result),
    c(
      "Variable",
      "TextVariable",
      "Missing",
      "Blank",
      "BlankPercent",
      "HasBlank"
    )
  )

})

test_that("blank_check returns one row per variable", {

  result <- blank_check(iris)

  expect_equal(
    nrow(result),
    ncol(iris)
  )

})


#==========================================================
# Character variables
#==========================================================

test_that("blank_check detects empty strings", {

  dat <- data.frame(
    x = c(
      "A",
      "",
      "B",
      ""
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_true(
    result$TextVariable
  )

  expect_equal(
    result$Blank,
    2L
  )

  expect_equal(
    result$BlankPercent,
    50
  )

  expect_true(
    result$HasBlank
  )

})

test_that("blank_check detects whitespace strings by default", {

  dat <- data.frame(
    x = c(
      "A",
      " ",
      "   ",
      "B"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$Blank,
    2L
  )

  expect_true(
    result$HasBlank
  )

})

test_that("blank_check detects tab as blank when trimming", {

  dat <- data.frame(
    x = c(
      "A",
      "\t",
      "B"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$Blank,
    1L
  )

})


#==========================================================
# trim = FALSE
#==========================================================

test_that("blank_check with trim FALSE counts only empty strings", {

  dat <- data.frame(
    x = c(
      "",
      " ",
      "   ",
      "A"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(
    dat,
    trim = FALSE
  )

  expect_equal(
    result$Blank,
    1L
  )

})

test_that("blank_check with trim FALSE does not classify whitespace as blank", {

  dat <- data.frame(
    x = c(
      " ",
      "   ",
      "\t"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(
    dat,
    trim = FALSE
  )

  expect_equal(
    result$Blank,
    0L
  )

  expect_false(
    result$HasBlank
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("blank_check separates NA from blank values", {

  dat <- data.frame(
    x = c(
      "A",
      "",
      NA,
      " "
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Blank,
    2L
  )

})

test_that("blank_check handles all missing character variable", {

  dat <- data.frame(
    x = rep(
      NA_character_,
      5
    )
  )

  result <- blank_check(dat)

  expect_equal(
    result$Missing,
    5L
  )

  expect_equal(
    result$Blank,
    0L
  )

  expect_false(
    result$HasBlank
  )

})


#==========================================================
# Factor variables
#==========================================================

test_that("blank_check handles factors", {

  dat <- data.frame(
    x = factor(
      c(
        "A",
        "",
        "B",
        ""
      )
    )
  )

  result <- blank_check(dat)

  expect_true(
    result$TextVariable
  )

  expect_equal(
    result$Blank,
    2L
  )

  expect_true(
    result$HasBlank
  )

})

test_that("blank_check handles whitespace in factors", {

  dat <- data.frame(
    x = factor(
      c(
        "A",
        " ",
        "B",
        "   "
      )
    )
  )

  result <- blank_check(dat)

  expect_equal(
    result$Blank,
    2L
  )

})


#==========================================================
# Numeric variables
#==========================================================

test_that("blank_check ignores numeric variables", {

  dat <- data.frame(
    x = c(
      1,
      2,
      NA,
      4
    )
  )

  result <- blank_check(dat)

  expect_false(
    result$TextVariable
  )

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Blank,
    0L
  )

  expect_false(
    result$HasBlank
  )

})


#==========================================================
# Logical variables
#==========================================================

test_that("blank_check ignores logical variables", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE,
      NA
    )
  )

  result <- blank_check(dat)

  expect_false(
    result$TextVariable
  )

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Blank,
    0L
  )

})


#==========================================================
# Date variables
#==========================================================

test_that("blank_check ignores Date variables", {

  dat <- data.frame(
    x = as.Date(
      c(
        "2026-01-01",
        NA,
        "2026-01-03"
      )
    )
  )

  result <- blank_check(dat)

  expect_false(
    result$TextVariable
  )

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Blank,
    0L
  )

})


#==========================================================
# POSIXct variables
#==========================================================

test_that("blank_check ignores POSIXct variables", {

  x <- as.POSIXct(
    c(
      "2026-01-01 10:00:00",
      "2026-01-02 10:00:00",
      "2026-01-03 10:00:00"
    ),
    tz = "UTC"
  )

  x[2] <- NA

  dat <- data.frame(
    x = x
  )

  result <- blank_check(dat)

  expect_false(
    result$TextVariable
  )

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Blank,
    0L
  )

})


#==========================================================
# Mixed variables
#==========================================================

test_that("blank_check handles mixed variable types", {

  dat <- data.frame(
    numeric = c(1, 2, NA, 4),
    character = c("A", "", "B", " "),
    factor = factor(
      c("X", "", "Y", "Z")
    ),
    logical = c(
      TRUE,
      FALSE,
      TRUE,
      NA
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$TextVariable,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$Blank,
    c(
      0L,
      2L,
      1L,
      0L
    )
  )

  expect_equal(
    result$HasBlank,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Percentage calculation
#==========================================================

test_that("blank_check calculates percentage correctly", {

  dat <- data.frame(
    x = c(
      "",
      "",
      "A",
      "B"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$BlankPercent,
    50
  )

})

test_that("blank_check rounds percentage to two decimals", {

  dat <- data.frame(
    x = c(
      "",
      "A",
      "B"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$BlankPercent,
    33.33
  )

})


#==========================================================
# No blanks
#==========================================================

test_that("blank_check reports no blanks when absent", {

  dat <- data.frame(
    x = c(
      "Dog",
      "Cat",
      "Cattle"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$Blank,
    0L
  )

  expect_false(
    result$HasBlank
  )

})


#==========================================================
# One-column data
#==========================================================

test_that("blank_check handles one-column data", {

  dat <- data.frame(
    Species = c(
      "Dog",
      "",
      "Cat"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    nrow(result),
    1L
  )

  expect_equal(
    result$Variable,
    "Species"
  )

  expect_equal(
    result$Blank,
    1L
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("blank_check handles zero-row data", {

  dat <- data.frame(
    x = character(0),
    y = numeric(0)
  )

  result <- blank_check(dat)

  expect_s3_class(
    result,
    "BlankCheck"
  )

  expect_equal(
    nrow(result),
    2L
  )

  expect_equal(
    result$Blank,
    c(
      0L,
      0L
    )
  )

  expect_true(
    all(is.na(result$BlankPercent))
  )

  expect_false(
    any(result$HasBlank)
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("blank_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = c(
      "A",
      "",
      "B"
    ),
    y = 1:3
  )

  result <- blank_check(dat)

  expect_s3_class(
    result,
    "BlankCheck"
  )

  expect_equal(
    result$Blank,
    c(
      1L,
      0L
    )
  )

})


#==========================================================
# Variable names
#==========================================================

test_that("blank_check preserves variable names", {

  dat <- data.frame(
    Animal_ID = 1:3,
    Species = c(
      "Dog",
      "",
      "Cat"
    ),
    stringsAsFactors = FALSE
  )

  result <- blank_check(dat)

  expect_equal(
    result$Variable,
    c(
      "Animal_ID",
      "Species"
    )
  )

})


#==========================================================
# trim validation
#==========================================================

test_that("blank_check rejects non-logical trim", {

  expect_error(
    blank_check(
      iris,
      trim = "TRUE"
    ),
    "'trim' must be TRUE or FALSE"
  )

})

test_that("blank_check rejects numeric trim", {

  expect_error(
    blank_check(
      iris,
      trim = 1
    ),
    "'trim' must be TRUE or FALSE"
  )

})

test_that("blank_check rejects vector trim", {

  expect_error(
    blank_check(
      iris,
      trim = c(TRUE, FALSE)
    ),
    "'trim' must be TRUE or FALSE"
  )

})

test_that("blank_check rejects NA trim", {

  expect_error(
    blank_check(
      iris,
      trim = NA
    ),
    "'trim' must be TRUE or FALSE"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("blank_check rejects NULL", {

  expect_error(
    blank_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("blank_check rejects numeric vector", {

  expect_error(
    blank_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("blank_check rejects character vector", {

  expect_error(
    blank_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("blank_check rejects list", {

  expect_error(
    blank_check(
      list(
        x = 1:5
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("blank_check rejects data.frame without variables", {

  expect_error(
    blank_check(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Group Sequence Check
#==========================================================

#' Check Sequence Integrity Within Groups
#'
#' Checks a numeric or integer sequence independently within
#' groups for missing values, duplicates, ordering problems,
#' gaps, and step mismatches.
#'
#' @param data A data.frame.
#' @param group A single character string specifying the
#'   grouping variable.
#' @param variable A single character string specifying the
#'   numeric or integer sequence variable.
#' @param step A positive finite numeric value specifying the
#'   expected increment. Default is 1.
#'
#' @return A data.frame of class \code{GroupSequenceCheck}
#'   containing row-level sequence diagnostics.
#'
#' @details
#' Sequence integrity is evaluated independently within each
#' group and in the original row order.
#'
#' Missing values in the grouping variable are retained and
#' treated as a separate missing group for diagnostic
#' purposes.
#'
#' Missing sequence values are reported separately and are
#' excluded when calculating previous values and differences.
#'
#' All occurrences of duplicated non-missing sequence values
#' within a group are flagged.
#'
#' A gap occurs when the difference from the previous
#' non-missing value within the same group is greater than
#' \code{step}.
#'
#' @examples
#' dat <- data.frame(
#'   Animal_ID = c(
#'     "A001", "A001", "A001",
#'     "A002", "A002", "A002"
#'   ),
#'   Visit = c(
#'     1, 2, 4,
#'     1, 2, 2
#'   )
#' )
#'
#' group_sequence_check(
#'   dat,
#'   group = "Animal_ID",
#'   variable = "Visit"
#' )
#'
#' @export

group_sequence_check <- function(
    data,
    group,
    variable,
    step = 1
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Helper: validate variable name
  #---------------------------------------------
  validate_name <- function(value, argument) {

    if (
      !is.character(value) ||
      length(value) != 1L ||
      is.na(value) ||
      !nzchar(value)
    ) {
      stop(
        sprintf(
          "'%s' must be a single non-empty character string.",
          argument
        ),
        call. = FALSE
      )
    }

    if (!value %in% names(data)) {
      stop(
        sprintf(
          "Variable '%s' was not found in 'data'.",
          value
        ),
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  #---------------------------------------------
  # Required arguments
  #---------------------------------------------
  if (missing(group)) {
    stop(
      "'group' must be supplied.",
      call. = FALSE
    )
  }

  if (missing(variable)) {
    stop(
      "'variable' must be supplied.",
      call. = FALSE
    )
  }

  validate_name(
    group,
    "group"
  )

  validate_name(
    variable,
    "variable"
  )

  #---------------------------------------------
  # Group and sequence variables must differ
  #---------------------------------------------
  if (identical(group, variable)) {
    stop(
      "'group' and 'variable' must refer to different variables.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate step
  #---------------------------------------------
  if (
    !is.numeric(step) ||
    length(step) != 1L ||
    is.na(step) ||
    !is.finite(step) ||
    step <= 0
  ) {
    stop(
      "'step' must be a single positive finite numeric value.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variables
  #---------------------------------------------
  g <- data[[group]]
  x <- data[[variable]]

  #---------------------------------------------
  # Validate sequence variable
  #---------------------------------------------
  if (!is.numeric(x)) {
    stop(
      "'variable' must refer to a numeric or integer variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Missing indicators
  #---------------------------------------------
  group_missing <- is.na(g)
  missing_value <- is.na(x)

  #---------------------------------------------
  # Initialize output
  #---------------------------------------------
  duplicate <- rep(
    FALSE,
    length(x)
  )

  previous_value <- rep(
    NA_real_,
    length(x)
  )

  difference <- rep(
    NA_real_,
    length(x)
  )

  order_issue <- rep(
    FALSE,
    length(x)
  )

  gap <- rep(
    FALSE,
    length(x)
  )

  step_mismatch <- rep(
    FALSE,
    length(x)
  )

  #---------------------------------------------
  # Construct grouping key
  #---------------------------------------------
  group_key <- as.character(g)

  group_key[group_missing] <-
    "<NA_GROUP>"

  group_indices <- split(
    seq_along(x),
    group_key
  )

  #---------------------------------------------
  # Evaluate each group
  #---------------------------------------------
  for (idx in group_indices) {

    available <-
      idx[!missing_value[idx]]

    #-------------------------------------------
    # Duplicate values within group
    #-------------------------------------------
    if (length(available) > 0L) {

      duplicate[available] <-
        duplicated(x[available]) |
        duplicated(
          x[available],
          fromLast = TRUE
        )
    }

    #-------------------------------------------
    # Sequence differences
    #-------------------------------------------
    if (length(available) >= 2L) {

      current_index <-
        available[-1L]

      previous_index <-
        available[-length(available)]

      previous_value[current_index] <-
        x[previous_index]

      difference[current_index] <-
        x[current_index] -
        x[previous_index]
    }
  }

  #---------------------------------------------
  # Comparable observations
  #---------------------------------------------
  comparable <-
    !is.na(difference)

  #---------------------------------------------
  # Ordering problems
  #---------------------------------------------
  order_issue[comparable] <-
    difference[comparable] <= 0

  #---------------------------------------------
  # Gaps
  #---------------------------------------------
  gap[comparable] <-
    difference[comparable] > step

  #---------------------------------------------
  # Step mismatch
  #---------------------------------------------
  step_mismatch[comparable] <-
    difference[comparable] > 0 &
    difference[comparable] != step

  #---------------------------------------------
  # Overall issue
  #---------------------------------------------
  sequence_issue <-
    group_missing |
    missing_value |
    duplicate |
    order_issue |
    step_mismatch

  valid_sequence <-
    !sequence_issue

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Group = g,
    Value = x,
    GroupMissing = group_missing,
    Missing = missing_value,
    Duplicate = duplicate,
    PreviousValue = previous_value,
    Difference = difference,
    OrderIssue = order_issue,
    Gap = gap,
    StepMismatch = step_mismatch,
    SequenceIssue = sequence_issue,
    ValidSequence = valid_sequence,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "GroupSequenceCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

test_that("group_sequence_check returns correct class", {

  dat <- data.frame(
    ID = c("A", "A", "B", "B"),
    Visit = c(1, 2, 1, 2)
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_s3_class(
    result,
    "GroupSequenceCheck"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("group_sequence_check returns expected columns", {

  dat <- data.frame(
    ID = c("A", "A"),
    Visit = c(1, 2)
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Group",
      "Value",
      "GroupMissing",
      "Missing",
      "Duplicate",
      "PreviousValue",
      "Difference",
      "OrderIssue",
      "Gap",
      "StepMismatch",
      "SequenceIssue",
      "ValidSequence"
    )
  )
})

test_that("valid sequences within groups pass", {

  dat <- data.frame(
    ID = c(
      "A", "A", "A",
      "B", "B", "B"
    ),
    Visit = c(
      1, 2, 3,
      1, 2, 3
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_true(
    all(result$ValidSequence)
  )

  expect_false(
    any(result$SequenceIssue)
  )
})

test_that("sequence restarts within each group", {

  dat <- data.frame(
    ID = c(
      "A", "A",
      "B", "B"
    ),
    Visit = c(
      1, 2,
      1, 2
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    result$PreviousValue,
    c(
      NA, 1,
      NA, 1
    )
  )

  expect_equal(
    result$Difference,
    c(
      NA, 1,
      NA, 1
    )
  )
})

test_that("gaps are detected within groups", {

  dat <- data.frame(
    ID = c(
      "A", "A", "A",
      "B", "B", "B"
    ),
    Visit = c(
      1, 2, 4,
      1, 2, 3
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    result$Gap,
    c(
      FALSE,
      FALSE,
      TRUE,
      FALSE,
      FALSE,
      FALSE
    )
  )

  expect_true(
    result$StepMismatch[3]
  )
})

test_that("duplicates are evaluated within groups only", {

  dat <- data.frame(
    ID = c(
      "A", "A",
      "B", "B"
    ),
    Visit = c(
      1, 1,
      1, 2
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    result$Duplicate,
    c(
      TRUE,
      TRUE,
      FALSE,
      FALSE
    )
  )
})

test_that("same sequence value across groups is not duplicate", {

  dat <- data.frame(
    ID = c(
      "A",
      "B"
    ),
    Visit = c(
      1,
      1
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_false(
    any(result$Duplicate)
  )
})

test_that("decreasing sequence is detected within group", {

  dat <- data.frame(
    ID = c(
      "A", "A", "A"
    ),
    Visit = c(
      1, 3, 2
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_true(
    result$OrderIssue[3]
  )

  expect_equal(
    result$Difference[3],
    -1
  )
})

test_that("missing sequence values are detected", {

  dat <- data.frame(
    ID = c(
      "A", "A", "A"
    ),
    Visit = c(
      1, NA, 2
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$PreviousValue,
    c(
      NA,
      NA,
      1
    )
  )

  expect_true(
    result$SequenceIssue[2]
  )
})

test_that("missing groups are flagged", {

  dat <- data.frame(
    ID = c(
      "A",
      NA,
      NA
    ),
    Visit = c(
      1,
      1,
      2
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    result$GroupMissing,
    c(
      FALSE,
      TRUE,
      TRUE
    )
  )

  expect_true(
    result$SequenceIssue[2]
  )

  expect_true(
    result$SequenceIssue[3]
  )
})

test_that("missing group rows can still be diagnosed as a group", {

  dat <- data.frame(
    ID = c(
      NA,
      NA,
      NA
    ),
    Visit = c(
      1,
      2,
      4
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_equal(
    result$Difference,
    c(
      NA,
      1,
      2
    )
  )

  expect_true(
    result$Gap[3]
  )
})

test_that("custom step works within groups", {

  dat <- data.frame(
    ID = c(
      "A", "A", "A",
      "B", "B", "B"
    ),
    Time = c(
      0, 5, 10,
      0, 5, 10
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Time",
    step = 5
  )

  expect_true(
    all(result$ValidSequence)
  )
})

test_that("custom step detects group-specific gap", {

  dat <- data.frame(
    ID = c(
      "A", "A", "A"
    ),
    Time = c(
      0, 5, 15
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Time",
    step = 5
  )

  expect_equal(
    result$Gap,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )
})

test_that("single observation groups are handled", {

  dat <- data.frame(
    ID = c(
      "A",
      "B",
      "C"
    ),
    Visit = c(
      1,
      1,
      1
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_true(
    all(is.na(result$PreviousValue))
  )

  expect_true(
    all(is.na(result$Difference))
  )

  expect_true(
    all(result$ValidSequence)
  )
})

test_that("all missing sequence values are handled", {

  dat <- data.frame(
    ID = c(
      "A",
      "A"
    ),
    Visit = c(
      NA_real_,
      NA_real_
    )
  )

  result <- group_sequence_check(
    dat,
    "ID",
    "Visit"
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$Duplicate)
  )

  expect_true(
    all(result$SequenceIssue)
  )
})

test_that("group is required", {

  dat <- data.frame(
    ID = "A",
    Visit = 1
  )

  expect_error(
    group_sequence_check(
      dat,
      variable = "Visit"
    ),
    "'group' must be supplied"
  )
})

test_that("variable is required", {

  dat <- data.frame(
    ID = "A",
    Visit = 1
  )

  expect_error(
    group_sequence_check(
      dat,
      group = "ID"
    ),
    "'variable' must be supplied"
  )
})

test_that("group argument is validated", {

  dat <- data.frame(
    ID = "A",
    Visit = 1
  )

  expect_error(
    group_sequence_check(
      dat,
      NULL,
      "Visit"
    ),
    "'group' must be a single non-empty character string"
  )

  expect_error(
    group_sequence_check(
      dat,
      "",
      "Visit"
    ),
    "'group' must be a single non-empty character string"
  )

  expect_error(
    group_sequence_check(
      dat,
      1,
      "Visit"
    ),
    "'group' must be a single non-empty character string"
  )

  expect_error(
    group_sequence_check(
      dat,
      NA_character_,
      "Visit"
    ),
    "'group' must be a single non-empty character string"
  )

  expect_error(
    group_sequence_check(
      dat,
      c("ID", "Visit"),
      "Visit"
    ),
    "'group' must be a single non-empty character string"
  )
})

test_that("variable argument is validated", {

  dat <- data.frame(
    ID = "A",
    Visit = 1
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      NULL
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      ""
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      1
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("unknown variables are rejected", {

  dat <- data.frame(
    ID = "A",
    Visit = 1
  )

  expect_error(
    group_sequence_check(
      dat,
      "Unknown",
      "Visit"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

test_that("group and variable must differ", {

  dat <- data.frame(
    ID = 1:3
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      "ID"
    ),
    "must refer to different variables"
  )
})

test_that("sequence variable must be numeric", {

  dat <- data.frame(
    ID = c("A", "A"),
    Visit = c("1", "2")
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      "Visit"
    ),
    "'variable' must refer to a numeric or integer variable"
  )
})

test_that("factor sequence variable is rejected", {

  dat <- data.frame(
    ID = c("A", "A"),
    Visit = factor(
      c("1", "2")
    )
  )

  expect_error(
    group_sequence_check(
      dat,
      "ID",
      "Visit"
    ),
    "'variable' must refer to a numeric or integer variable"
  )
})

test_that("invalid step is rejected", {

  dat <- data.frame(
    ID = c("A", "A"),
    Visit = c(1, 2)
  )

  invalid_steps <- list(
    0,
    -1,
    NA_real_,
    Inf,
    -Inf,
    "1",
    c(1, 2)
  )

  for (s in invalid_steps) {

    expect_error(
      group_sequence_check(
        dat,
        "ID",
        "Visit",
        step = s
      ),
      "'step' must be a single positive finite numeric value"
    )
  }
})

test_that("invalid data is rejected", {

  expect_error(
    group_sequence_check(
      NULL,
      "ID",
      "Visit"
    )
  )

  expect_error(
    group_sequence_check(
      1:5,
      "ID",
      "Visit"
    )
  )

  expect_error(
    group_sequence_check(
      list(
        ID = "A",
        Visit = 1
      ),
      "ID",
      "Visit"
    )
  )

  expect_error(
    group_sequence_check(
      data.frame(),
      "ID",
      "Visit"
    )
  )
})

#==========================================================
# Dependency Check
#==========================================================

#' Check Conditional Variable Dependency
#'
#' Checks whether a target variable satisfies a requirement
#' when a specified condition is met in another variable.
#'
#' @param data A data.frame.
#' @param condition_variable A single character string
#'   specifying the variable defining the condition.
#' @param condition_value The value of \code{condition_variable}
#'   that activates the dependency rule.
#' @param target_variable A single character string specifying
#'   the variable that is required when the condition is met.
#' @param allow_blank Logical. If FALSE, blank character values
#'   in the target variable are treated as missing. Default is
#'   FALSE.
#'
#' @return A data.frame of class \code{DependencyCheck}
#'   containing row-level dependency diagnostics.
#'
#' @details
#' The dependency rule is interpreted as:
#'
#' If \code{condition_variable == condition_value}, then
#' \code{target_variable} must contain an acceptable value.
#'
#' Missing values in the condition variable do not activate
#' the dependency rule.
#'
#' When \code{allow_blank = FALSE}, empty strings and strings
#' containing only whitespace are treated as unavailable target
#' values.
#'
#' The function does not modify the original data.
#'
#' @examples
#' dat <- data.frame(
#'   Disease_Status = c(
#'     "Positive",
#'     "Negative",
#'     "Positive",
#'     "Positive"
#'   ),
#'   Test_Result = c(
#'     "PCR positive",
#'     NA,
#'     NA,
#'     "Detected"
#'   )
#' )
#'
#' dependency_check(
#'   dat,
#'   condition_variable = "Disease_Status",
#'   condition_value = "Positive",
#'   target_variable = "Test_Result"
#' )
#'
#' @export

dependency_check <- function(
    data,
    condition_variable,
    condition_value,
    target_variable,
    allow_blank = FALSE
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Helper: validate variable name
  #---------------------------------------------
  validate_variable <- function(variable, argument) {

    if (
      !is.character(variable) ||
      length(variable) != 1L ||
      is.na(variable) ||
      !nzchar(variable)
    ) {
      stop(
        sprintf(
          "'%s' must be a single non-empty character string.",
          argument
        ),
        call. = FALSE
      )
    }

    if (!variable %in% names(data)) {
      stop(
        sprintf(
          "Variable '%s' was not found in 'data'.",
          variable
        ),
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  #---------------------------------------------
  # Validate required arguments
  #---------------------------------------------
  if (missing(condition_variable)) {
    stop(
      "'condition_variable' must be supplied.",
      call. = FALSE
    )
  }

  if (missing(condition_value)) {
    stop(
      "'condition_value' must be supplied.",
      call. = FALSE
    )
  }

  if (missing(target_variable)) {
    stop(
      "'target_variable' must be supplied.",
      call. = FALSE
    )
  }

  validate_variable(
    condition_variable,
    "condition_variable"
  )

  validate_variable(
    target_variable,
    "target_variable"
  )

  #---------------------------------------------
  # Variables must differ
  #---------------------------------------------
  if (identical(
    condition_variable,
    target_variable
  )) {
    stop(
      paste0(
        "'condition_variable' and 'target_variable' ",
        "must refer to different variables."
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate condition value
  #---------------------------------------------
  if (
    length(condition_value) != 1L ||
    is.na(condition_value)
  ) {
    stop(
      "'condition_value' must contain exactly one non-missing value.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate allow_blank
  #---------------------------------------------
  if (
    !is.logical(allow_blank) ||
    length(allow_blank) != 1L ||
    is.na(allow_blank)
  ) {
    stop(
      "'allow_blank' must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variables
  #---------------------------------------------
  condition <- data[[condition_variable]]
  target <- data[[target_variable]]

  #---------------------------------------------
  # Determine activated rows
  #---------------------------------------------
  condition_missing <- is.na(condition)

  condition_met <- rep(
    FALSE,
    length(condition)
  )

  comparable <- which(
    !condition_missing
  )

  if (length(comparable) > 0L) {

    condition_met[comparable] <-
      condition[comparable] ==
      condition_value

    condition_met[
      is.na(condition_met)
    ] <- FALSE
  }

  #---------------------------------------------
  # Target missing values
  #---------------------------------------------
  target_missing <- is.na(target)

  #---------------------------------------------
  # Blank target values
  #---------------------------------------------
  target_blank <- rep(
    FALSE,
    length(target)
  )

  if (
    is.character(target) ||
    is.factor(target)
  ) {

    target_character <-
      as.character(target)

    non_missing_target <-
      !target_missing

    target_blank[non_missing_target] <-
      grepl(
        "^[[:space:]]*$",
        target_character[
          non_missing_target
        ]
      )
  }

  #---------------------------------------------
  # Determine target availability
  #---------------------------------------------
  target_available <-
    !target_missing

  if (!allow_blank) {
    target_available <-
      target_available &
      !target_blank
  }

  #---------------------------------------------
  # Dependency violation
  #---------------------------------------------
  dependency_violation <-
    condition_met &
    !target_available

  #---------------------------------------------
  # Valid dependency
  #---------------------------------------------
  valid_dependency <-
    !condition_met |
    target_available

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(condition),
    ConditionValue = condition,
    TargetValue = target,
    ConditionMissing = condition_missing,
    ConditionMet = condition_met,
    TargetMissing = target_missing,
    TargetBlank = target_blank,
    DependencyViolation = dependency_violation,
    ValidDependency = valid_dependency,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "DependencyCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("dependency_check returns correct class", {

  dat <- data.frame(
    Status = c("Yes", "No"),
    Result = c("Positive", NA)
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_s3_class(
    result,
    "DependencyCheck"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("dependency_check returns expected columns", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Positive"
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "ConditionValue",
      "TargetValue",
      "ConditionMissing",
      "ConditionMet",
      "TargetMissing",
      "TargetBlank",
      "DependencyViolation",
      "ValidDependency"
    )
  )
})

#==========================================================
# Condition activation
#==========================================================

test_that("dependency_check identifies activated rows", {

  dat <- data.frame(
    Status = c(
      "Positive",
      "Negative",
      "Positive"
    ),
    Result = c(
      "Detected",
      NA,
      "Detected"
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Positive",
    "Result"
  )

  expect_equal(
    result$ConditionMet,
    c(TRUE, FALSE, TRUE)
  )
})

#==========================================================
# Missing target
#==========================================================

test_that("dependency_check detects missing required target", {

  dat <- data.frame(
    Status = c(
      "Positive",
      "Negative",
      "Positive"
    ),
    Result = c(
      "Detected",
      NA,
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Positive",
    "Result"
  )

  expect_equal(
    result$DependencyViolation,
    c(FALSE, FALSE, TRUE)
  )

  expect_equal(
    result$ValidDependency,
    c(TRUE, TRUE, FALSE)
  )
})

test_that("missing target is allowed when condition is not met", {

  dat <- data.frame(
    Status = c(
      "Negative",
      "Negative"
    ),
    Result = c(
      NA,
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Positive",
    "Result"
  )

  expect_false(
    any(result$DependencyViolation)
  )

  expect_true(
    all(result$ValidDependency)
  )
})

#==========================================================
# Blank target
#==========================================================

test_that("blank target violates dependency by default", {

  dat <- data.frame(
    Status = c(
      "Yes",
      "Yes",
      "Yes"
    ),
    Result = c(
      "",
      "   ",
      "Available"
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_equal(
    result$TargetBlank,
    c(TRUE, TRUE, FALSE)
  )

  expect_equal(
    result$DependencyViolation,
    c(TRUE, TRUE, FALSE)
  )
})

test_that("blank target can be allowed", {

  dat <- data.frame(
    Status = c(
      "Yes",
      "Yes"
    ),
    Result = c(
      "",
      " "
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result",
    allow_blank = TRUE
  )

  expect_false(
    any(result$DependencyViolation)
  )

  expect_true(
    all(result$ValidDependency)
  )
})

#==========================================================
# Missing condition
#==========================================================

test_that("missing condition does not activate dependency", {

  dat <- data.frame(
    Status = c(
      NA,
      "Yes",
      "No"
    ),
    Result = c(
      NA,
      "Available",
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_equal(
    result$ConditionMissing,
    c(TRUE, FALSE, FALSE)
  )

  expect_false(
    result$ConditionMet[1]
  )

  expect_false(
    result$DependencyViolation[1]
  )
})

test_that("all missing conditions are handled", {

  dat <- data.frame(
    Status = c(
      NA_character_,
      NA_character_
    ),
    Result = c(
      NA_character_,
      NA_character_
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_true(
    all(result$ConditionMissing)
  )

  expect_false(
    any(result$ConditionMet)
  )

  expect_false(
    any(result$DependencyViolation)
  )
})

#==========================================================
# Numeric condition
#==========================================================

test_that("dependency_check supports numeric conditions", {

  dat <- data.frame(
    Disease = c(1, 0, 1),
    Result = c(
      "Positive",
      NA,
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Disease",
    1,
    "Result"
  )

  expect_equal(
    result$DependencyViolation,
    c(FALSE, FALSE, TRUE)
  )
})

#==========================================================
# Logical condition
#==========================================================

test_that("dependency_check supports logical conditions", {

  dat <- data.frame(
    Tested = c(
      TRUE,
      FALSE,
      TRUE
    ),
    Result = c(
      "Positive",
      NA,
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Tested",
    TRUE,
    "Result"
  )

  expect_equal(
    result$DependencyViolation,
    c(FALSE, FALSE, TRUE)
  )
})

#==========================================================
# Factor variables
#==========================================================

test_that("dependency_check supports factor condition", {

  dat <- data.frame(
    Status = factor(
      c(
        "Yes",
        "No",
        "Yes"
      )
    ),
    Result = c(
      "Available",
      NA,
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_equal(
    result$DependencyViolation,
    c(FALSE, FALSE, TRUE)
  )
})

test_that("dependency_check detects blanks in factor target", {

  dat <- data.frame(
    Status = c(
      "Yes",
      "Yes"
    ),
    Result = factor(
      c(
        "",
        "Available"
      )
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_equal(
    result$TargetBlank,
    c(TRUE, FALSE)
  )
})

#==========================================================
# Numeric target
#==========================================================

test_that("numeric target does not use blank detection", {

  dat <- data.frame(
    Status = c(
      "Yes",
      "Yes"
    ),
    Result = c(
      10,
      NA
    )
  )

  result <- dependency_check(
    dat,
    "Status",
    "Yes",
    "Result"
  )

  expect_false(
    any(result$TargetBlank)
  )

  expect_equal(
    result$DependencyViolation,
    c(FALSE, TRUE)
  )
})

#==========================================================
# Required arguments
#==========================================================

test_that("condition_variable is required", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      condition_value = "Yes",
      target_variable = "Result"
    ),
    "'condition_variable' must be supplied"
  )
})

test_that("condition_value is required", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      condition_variable = "Status",
      target_variable = "Result"
    ),
    "'condition_value' must be supplied"
  )
})

test_that("target_variable is required", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      condition_variable = "Status",
      condition_value = "Yes"
    ),
    "'target_variable' must be supplied"
  )
})

#==========================================================
# Variable validation
#==========================================================

test_that("condition_variable is validated", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      NULL,
      "Yes",
      "Result"
    ),
    "'condition_variable' must be a single non-empty character string"
  )

  expect_error(
    dependency_check(
      dat,
      "",
      "Yes",
      "Result"
    ),
    "'condition_variable' must be a single non-empty character string"
  )

  expect_error(
    dependency_check(
      dat,
      1,
      "Yes",
      "Result"
    ),
    "'condition_variable' must be a single non-empty character string"
  )

  expect_error(
    dependency_check(
      dat,
      c("Status", "Result"),
      "Yes",
      "Result"
    ),
    "'condition_variable' must be a single non-empty character string"
  )

  expect_error(
    dependency_check(
      dat,
      NA_character_,
      "Yes",
      "Result"
    ),
    "'condition_variable' must be a single non-empty character string"
  )
})

test_that("target_variable is validated", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      NULL
    ),
    "'target_variable' must be a single non-empty character string"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      ""
    ),
    "'target_variable' must be a single non-empty character string"
  )
})

test_that("unknown variables are rejected", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      "Unknown",
      "Yes",
      "Result"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

test_that("condition and target variables must differ", {

  dat <- data.frame(
    Status = "Yes"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      "Status"
    ),
    "must refer to different variables"
  )
})

#==========================================================
# condition_value validation
#==========================================================

test_that("condition_value must be scalar and non-missing", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      NA_character_,
      "Result"
    ),
    "'condition_value' must contain exactly one non-missing value"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      c("Yes", "No"),
      "Result"
    ),
    "'condition_value' must contain exactly one non-missing value"
  )
})

#==========================================================
# allow_blank validation
#==========================================================

test_that("allow_blank must be logical scalar", {

  dat <- data.frame(
    Status = "Yes",
    Result = "Available"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      "Result",
      allow_blank = 1
    ),
    "'allow_blank' must be TRUE or FALSE"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      "Result",
      allow_blank = NA
    ),
    "'allow_blank' must be TRUE or FALSE"
  )

  expect_error(
    dependency_check(
      dat,
      "Status",
      "Yes",
      "Result",
      allow_blank = c(TRUE, FALSE)
    ),
    "'allow_blank' must be TRUE or FALSE"
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("dependency_check rejects invalid data", {

  expect_error(
    dependency_check(
      NULL,
      "Status",
      "Yes",
      "Result"
    )
  )

  expect_error(
    dependency_check(
      1:5,
      "Status",
      "Yes",
      "Result"
    )
  )

  expect_error(
    dependency_check(
      list(
        Status = "Yes",
        Result = "Available"
      ),
      "Status",
      "Yes",
      "Result"
    )
  )

  expect_error(
    dependency_check(
      data.frame(),
      "Status",
      "Yes",
      "Result"
    )
  )
})

#==========================================================
# Summary DataAudit Rule Result
#==========================================================

#' Summarize a DataAudit Rule Result
#'
#' Produces a compact summary of custom validation-rule
#' results generated by \code{\link{apply_audit_rules}}.
#'
#' @param object An object of class
#'   \code{DataAuditRuleResult}.
#' @param ... Additional arguments, currently unused.
#'
#' @return An object of class
#'   \code{summary.DataAuditRuleResult}.
#'
#' @export

summary.DataAuditRuleResult <- function(
    object,
    ...
) {

  #---------------------------------------------
  # Validate object
  #---------------------------------------------
  required_components <- c(
    "status",
    "total_violations",
    "summary",
    "range",
    "category",
    "unique",
    "required",
    "pattern"
  )

  if (
    !inherits(
      object,
      "DataAuditRuleResult"
    ) ||
    !all(
      required_components %in%
      names(object)
    )
  ) {

    stop(
      "Invalid 'DataAuditRuleResult' object.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate summary
  #---------------------------------------------
  if (!is.data.frame(object$summary)) {

    stop(
      "Invalid rule summary in 'DataAuditRuleResult'.",
      call. = FALSE
    )
  }

  required_summary <- c(
    "Rule",
    "Violations"
  )

  if (
    !all(
      required_summary %in%
      names(object$summary)
    )
  ) {

    stop(
      "Invalid rule summary in 'DataAuditRuleResult'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Helper: collect violation rows
  #---------------------------------------------
  collect_violations <- function(
    group,
    rule_name
  ) {

    if (length(group) == 0L) {

      return(
        data.frame(
          Rule = character(),
          Variable = character(),
          Row = integer(),
          Value = character(),
          stringsAsFactors = FALSE
        )
      )
    }

    pieces <- list()

    for (variable in names(group)) {

      x <- group[[variable]]

      if (
        !is.data.frame(x) ||
        !"Violation" %in% names(x)
      ) {
        next
      }

      bad <- which(
        !is.na(x$Violation) &
          x$Violation
      )

      if (length(bad) == 0L) {
        next
      }

      value <- if (
        "Value" %in% names(x)
      ) {
        as.character(
          x$Value[bad]
        )
      } else {
        rep(
          NA_character_,
          length(bad)
        )
      }

      row_number <- if (
        "Row" %in% names(x)
      ) {
        as.integer(
          x$Row[bad]
        )
      } else {
        as.integer(bad)
      }

      pieces[[length(pieces) + 1L]] <-
        data.frame(
          Rule = rule_name,
          Variable = variable,
          Row = row_number,
          Value = value,
          stringsAsFactors = FALSE
        )
    }

    if (length(pieces) == 0L) {

      return(
        data.frame(
          Rule = character(),
          Variable = character(),
          Row = integer(),
          Value = character(),
          stringsAsFactors = FALSE
        )
      )
    }

    do.call(
      rbind,
      pieces
    )
  }

  #---------------------------------------------
  # Collect violations
  #---------------------------------------------
  violations <- rbind(
    collect_violations(
      object$range,
      "Range"
    ),
    collect_violations(
      object$category,
      "Category"
    ),
    collect_violations(
      object$unique,
      "Unique"
    ),
    collect_violations(
      object$required,
      "Required"
    ),
    collect_violations(
      object$pattern,
      "Pattern"
    )
  )

  rownames(violations) <- NULL

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- list(
    status = object$status,
    total_violations =
      object$total_violations,
    by_rule = object$summary,
    violations = violations
  )

  class(result) <- c(
    "summary.DataAuditRuleResult",
    "list"
  )

  result
}

#==========================================================
# Data Audit Score
#==========================================================

#' Calculate a Data Quality Score
#'
#' Calculates a standardized data-quality score from a
#' \code{DataAuditReport} produced by \code{\link{audit_data}}.
#'
#' @param x An object of class \code{DataAuditReport}.
#'
#' @return An object of class \code{DataAuditScore}.
#'
#' @details
#' The score ranges from 0 to 100, where higher values
#' indicate better data quality.
#'
#' The score is based on automatic audit issues and custom
#' rule violations relative to the number of observations
#' and variables in the audited dataset.
#'
#' @examples
#' dat <- data.frame(
#'   ID = c(1, 2, 2, 4),
#'   Value = c(10, NA, Inf, 20)
#' )
#'
#' report <- audit_data(dat)
#' audit_score(report)
#'
#' @export

audit_score <- function(x) {

  #---------------------------------------------
  # Validate object
  #---------------------------------------------
  if (!inherits(x, "DataAuditReport")) {

    stop(
      "'x' must be a DataAuditReport object.",
      call. = FALSE
    )
  }

  required_components <- c(
    "variables",
    "summary"
  )

  if (
    !all(
      required_components %in%
      names(x)
    )
  ) {

    stop(
      "Invalid 'DataAuditReport' object.",
      call. = FALSE
    )
  }

  if (
    !is.data.frame(x$summary) ||
    nrow(x$summary) != 1L
  ) {

    stop(
      "Invalid summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  required_summary <- c(
    "Rows",
    "Variables",
    "TotalIssues",
    "RuleViolations",
    "OverallIssues"
  )

  if (
    !all(
      required_summary %in%
      names(x$summary)
    )
  ) {

    stop(
      "Invalid summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract values
  #---------------------------------------------
  n_rows <- x$summary$Rows[[1L]]
  n_variables <- x$summary$Variables[[1L]]
  automatic_issues <-
    x$summary$TotalIssues[[1L]]
  rule_violations <-
    x$summary$RuleViolations[[1L]]
  overall_issues <-
    x$summary$OverallIssues[[1L]]

  #---------------------------------------------
  # Validate values
  #---------------------------------------------
  values <- c(
    n_rows,
    n_variables,
    automatic_issues,
    rule_violations,
    overall_issues
  )

  if (
    any(!is.numeric(values)) ||
    any(is.na(values)) ||
    any(!is.finite(values)) ||
    n_rows <= 0 ||
    n_variables <= 0 ||
    automatic_issues < 0 ||
    rule_violations < 0 ||
    overall_issues < 0
  ) {

    stop(
      "Invalid values in DataAudit summary.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Number of data cells
  #---------------------------------------------
  total_cells <-
    n_rows * n_variables

  #---------------------------------------------
  # Issue burden
  #
  # log1p prevents very large numbers of
  # violations from producing unstable scores.
  #---------------------------------------------
  burden <-
    overall_issues / total_cells

  #---------------------------------------------
  # Convert burden to 0-100 score
  #---------------------------------------------
  score <-
    100 / (1 + burden)

  score <- max(
    0,
    min(
      100,
      score
    )
  )

  #---------------------------------------------
  # Quality category
  #---------------------------------------------
  quality <- if (score >= 95) {

    "Excellent"

  } else if (score >= 85) {

    "Good"

  } else if (score >= 70) {

    "Moderate"

  } else if (score >= 50) {

    "Poor"

  } else {

    "Critical"
  }

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- list(
    score = as.numeric(score),
    quality = quality,
    automatic_issues =
      as.integer(automatic_issues),
    rule_violations =
      as.integer(rule_violations),
    overall_issues =
      as.integer(overall_issues),
    rows = as.integer(n_rows),
    variables = as.integer(n_variables),
    total_cells =
      as.numeric(total_cells)
  )

  class(result) <- c(
    "DataAuditScore",
    "list"
  )

  result
}

library(testthat)
library(DataAudit)


test_that("audit_score returns correct class", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- audit_score(report)

  expect_s3_class(
    result,
    "DataAuditScore"
  )

  expect_type(
    result,
    "list"
  )
})


test_that("audit_score returns expected components", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- audit_score(report)

  expect_equal(
    names(result),
    c(
      "score",
      "quality",
      "automatic_issues",
      "rule_violations",
      "overall_issues",
      "rows",
      "variables",
      "total_cells"
    )
  )
})


test_that("clean data receives score 100", {

  report <- audit_data(
    data.frame(
      x = 1:10,
      y = letters[1:10]
    ),
    include_nzv = FALSE
  )

  result <- audit_score(report)

  expect_equal(
    result$score,
    100
  )

  expect_equal(
    result$quality,
    "Excellent"
  )
})


test_that("issues reduce audit score", {

  clean <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  dirty <- audit_data(
    data.frame(
      x = c(
        1,
        NA,
        Inf,
        4,
        5
      )
    ),
    include_nzv = FALSE
  )

  expect_lt(
    audit_score(dirty)$score,
    audit_score(clean)$score
  )
})


test_that("rule violations reduce score", {

  dat <- data.frame(
    Age = c(
      5,
      40,
      -2
    )
  )

  without_rules <- audit_data(
    dat,
    include_nzv = FALSE
  )

  with_rules <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  expect_lt(
    audit_score(with_rules)$score,
    audit_score(without_rules)$score
  )
})


test_that("score remains between zero and 100", {

  report <- audit_data(
    data.frame(
      x = c(
        NA,
        Inf,
        -Inf
      )
    ),
    include_nzv = TRUE
  )

  result <- audit_score(report)

  expect_gte(
    result$score,
    0
  )

  expect_lte(
    result$score,
    100
  )
})


test_that("audit_score reports dataset dimensions", {

  report <- audit_data(
    data.frame(
      x = 1:5,
      y = letters[1:5]
    ),
    include_nzv = FALSE
  )

  result <- audit_score(report)

  expect_equal(
    result$rows,
    5L
  )

  expect_equal(
    result$variables,
    2L
  )

  expect_equal(
    result$total_cells,
    10
  )
})


test_that("audit_score reports issue counts", {

  report <- audit_data(
    data.frame(
      x = c(
        1,
        NA,
        Inf
      )
    ),
    include_nzv = FALSE
  )

  result <- audit_score(report)

  expect_equal(
    result$automatic_issues,
    report$summary$TotalIssues
  )

  expect_equal(
    result$overall_issues,
    report$summary$OverallIssues
  )
})


test_that("audit_score rejects non-report object", {

  expect_error(
    audit_score(
      data.frame(
        x = 1:5
      )
    ),
    "'x' must be a DataAuditReport object"
  )
})


test_that("audit_score rejects malformed report", {

  x <- list()

  class(x) <- c(
    "DataAuditReport",
    "list"
  )

  expect_error(
    audit_score(x),
    "Invalid 'DataAuditReport' object"
  )
})


test_that("audit_score rejects invalid summary", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$summary <- "invalid"

  expect_error(
    audit_score(report),
    "Invalid summary in 'DataAuditReport'"
  )
})


test_that("audit_score rejects summary missing fields", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$summary <- data.frame(
    Rows = 5
  )

  expect_error(
    audit_score(report),
    "Invalid summary in 'DataAuditReport'"
  )
})


test_that("audit_score rejects invalid summary values", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$summary$Rows <- 0

  expect_error(
    audit_score(report),
    "Invalid values in DataAudit summary"
  )
})
test_that("quality categories use expected thresholds", {

  make_report <- function(score_target) {

    report <- audit_data(
      data.frame(
        x = 1:100
      ),
      include_nzv = FALSE
    )

    burden <-
      100 / score_target - 1

    issues <-
      burden * 100

    report$summary$TotalIssues <-
      issues

    report$summary$RuleViolations <-
      0

    report$summary$OverallIssues <-
      issues

    report
  }

  expect_equal(
    audit_score(
      make_report(95)
    )$quality,
    "Excellent"
  )

  expect_equal(
    audit_score(
      make_report(85)
    )$quality,
    "Good"
  )

  expect_equal(
    audit_score(
      make_report(70)
    )$quality,
    "Moderate"
  )

  expect_equal(
    audit_score(
      make_report(50)
    )$quality,
    "Poor"
  )

  expect_equal(
    audit_score(
      make_report(40)
    )$quality,
    "Critical"
  )
})

library(testthat)
library(DataAudit)


test_that("print DataAuditScore returns object invisibly", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  score <- audit_score(report)

  result <- withVisible(
    print(score)
  )

  expect_false(
    result$visible
  )

  expect_identical(
    result$value,
    score
  )
})


test_that("print displays score header", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5
      ),
      include_nzv = FALSE
    )
  )

  expect_output(
    print(score),
    "DataAudit Quality Score"
  )
})


test_that("print displays score and quality", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5
      ),
      include_nzv = FALSE
    )
  )

  output <- capture.output(
    print(score)
  )

  expect_true(
    any(
      grepl(
        "Score:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "100.00 / 100",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Quality:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Excellent",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print displays dataset information", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5,
        y = letters[1:5]
      ),
      include_nzv = FALSE
    )
  )

  output <- capture.output(
    print(score)
  )

  expect_true(
    any(
      grepl(
        "Dataset",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Rows:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Variables:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Data cells:",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print displays issue information", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = c(
          1,
          NA,
          Inf
        )
      ),
      include_nzv = FALSE
    )
  )

  output <- capture.output(
    print(score)
  )

  expect_true(
    any(
      grepl(
        "Automatic issues:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Rule violations:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Overall issues:",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print displays custom rule violations", {

  dat <- data.frame(
    Age = c(
      5,
      40,
      -2
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  score <- audit_score(report)

  output <- capture.output(
    print(score)
  )

  expect_true(
    any(
      grepl(
        "Rule violations:     2",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print rejects malformed DataAuditScore", {

  x <- list()

  class(x) <- c(
    "DataAuditScore",
    "list"
  )

  expect_error(
    print(x),
    "Invalid 'DataAuditScore' object"
  )
})


test_that("print rejects object without DataAuditScore class", {

  x <- list(
    score = 100,
    quality = "Excellent",
    automatic_issues = 0L,
    rule_violations = 0L,
    overall_issues = 0L,
    rows = 5L,
    variables = 1L,
    total_cells = 5
  )

  expect_error(
    print.DataAuditScore(x),
    "Invalid 'DataAuditScore' object"
  )
})


test_that("print rejects invalid score", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5
      ),
      include_nzv = FALSE
    )
  )

  score$score <- NA_real_

  expect_error(
    print(score),
    "Invalid score in 'DataAuditScore'"
  )
})


test_that("print rejects score above 100", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5
      ),
      include_nzv = FALSE
    )
  )

  score$score <- 101

  expect_error(
    print(score),
    "Invalid score in 'DataAuditScore'"
  )
})


test_that("print rejects score below zero", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5
      ),
      include_nzv = FALSE
    )
  )

  score$score <- -1

  expect_error(
    print(score),
    "Invalid score in 'DataAuditScore'"
  )
})


test_that("print rejects invalid quality category", {

  score <- audit_score(
    audit_data(
      data.frame(
        x = 1:5
      ),
      include_nzv = FALSE
    )
  )

  score$quality <- "Unknown"

  expect_error(
    print(score),
    "Invalid quality category in 'DataAuditScore'"
  )
})

#==========================================================
# Internal Validation Functions
#==========================================================


#==========================================================
# Validate Data
#==========================================================

#' Validate Data
#'
#' Internal validation function used throughout DataAudit.
#'
#' @param data An object expected to be a data.frame.
#'
#' @return Invisibly returns \code{TRUE} when validation
#'   succeeds.
#'
#' @keywords internal
#' @noRd

validate_data <- function(data) {

  #---------------------------------------------
  # NULL
  #---------------------------------------------
  if (is.null(data)) {
    stop(
      "'data' cannot be NULL.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Data frame
  #---------------------------------------------
  if (!is.data.frame(data)) {
    stop(
      "'data' must be a data.frame.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # At least one variable
  #---------------------------------------------
  if (ncol(data) == 0L) {
    stop(
      "'data' must contain at least one variable.",
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#==========================================================
# Validate Column
#==========================================================

#' Validate Column
#'
#' Internal validation function for a column name.
#'
#' @param data A data.frame.
#' @param column A column name.
#'
#' @return Invisibly returns \code{TRUE} when validation
#'   succeeds.
#'
#' @keywords internal
#' @noRd

.validate_column <- function(data, column) {

  validate_data(data)

  if (
    !is.character(column) ||
    length(column) != 1L ||
    is.na(column)
  ) {
    stop(
      "'column' must be a single non-missing character string.",
      call. = FALSE
    )
  }

  if (!nzchar(column)) {
    stop(
      "'column' cannot be empty.",
      call. = FALSE
    )
  }

  if (!column %in% names(data)) {
    stop(
      sprintf(
        "Column '%s' was not found in 'data'.",
        column
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#==========================================================
# Validate Numeric Scalar
#==========================================================

#' Validate Numeric Scalar
#'
#' Internal validation function for finite numeric scalar
#' arguments.
#'
#' @param x Object to validate.
#' @param name Argument name used in the error message.
#'
#' @return Invisibly returns \code{TRUE}.
#'
#' @keywords internal
#' @noRd

.validate_numeric_scalar <- function(x, name) {

  if (
    !is.numeric(x) ||
    length(x) != 1L ||
    is.na(x) ||
    !is.finite(x)
  ) {
    stop(
      sprintf(
        "'%s' must be a single finite numeric value.",
        name
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#==========================================================
# Validate Logical Scalar
#==========================================================

#' Validate Logical Scalar
#'
#' @param x Object to validate.
#' @param name Argument name.
#'
#' @return Invisibly returns \code{TRUE}.
#'
#' @keywords internal
#' @noRd

.validate_logical_scalar <- function(x, name) {

  if (
    !is.logical(x) ||
    length(x) != 1L ||
    is.na(x)
  ) {
    stop(
      sprintf(
        "'%s' must be TRUE or FALSE.",
        name
      ),
      call. = FALSE
    )
  }

  invisible(TRUE)
}


#==========================================================
# Detect Blank Strings
#==========================================================

#' Detect Blank Strings
#'
#' Internal helper for identifying empty or whitespace-only
#' character values.
#'
#' @param x A vector.
#'
#' @return A logical vector.
#'
#' @keywords internal
#' @noRd

.is_blank <- function(x) {

  if (!is.character(x)) {
    return(
      rep(FALSE, length(x))
    )
  }

  !is.na(x) & trimws(x) == ""
}

library(testthat)
library(DataAudit)

#==========================================================
# validate_data()
#==========================================================

test_that("validate_data accepts data.frame", {

  expect_true(
    DataAudit:::validate_data(iris)
  )

})

test_that("validate_data accepts one-column data.frame", {

  dat <- data.frame(
    x = 1:10
  )

  expect_true(
    DataAudit:::validate_data(dat)
  )

})

test_that("validate_data accepts zero-row data.frame", {

  dat <- data.frame(
    x = numeric(0)
  )

  expect_true(
    DataAudit:::validate_data(dat)
  )

})

test_that("validate_data accepts tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::as_tibble(iris)

  expect_true(
    DataAudit:::validate_data(dat)
  )

})

test_that("validate_data rejects NULL", {

  expect_error(
    DataAudit:::validate_data(NULL),
    "'data' cannot be NULL"
  )

})

test_that("validate_data rejects numeric vector", {

  expect_error(
    DataAudit:::validate_data(1:10),
    "'data' must be a data.frame"
  )

})

test_that("validate_data rejects character vector", {

  expect_error(
    DataAudit:::validate_data(letters),
    "'data' must be a data.frame"
  )

})

test_that("validate_data rejects list", {

  expect_error(
    DataAudit:::validate_data(
      list(x = 1:10)
    ),
    "'data' must be a data.frame"
  )

})

test_that("validate_data rejects data.frame without variables", {

  expect_error(
    DataAudit:::validate_data(
      data.frame()
    ),
    "at least one variable"
  )

})


#==========================================================
# .validate_column()
#==========================================================

test_that(".validate_column accepts valid column", {

  expect_true(
    DataAudit:::.validate_column(
      iris,
      "Sepal.Length"
    )
  )

})

test_that(".validate_column rejects NULL column", {

  expect_error(
    DataAudit:::.validate_column(
      iris,
      NULL
    ),
    "single non-missing character string"
  )

})

test_that(".validate_column rejects numeric column argument", {

  expect_error(
    DataAudit:::.validate_column(
      iris,
      1
    ),
    "single non-missing character string"
  )

})

test_that(".validate_column rejects multiple names", {

  expect_error(
    DataAudit:::.validate_column(
      iris,
      c("Sepal.Length", "Sepal.Width")
    ),
    "single non-missing character string"
  )

})

test_that(".validate_column rejects NA", {

  expect_error(
    DataAudit:::.validate_column(
      iris,
      NA_character_
    ),
    "single non-missing character string"
  )

})

test_that(".validate_column rejects empty name", {

  expect_error(
    DataAudit:::.validate_column(
      iris,
      ""
    ),
    "'column' cannot be empty"
  )

})

test_that(".validate_column rejects unknown column", {

  expect_error(
    DataAudit:::.validate_column(
      iris,
      "ABC"
    ),
    "was not found"
  )

})


#==========================================================
# .validate_numeric_scalar()
#==========================================================

test_that(".validate_numeric_scalar accepts numeric scalar", {

  expect_true(
    DataAudit:::.validate_numeric_scalar(
      1.5,
      "cutoff"
    )
  )

})

test_that(".validate_numeric_scalar accepts integer scalar", {

  expect_true(
    DataAudit:::.validate_numeric_scalar(
      2L,
      "cutoff"
    )
  )

})

test_that(".validate_numeric_scalar rejects character value", {

  expect_error(
    DataAudit:::.validate_numeric_scalar(
      "1",
      "cutoff"
    ),
    "'cutoff' must be a single finite numeric value"
  )

})

test_that(".validate_numeric_scalar rejects multiple values", {

  expect_error(
    DataAudit:::.validate_numeric_scalar(
      c(1, 2),
      "cutoff"
    ),
    "single finite numeric value"
  )

})

test_that(".validate_numeric_scalar rejects NA", {

  expect_error(
    DataAudit:::.validate_numeric_scalar(
      NA_real_,
      "cutoff"
    ),
    "single finite numeric value"
  )

})

test_that(".validate_numeric_scalar rejects Inf", {

  expect_error(
    DataAudit:::.validate_numeric_scalar(
      Inf,
      "cutoff"
    ),
    "single finite numeric value"
  )

})

test_that(".validate_numeric_scalar rejects NaN", {

  expect_error(
    DataAudit:::.validate_numeric_scalar(
      NaN,
      "cutoff"
    ),
    "single finite numeric value"
  )

})


#==========================================================
# .validate_logical_scalar()
#==========================================================

test_that(".validate_logical_scalar accepts TRUE and FALSE", {

  expect_true(
    DataAudit:::.validate_logical_scalar(
      TRUE,
      "remove"
    )
  )

  expect_true(
    DataAudit:::.validate_logical_scalar(
      FALSE,
      "remove"
    )
  )

})

test_that(".validate_logical_scalar rejects numeric", {

  expect_error(
    DataAudit:::.validate_logical_scalar(
      1,
      "remove"
    ),
    "'remove' must be TRUE or FALSE"
  )

})

test_that(".validate_logical_scalar rejects NA", {

  expect_error(
    DataAudit:::.validate_logical_scalar(
      NA,
      "remove"
    ),
    "'remove' must be TRUE or FALSE"
  )

})

test_that(".validate_logical_scalar rejects multiple values", {

  expect_error(
    DataAudit:::.validate_logical_scalar(
      c(TRUE, FALSE),
      "remove"
    ),
    "'remove' must be TRUE or FALSE"
  )

})


#==========================================================
# .is_blank()
#==========================================================

test_that(".is_blank detects empty strings", {

  x <- c(
    "",
    " ",
    "   ",
    "A",
    NA
  )

  result <- DataAudit:::.is_blank(x)

  expect_equal(
    result,
    c(
      TRUE,
      TRUE,
      TRUE,
      FALSE,
      FALSE
    )
  )

})

test_that(".is_blank returns FALSE for numeric values", {

  expect_equal(
    DataAudit:::.is_blank(1:5),
    rep(FALSE, 5)
  )

})

test_that(".is_blank returns FALSE for logical values", {

  expect_equal(
    DataAudit:::.is_blank(
      c(TRUE, FALSE, NA)
    ),
    c(FALSE, FALSE, FALSE)
  )

})

test_that(".is_blank handles zero-length character vector", {

  expect_length(
    DataAudit:::.is_blank(
      character(0)
    ),
    0
  )

})

#==========================================================
# Print DataAudit Report
#==========================================================

#' Print a DataAudit Report
#'
#' Prints a concise summary of a data audit produced by
#' \code{\link{audit_data}}.
#'
#' @param x An object of class \code{DataAuditReport}.
#' @param ... Additional arguments, currently unused.
#'
#' @return The input object \code{x}, invisibly.
#'
#' @export
print.DataAuditReport <- function(
    x,
    ...
) {

  #---------------------------------------------
  # Validate object
  #---------------------------------------------
  required_components <- c(
    "overview",
    "variables",
    "duplicates",
    "summary",
    "rules"
  )

  if (
    !inherits(x, "DataAuditReport") ||
    !all(required_components %in% names(x))
  ) {

    stop(
      "Invalid 'DataAuditReport' object.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate summary
  #---------------------------------------------
  if (!is.data.frame(x$summary)) {

    stop(
      "Invalid summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  required_summary <- c(
    "Rows",
    "Variables",
    "VariablesWithIssues",
    "DuplicateRows",
    "TotalIssues",
    "RuleViolations",
    "OverallIssues"
  )

  if (
    !all(
      required_summary %in%
      names(x$summary)
    )
  ) {

    stop(
      "Invalid summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate variable summary
  #---------------------------------------------
  if (!is.data.frame(x$variables)) {

    stop(
      "Invalid variable summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  if (!"Issue" %in% names(x$variables)) {

    stop(
      "Invalid variable summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate duplicates
  #---------------------------------------------
  if (!is.data.frame(x$duplicates)) {

    stop(
      "Invalid duplicate summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  if (!"Duplicate" %in% names(x$duplicates)) {

    stop(
      "Invalid duplicate summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract summary
  #---------------------------------------------
  s <- x$summary[1L, ]

  #---------------------------------------------
  # Header
  #---------------------------------------------
  cat(
    "DataAudit Report\n"
  )

  cat(
    "================\n\n"
  )

  #---------------------------------------------
  # Dataset summary
  #---------------------------------------------
  cat(
    sprintf(
      "Rows:                  %s\n",
      s$Rows
    )
  )

  cat(
    sprintf(
      "Variables:             %s\n",
      s$Variables
    )
  )

  cat(
    sprintf(
      "Variables with issues: %s\n",
      s$VariablesWithIssues
    )
  )

  cat(
    sprintf(
      "Duplicated rows:       %s\n",
      s$DuplicateRows
    )
  )

  cat(
    sprintf(
      "Total issues:          %s\n",
      s$TotalIssues
    )
  )

  cat(
    sprintf(
      "Rule violations:       %s\n",
      s$RuleViolations
    )
  )

  cat(
    sprintf(
      "Overall issues:        %s\n",
      s$OverallIssues
    )
  )

  #---------------------------------------------
  # Variables with issues
  #---------------------------------------------
  issue_variables <-
    x$variables[
      x$variables$Issue,
      ,
      drop = FALSE
    ]

  if (nrow(issue_variables) > 0L) {

    cat(
      "\nVariables with issues\n"
    )

    cat(
      "---------------------\n"
    )

    print(
      issue_variables,
      row.names = FALSE
    )

  } else {

    cat(
      "\nNo variable-level issues detected.\n"
    )
  }

  #---------------------------------------------
  # Duplicate rows
  #---------------------------------------------
  duplicate_rows <-
    x$duplicates[
      x$duplicates$Duplicate,
      ,
      drop = FALSE
    ]

  if (nrow(duplicate_rows) > 0L) {

    cat(
      "\nDuplicated rows\n"
    )

    cat(
      "---------------\n"
    )

    print(
      duplicate_rows,
      row.names = FALSE
    )

  } else {

    cat(
      "\nNo duplicated rows detected.\n"
    )
  }

  #---------------------------------------------
  # Custom rule validation
  #---------------------------------------------
  if (!is.null(x$rules)) {

    if (
      !inherits(
        x$rules,
        "DataAuditRuleResult"
      )
    ) {

      stop(
        "Invalid rule results in 'DataAuditReport'.",
        call. = FALSE
      )
    }

    cat(
      "\nCustom rule validation\n"
    )

    cat(
      "----------------------\n"
    )

    print(
      x$rules$summary,
      row.names = FALSE
    )

    cat(
      sprintf(
        "\nTotal rule violations: %s\n",
        x$rules$total_violations
      )
    )
  }

  invisible(x)
}

library(testthat)
library(DataAudit)


#==========================================================
# Basic print behaviour
#==========================================================

test_that("print method returns object invisibly", {

  dat <- data.frame(
    x = 1:5
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE
  )

  result <- withVisible(
    print(report)
  )

  expect_false(
    result$visible
  )

  expect_identical(
    result$value,
    report
  )
})


test_that("print method displays report header", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  expect_output(
    print(report),
    "DataAudit Report"
  )
})


test_that("print method displays dimensions", {

  report <- audit_data(
    data.frame(
      x = 1:4,
      y = letters[1:4]
    ),
    include_nzv = FALSE
  )

  output <- capture.output(
    print(report)
  )

  expect_true(
    any(
      grepl(
        "Rows:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Variables:",
        output,
        fixed = TRUE
      )
    )
  )
})


#==========================================================
# Automatic audit results
#==========================================================

test_that("print method reports variables with issues", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      3
    ),
    y = 1:3
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE
  )

  output <- capture.output(
    print(report)
  )

  expect_true(
    any(
      grepl(
        "Variables with issues",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "x",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method reports no variable issues", {

  dat <- data.frame(
    x = 1:5,
    y = letters[1:5],
    stringsAsFactors = FALSE
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_output(
    print(report),
    "No variable-level issues detected"
  )
})


test_that("print method reports duplicate rows", {

  dat <- data.frame(
    x = c(
      1,
      2,
      2,
      3
    ),
    y = c(
      "A",
      "B",
      "B",
      "C"
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE
  )

  output <- capture.output(
    print(report)
  )

  expect_true(
    any(
      grepl(
        "Duplicated rows",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "2",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "3",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method reports absence of duplicates", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  expect_output(
    print(report),
    "No duplicated rows detected"
  )
})


test_that("print method displays issue counts", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      Inf
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE
  )

  output <- capture.output(
    print(report)
  )

  expect_true(
    any(
      grepl(
        "Total issues:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Rule violations:",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Overall issues:",
        output,
        fixed = TRUE
      )
    )
  )
})


#==========================================================
# Custom audit rules
#==========================================================

test_that("print method displays custom rule validation", {

  dat <- data.frame(
    Age = c(
      5,
      40,
      -2
    )
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = rules
  )

  output <- capture.output(
    print(report)
  )

  expect_true(
    any(
      grepl(
        "Custom rule validation",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Total rule violations: 2",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method displays rule summary", {

  dat <- data.frame(
    Age = c(
      5,
      40
    )
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = rules
  )

  output <- capture.output(
    print(report)
  )

  expect_true(
    any(
      grepl(
        "Range",
        output,
        fixed = TRUE
      )
    )
  )

  expect_true(
    any(
      grepl(
        "Violations",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method omits custom rule section without rules", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  output <- capture.output(
    print(report)
  )

  expect_false(
    any(
      grepl(
        "Custom rule validation",
        output,
        fixed = TRUE
      )
    )
  )
})


test_that("print method displays multiple rule types", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002"
    ),
    Species = c(
      "Dog",
      "Horse",
      ""
    ),
    Age = c(
      5,
      40,
      -2
    ),
    stringsAsFactors = FALSE
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    ),
    category = list(
      Species = c(
        "Dog",
        "Cat"
      )
    ),
    unique = "ID",
    required = "Species",
    pattern = list(
      ID = "^A[0-9]{3}$"
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = rules
  )

  output <- capture.output(
    print(report)
  )

  for (
    rule_name in c(
      "Range",
      "Category",
      "Unique",
      "Required",
      "Pattern"
    )
  ) {

    expect_true(
      any(
        grepl(
          rule_name,
          output,
          fixed = TRUE
        )
      )
    )
  }
})


#==========================================================
# Defensive validation
#==========================================================

test_that("print method rejects object without class", {

  x <- list(
    overview = data.frame(),
    variables = data.frame(),
    duplicates = data.frame(),
    summary = data.frame(),
    rules = NULL
  )

  expect_error(
    print.DataAuditReport(x),
    "Invalid 'DataAuditReport' object"
  )
})


test_that("print method rejects malformed report", {

  x <- list(
    overview = data.frame()
  )

  class(x) <- c(
    "DataAuditReport",
    "list"
  )

  expect_error(
    print(x),
    "Invalid 'DataAuditReport' object"
  )
})


test_that("print method rejects non-data-frame summary", {

  x <- list(
    overview = data.frame(),
    variables = data.frame(
      Issue = logical()
    ),
    duplicates = data.frame(
      Row = integer(),
      Duplicate = logical()
    ),
    summary = "invalid",
    rules = NULL
  )

  class(x) <- c(
    "DataAuditReport",
    "list"
  )

  expect_error(
    print(x),
    "Invalid summary in 'DataAuditReport'"
  )
})


test_that("print method rejects summary missing required columns", {

  x <- list(
    overview = data.frame(),
    variables = data.frame(
      Issue = logical()
    ),
    duplicates = data.frame(
      Row = integer(),
      Duplicate = logical()
    ),
    summary = data.frame(
      Wrong = 1
    ),
    rules = NULL
  )

  class(x) <- c(
    "DataAuditReport",
    "list"
  )

  expect_error(
    print(x),
    "Invalid summary in 'DataAuditReport'"
  )
})


test_that("print method rejects invalid variable summary", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$variables <- "invalid"

  expect_error(
    print(report),
    "Invalid variable summary in 'DataAuditReport'"
  )
})


test_that("print method rejects variable summary without Issue", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$variables <- data.frame(
    Variable = "x"
  )

  expect_error(
    print(report),
    "Invalid variable summary in 'DataAuditReport'"
  )
})


test_that("print method rejects invalid duplicate summary", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$duplicates <- "invalid"

  expect_error(
    print(report),
    "Invalid duplicate summary in 'DataAuditReport'"
  )
})


test_that("print method rejects duplicate summary without Duplicate", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$duplicates <- data.frame(
    Row = 1:5
  )

  expect_error(
    print(report),
    "Invalid duplicate summary in 'DataAuditReport'"
  )
})


test_that("print method rejects invalid rule result", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$rules <- list(
    invalid = TRUE
  )

  expect_error(
    print(report),
    "Invalid rule results in 'DataAuditReport'"
  )
})

#==========================================================
# Identifier Check
#==========================================================

#' Check Identifier Integrity
#'
#' Checks an identifier variable for missing values, blank
#' values, duplicates, leading or trailing whitespace, and
#' optionally values that do not match a regular expression.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   identifier variable to check.
#' @param pattern Optional regular expression that valid
#'   identifiers must match.
#'
#' @return A data.frame of class \code{IDCheck} containing
#'   row-level identifier diagnostics.
#'
#' @details
#' Character, factor, numeric, and integer identifier
#' variables are supported.
#'
#' Duplicate detection excludes missing and blank values.
#'
#' Whitespace checks are performed on the character
#' representation of the identifier.
#'
#' If \code{pattern} is supplied, non-missing and non-blank
#' identifiers that do not match the pattern are flagged.
#'
#' The function does not modify the original data.
#'
#' @examples
#' dat <- data.frame(
#'   Animal_ID = c(
#'     "A001",
#'     "A002",
#'     "A002",
#'     " A003",
#'     "",
#'     NA
#'   )
#' )
#'
#' id_check(
#'   dat,
#'   variable = "Animal_ID",
#'   pattern = "^A[0-9]{3}$"
#' )
#'
#' @export

id_check <- function(
    data,
    variable,
    pattern = NULL
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (
    missing(variable) ||
    !is.character(variable) ||
    length(variable) != 1L ||
    is.na(variable) ||
    !nzchar(variable)
  ) {
    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {
    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate pattern
  #---------------------------------------------
  if (!is.null(pattern)) {

    if (
      !is.character(pattern) ||
      length(pattern) != 1L ||
      is.na(pattern) ||
      !nzchar(pattern)
    ) {
      stop(
        "'pattern' must be NULL or a single non-empty character string.",
        call. = FALSE
      )
    }

    pattern_valid <- tryCatch(
      {
        suppressWarnings(
          grepl(
            pattern,
            "",
            perl = TRUE
          )
        )
        TRUE
      },
      warning = function(w) {
        FALSE
      },
      error = function(e) {
        FALSE
      }
    )

    if (!pattern_valid) {
      stop(
        "'pattern' is not a valid regular expression.",
        call. = FALSE
      )
    }
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Validate identifier type
  #---------------------------------------------
  if (
    !is.character(x) &&
    !is.factor(x) &&
    !is.numeric(x)
  ) {
    stop(
      paste0(
        "'variable' must refer to a character, factor, ",
        "numeric, or integer variable."
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Character representation
  #---------------------------------------------
  value <- as.character(x)

  #---------------------------------------------
  # Missing
  #---------------------------------------------
  missing_value <- is.na(value)

  #---------------------------------------------
  # Blank
  #---------------------------------------------
  blank <- rep(
    FALSE,
    length(value)
  )

  blank[!missing_value] <-
    value[!missing_value] == ""

  #---------------------------------------------
  # Whitespace only
  #---------------------------------------------
  whitespace_only <- rep(
    FALSE,
    length(value)
  )

  whitespace_only[!missing_value] <-
    grepl(
      "^[[:space:]]+$",
      value[!missing_value]
    )

  #---------------------------------------------
  # Leading whitespace
  #---------------------------------------------
  leading_whitespace <- rep(
    FALSE,
    length(value)
  )

  leading_whitespace[!missing_value] <-
    grepl(
      "^[[:space:]]+",
      value[!missing_value]
    )

  #---------------------------------------------
  # Trailing whitespace
  #---------------------------------------------
  trailing_whitespace <- rep(
    FALSE,
    length(value)
  )

  trailing_whitespace[!missing_value] <-
    grepl(
      "[[:space:]]+$",
      value[!missing_value]
    )

  #---------------------------------------------
  # Duplicate identifiers
  #---------------------------------------------
  duplicate <- rep(
    FALSE,
    length(value)
  )

  duplicate_candidates <-
    !missing_value &
    !blank &
    !whitespace_only

  duplicate[duplicate_candidates] <-
    duplicated(
      value[duplicate_candidates]
    ) |
    duplicated(
      value[duplicate_candidates],
      fromLast = TRUE
    )

  #---------------------------------------------
  # Pattern mismatch
  #---------------------------------------------
  pattern_mismatch <- rep(
    FALSE,
    length(value)
  )

  if (!is.null(pattern)) {

    pattern_candidates <-
      !missing_value &
      !blank &
      !whitespace_only

    pattern_mismatch[pattern_candidates] <-
      !grepl(
        pattern,
        value[pattern_candidates],
        perl = TRUE
      )
  }

  #---------------------------------------------
  # Overall validity
  #---------------------------------------------
  valid_id <-
    !missing_value &
    !blank &
    !whitespace_only &
    !leading_whitespace &
    !trailing_whitespace &
    !duplicate &
    !pattern_mismatch

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = value,
    Missing = missing_value,
    Blank = blank,
    WhitespaceOnly = whitespace_only,
    LeadingWhitespace = leading_whitespace,
    TrailingWhitespace = trailing_whitespace,
    Duplicate = duplicate,
    PatternMismatch = pattern_mismatch,
    ValidID = valid_id,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "IDCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("id_check returns IDCheck object", {

  dat <- data.frame(
    ID = c("A001", "A002")
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_s3_class(result, "IDCheck")
  expect_s3_class(result, "data.frame")
})

test_that("id_check returns expected columns", {

  dat <- data.frame(
    ID = c("A001", "A002")
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "Blank",
      "WhitespaceOnly",
      "LeadingWhitespace",
      "TrailingWhitespace",
      "Duplicate",
      "PatternMismatch",
      "ValidID"
    )
  )
})

#==========================================================
# Valid identifiers
#==========================================================

test_that("valid identifiers are not flagged", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A003"
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_true(
    all(result$ValidID)
  )

  expect_false(
    any(result$Duplicate)
  )
})

#==========================================================
# Missing and blank
#==========================================================

test_that("id_check detects missing identifiers", {

  dat <- data.frame(
    ID = c(
      "A001",
      NA,
      "A003"
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Missing,
    c(FALSE, TRUE, FALSE)
  )

  expect_false(
    result$ValidID[2]
  )
})

test_that("id_check detects blank identifiers", {

  dat <- data.frame(
    ID = c(
      "A001",
      "",
      "A003"
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Blank,
    c(FALSE, TRUE, FALSE)
  )

  expect_false(
    result$ValidID[2]
  )
})

#==========================================================
# Whitespace
#==========================================================

test_that("id_check detects whitespace-only identifiers", {

  dat <- data.frame(
    ID = c(
      "A001",
      " ",
      "   "
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$WhitespaceOnly,
    c(FALSE, TRUE, TRUE)
  )

  expect_false(
    result$ValidID[2]
  )
})

test_that("id_check detects leading whitespace", {

  dat <- data.frame(
    ID = c(
      "A001",
      " A002"
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$LeadingWhitespace,
    c(FALSE, TRUE)
  )
})

test_that("id_check detects trailing whitespace", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002 "
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$TrailingWhitespace,
    c(FALSE, TRUE)
  )
})

#==========================================================
# Duplicates
#==========================================================

test_that("id_check flags all occurrences of duplicated IDs", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002",
      "A003"
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$ValidID,
    c(
      TRUE,
      FALSE,
      FALSE,
      TRUE
    )
  )
})

test_that("missing and blank IDs are not treated as duplicates", {

  dat <- data.frame(
    ID = c(
      NA,
      NA,
      "",
      ""
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_false(
    any(result$Duplicate)
  )
})

#==========================================================
# Pattern
#==========================================================

test_that("id_check detects pattern mismatches", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "B003",
      "A04"
    )
  )

  result <- id_check(
    dat,
    "ID",
    pattern = "^A[0-9]{3}$"
  )

  expect_equal(
    result$PatternMismatch,
    c(
      FALSE,
      FALSE,
      TRUE,
      TRUE
    )
  )
})

test_that("pattern is optional", {

  dat <- data.frame(
    ID = c(
      "ABC",
      "123"
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_false(
    any(result$PatternMismatch)
  )
})

test_that("missing blank and whitespace IDs are excluded from pattern check", {

  dat <- data.frame(
    ID = c(
      NA,
      "",
      " ",
      "A001"
    )
  )

  result <- id_check(
    dat,
    "ID",
    pattern = "^A[0-9]{3}$"
  )

  expect_false(
    any(result$PatternMismatch)
  )
})

#==========================================================
# Factor identifiers
#==========================================================

test_that("id_check supports factor identifiers", {

  dat <- data.frame(
    ID = factor(
      c(
        "A001",
        "A002",
        "A002"
      )
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE
    )
  )
})

#==========================================================
# Numeric identifiers
#==========================================================

test_that("id_check supports numeric identifiers", {

  dat <- data.frame(
    ID = c(
      101,
      102,
      102,
      103
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )
})

test_that("id_check supports integer identifiers", {

  dat <- data.frame(
    ID = c(
      1L,
      2L,
      3L
    )
  )

  result <- id_check(
    dat,
    "ID"
  )

  expect_true(
    all(result$ValidID)
  )
})

#==========================================================
# Invalid variable argument
#==========================================================

test_that("id_check rejects invalid variable argument", {

  expect_error(
    id_check(iris),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    id_check(iris, NULL),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    id_check(iris, ""),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    id_check(iris, 1),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    id_check(
      iris,
      c("Species", "Sepal.Length")
    ),
    "'variable' must be a single non-empty character string"
  )

  expect_error(
    id_check(
      iris,
      NA_character_
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("id_check rejects unknown variable", {

  expect_error(
    id_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

#==========================================================
# Invalid pattern
#==========================================================

test_that("id_check rejects invalid pattern argument", {

  dat <- data.frame(
    ID = "A001"
  )

  expect_error(
    id_check(
      dat,
      "ID",
      pattern = ""
    ),
    "'pattern' must be NULL or a single non-empty character string"
  )

  expect_error(
    id_check(
      dat,
      "ID",
      pattern = NA_character_
    ),
    "'pattern' must be NULL or a single non-empty character string"
  )

  expect_error(
    id_check(
      dat,
      "ID",
      pattern = 1
    ),
    "'pattern' must be NULL or a single non-empty character string"
  )

  expect_error(
    id_check(
      dat,
      "ID",
      pattern = c("^A", "^B")
    ),
    "'pattern' must be NULL or a single non-empty character string"
  )
})

test_that("id_check rejects malformed regular expression", {

  dat <- data.frame(
    ID = "A001"
  )

  expect_error(
    id_check(
      dat,
      "ID",
      pattern = "["
    ),
    "'pattern' is not a valid regular expression"
  )
})

#==========================================================
# Unsupported identifier types
#==========================================================

test_that("id_check rejects logical identifier", {

  dat <- data.frame(
    ID = c(
      TRUE,
      FALSE
    )
  )

  expect_error(
    id_check(
      dat,
      "ID"
    ),
    "'variable' must refer to a character, factor, numeric, or integer variable"
  )
})

test_that("id_check rejects Date identifier", {

  dat <- data.frame(
    ID = as.Date(
      c(
        "2026-01-01",
        "2026-01-02"
      )
    )
  )

  expect_error(
    id_check(
      dat,
      "ID"
    ),
    "'variable' must refer to a character, factor, numeric, or integer variable"
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("id_check rejects invalid data", {

  expect_error(
    id_check(
      NULL,
      "ID"
    )
  )

  expect_error(
    id_check(
      1:5,
      "ID"
    )
  )

  expect_error(
    id_check(
      list(ID = "A001"),
      "ID"
    )
  )

  expect_error(
    id_check(
      data.frame(),
      "ID"
    )
  )
})

#==========================================================
# Data Audit
#==========================================================

#' Perform a General Data Quality Audit
#'
#' Performs a general audit of a data frame and summarizes
#' common data-quality problems including missing values,
#' blank values, infinite values, constant variables,
#' near-zero variance variables, and duplicated rows.
#'
#' @param data A data.frame.
#' @param include_nzv Logical. If TRUE, near-zero variance
#'   variables are included in the audit. Default is TRUE.
#' @param rules Optional object of class \code{DataAuditRules}
#'   created by \code{\link{audit_rules}}. When supplied,
#'   the rules are evaluated and included in the audit
#'   report.
#'
#' @return An object of class \code{DataAuditReport}.
#'
#' @details
#' \code{audit_data()} is intended as a general first-pass
#' audit requiring minimal user configuration.
#'
#' More specialized checks such as ranges, categories,
#' patterns, identifiers, dates, dependencies, and
#' cross-variable consistency can be performed using the
#' corresponding DataAudit functions.
#'
#' Custom validation rules created with
#' \code{\link{audit_rules}} can be supplied through
#' \code{rules}. These rules are evaluated using
#' \code{\link{apply_audit_rules}}.
#'
#' The returned object contains:
#'
#' \itemize{
#'   \item \code{overview}: basic dataset information.
#'   \item \code{variables}: variable-level audit summary.
#'   \item \code{duplicates}: duplicated-row diagnostics.
#'   \item \code{summary}: overall audit statistics.
#'   \item \code{rules}: custom rule results, or NULL when
#'     no rules are supplied.
#' }
#'
#' @examples
#' dat <- data.frame(
#'   ID = c(1, 2, 2, 4),
#'   Species = c("Dog", "", "", "Cat"),
#'   Value = c(10, NA, Inf, 20),
#'   Constant = 1
#' )
#'
#' audit_data(dat)
#'
#' rules <- audit_rules(
#'   range = list(
#'     Value = c(0, 100)
#'   ),
#'   required = "Species"
#' )
#'
#' audit_data(
#'   dat,
#'   rules = rules
#' )
#'
#' @export

audit_data <- function(
    data,
    include_nzv = TRUE,
    rules = NULL
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate include_nzv
  #---------------------------------------------
  if (
    !is.logical(include_nzv) ||
    length(include_nzv) != 1L ||
    is.na(include_nzv)
  ) {

    stop(
      "'include_nzv' must be TRUE or FALSE.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate rules
  #---------------------------------------------
  if (
    !is.null(rules) &&
    !inherits(
      rules,
      "DataAuditRules"
    )
  ) {

    stop(
      "'rules' must be NULL or an object created by audit_rules().",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Basic dimensions
  #---------------------------------------------
  n_rows <- nrow(data)
  n_variables <- ncol(data)

  variable_names <- names(data)

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_n <- vapply(
    data,
    function(x) {
      sum(is.na(x))
    },
    numeric(1)
  )

  missing_percent <-
    100 * missing_n / n_rows

  #---------------------------------------------
  # Blank values
  #---------------------------------------------
  blank_n <- vapply(
    data,
    function(x) {

      if (
        is.character(x) ||
        is.factor(x)
      ) {

        y <- as.character(x)

        return(
          sum(
            !is.na(y) &
              grepl(
                "^[[:space:]]*$",
                y
              )
          )
        )
      }

      0
    },
    numeric(1)
  )

  #---------------------------------------------
  # Infinite values
  #---------------------------------------------
  infinite_n <- vapply(
    data,
    function(x) {

      if (is.numeric(x)) {

        return(
          sum(
            is.infinite(x),
            na.rm = TRUE
          )
        )
      }

      0
    },
    numeric(1)
  )

  #---------------------------------------------
  # Unique non-missing values
  #---------------------------------------------
  unique_n <- vapply(
    data,
    function(x) {

      length(
        unique(
          x[!is.na(x)]
        )
      )
    },
    integer(1)
  )

  #---------------------------------------------
  # Constant variables
  #---------------------------------------------
  constant <-
    unique_n <= 1L

  #---------------------------------------------
  # Near-zero variance
  #---------------------------------------------
  near_zero_variance <- rep(
    FALSE,
    n_variables
  )

  if (include_nzv) {

    near_zero_variance <- vapply(
      data,
      function(x) {

        y <- x[!is.na(x)]

        if (length(y) == 0L) {
          return(TRUE)
        }

        tab <- table(
          y,
          useNA = "no"
        )

        if (length(tab) <= 1L) {
          return(TRUE)
        }

        sorted_counts <- sort(
          as.numeric(tab),
          decreasing = TRUE
        )

        frequency_ratio <-
          sorted_counts[1L] /
          sorted_counts[2L]

        percent_unique <-
          100 *
          length(tab) /
          length(y)

        frequency_ratio > 19 &&
          percent_unique <= 10
      },
      logical(1)
    )
  }

  #---------------------------------------------
  # Variable classes
  #---------------------------------------------
  variable_class <- vapply(
    data,
    function(x) {
      class(x)[1L]
    },
    character(1)
  )

  #---------------------------------------------
  # Variable-level issue
  #---------------------------------------------
  variable_issue <-
    missing_n > 0 |
    blank_n > 0 |
    infinite_n > 0 |
    constant |
    near_zero_variance

  #---------------------------------------------
  # Variable summary
  #---------------------------------------------
  variables <- data.frame(
    Variable = variable_names,
    Class = variable_class,
    Missing = as.integer(missing_n),
    MissingPercent =
      as.numeric(missing_percent),
    Blank = as.integer(blank_n),
    Infinite = as.integer(infinite_n),
    Unique = as.integer(unique_n),
    Constant = constant,
    NearZeroVariance =
      near_zero_variance,
    Issue = variable_issue,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Duplicate rows
  #---------------------------------------------
  duplicate_flag <-
    duplicated(data) |
    duplicated(
      data,
      fromLast = TRUE
    )

  duplicates <- data.frame(
    Row = seq_len(n_rows),
    Duplicate = duplicate_flag,
    stringsAsFactors = FALSE
  )

  #---------------------------------------------
  # Overview
  #---------------------------------------------
  overview <- data.frame(
    Metric = c(
      "Rows",
      "Variables",
      "Missing values",
      "Blank values",
      "Infinite values",
      "Duplicated rows",
      "Constant variables",
      "Near-zero variance variables",
      "Variables with issues"
    ),
    Value = c(
      n_rows,
      n_variables,
      sum(missing_n),
      sum(blank_n),
      sum(infinite_n),
      sum(duplicate_flag),
      sum(constant),
      sum(near_zero_variance),
      sum(variable_issue)
    ),
    stringsAsFactors = FALSE
  )

  #---------------------------------------------
  # Automatic audit issue count
  #---------------------------------------------
  total_issues <-
    sum(missing_n) +
    sum(blank_n) +
    sum(infinite_n) +
    sum(duplicate_flag) +
    sum(constant) +
    sum(near_zero_variance)

  #---------------------------------------------
  # Apply custom audit rules
  #---------------------------------------------
  rule_results <- NULL

  if (!is.null(rules)) {

    rule_results <- apply_audit_rules(
      data,
      rules
    )
  }

  #---------------------------------------------
  # Rule violation count
  #---------------------------------------------
  rule_violations <- if (
    is.null(rule_results)
  ) {

    0L

  } else {

    as.integer(
      rule_results$total_violations
    )
  }

  #---------------------------------------------
  # Overall summary
  #---------------------------------------------
  summary <- data.frame(
    Rows = n_rows,
    Variables = n_variables,
    VariablesWithIssues =
      sum(variable_issue),
    DuplicateRows =
      sum(duplicate_flag),
    TotalIssues =
      total_issues,
    RuleViolations =
      rule_violations,
    OverallIssues =
      total_issues +
      rule_violations,
    stringsAsFactors = FALSE
  )

  #---------------------------------------------
  # Construct report
  #---------------------------------------------
  result <- list(
    overview = overview,
    variables = variables,
    duplicates = duplicates,
    summary = summary,
    rules = rule_results
  )

  class(result) <- c(
    "DataAuditReport",
    "list"
  )

  result
}

library(testthat)
library(DataAudit)


test_that("summary returns correct class", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_s3_class(
    result,
    "summary.DataAuditReport"
  )

  expect_type(
    result,
    "list"
  )
})


test_that("summary contains expected components", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_equal(
    names(result),
    c(
      "dataset",
      "variables_with_issues",
      "duplicate_rows",
      "rules"
    )
  )
})


test_that("dataset summary contains expected fields", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_equal(
    names(result$dataset),
    c(
      "Rows",
      "Variables",
      "VariablesWithIssues",
      "DuplicateRows",
      "TotalIssues",
      "RuleViolations",
      "OverallIssues"
    )
  )
})


test_that("summary identifies variables with issues", {

  report <- audit_data(
    data.frame(
      x = c(1, NA, 3),
      y = 1:3
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_equal(
    result$variables_with_issues$Variable,
    "x"
  )
})


test_that("summary returns empty variable table when clean", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_equal(
    nrow(result$variables_with_issues),
    0L
  )
})


test_that("summary identifies duplicate rows", {

  report <- audit_data(
    data.frame(
      x = c(
        1,
        2,
        2,
        3
      )
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_equal(
    result$duplicate_rows$Row,
    c(2L, 3L)
  )
})


test_that("summary returns empty duplicate table when clean", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_equal(
    nrow(result$duplicate_rows),
    0L
  )
})


test_that("summary returns NULL rules without custom rules", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  result <- summary(report)

  expect_null(
    result$rules
  )

  expect_equal(
    result$dataset$RuleViolations,
    0L
  )
})


test_that("summary includes custom rule results", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002"
    ),
    Age = c(
      5,
      40,
      -2
    )
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    ),
    unique = "ID"
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = rules
  )

  result <- summary(report)

  expect_true(
    is.data.frame(result$rules)
  )

  expect_equal(
    names(result$rules),
    c(
      "Rule",
      "Violations"
    )
  )

  expect_equal(
    result$dataset$RuleViolations,
    report$rules$total_violations
  )
})


test_that("summary reports combined overall issues", {

  dat <- data.frame(
    Age = c(
      5,
      NA,
      40
    )
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    )
  )

  report <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = rules
  )

  result <- summary(report)

  expect_equal(
    result$dataset$OverallIssues,
    result$dataset$TotalIssues +
      result$dataset$RuleViolations
  )
})


#==========================================================
# Defensive validation
#==========================================================

test_that("summary rejects object without correct class", {

  x <- list(
    overview = data.frame(),
    variables = data.frame(),
    duplicates = data.frame(),
    summary = data.frame(),
    rules = NULL
  )

  expect_error(
    summary.DataAuditReport(x),
    "Invalid 'DataAuditReport' object"
  )
})


test_that("summary rejects malformed report", {

  x <- list(
    overview = data.frame()
  )

  class(x) <- c(
    "DataAuditReport",
    "list"
  )

  expect_error(
    summary(x),
    "Invalid 'DataAuditReport' object"
  )
})


test_that("summary rejects invalid dataset summary", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$summary <- "invalid"

  expect_error(
    summary(report),
    "Invalid summary in 'DataAuditReport'"
  )
})


test_that("summary rejects missing dataset summary columns", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$summary <- data.frame(
    Wrong = 1
  )

  expect_error(
    summary(report),
    "Invalid summary in 'DataAuditReport'"
  )
})


test_that("summary rejects invalid variable summary", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$variables <- "invalid"

  expect_error(
    summary(report),
    "Invalid variable summary in 'DataAuditReport'"
  )
})


test_that("summary rejects variable summary without Issue", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$variables <- data.frame(
    Variable = "x"
  )

  expect_error(
    summary(report),
    "Invalid variable summary in 'DataAuditReport'"
  )
})


test_that("summary rejects invalid duplicate summary", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$duplicates <- "invalid"

  expect_error(
    summary(report),
    "Invalid duplicate summary in 'DataAuditReport'"
  )
})


test_that("summary rejects duplicate summary without Duplicate", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$duplicates <- data.frame(
    Row = 1:5
  )

  expect_error(
    summary(report),
    "Invalid duplicate summary in 'DataAuditReport'"
  )
})


test_that("summary rejects invalid rule result", {

  report <- audit_data(
    data.frame(
      x = 1:5
    ),
    include_nzv = FALSE
  )

  report$rules <- list(
    invalid = TRUE
  )

  expect_error(
    summary(report),
    "Invalid rule results in 'DataAuditReport'"
  )
})


test_that("summary rejects non-data-frame rule summary", {

  report <- audit_data(
    data.frame(
      Age = c(5, 40)
    ),
    include_nzv = FALSE,
    rules = audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  report$rules$summary <- "invalid"

  expect_error(
    summary(report),
    "Invalid rule summary in 'DataAuditReport'"
  )
})


test_that("summary rejects malformed rule summary", {

  report <- audit_data(
    data.frame(
      Age = c(5, 40)
    ),
    include_nzv = FALSE,
    rules = audit_rules(
      range = list(
        Age = c(0, 30)
      )
    )
  )

  report$rules$summary <- data.frame(
    Wrong = 1
  )

  expect_error(
    summary(report),
    "Invalid rule summary in 'DataAuditReport'"
  )
})

library(testthat)
library(DataAudit)

test_that("audit_data returns correct class", {

  dat <- data.frame(
    x = 1:5,
    y = letters[1:5]
  )

  result <- audit_data(dat)

  expect_s3_class(
    result,
    "DataAuditReport"
  )

  expect_type(
    result,
    "list"
  )
})


test_that("audit_data returns expected components", {

  dat <- data.frame(
    x = 1:3
  )

  result <- audit_data(dat)

  expect_equal(
    names(result),
    c(
      "overview",
      "variables",
      "duplicates",
      "summary",
      "rules"
    )
  )

  expect_null(
    result$rules
  )
})

test_that("audit_data calculates dimensions", {

  dat <- data.frame(
    x = 1:4,
    y = letters[1:4]
  )

  result <- audit_data(dat)

  expect_equal(
    result$summary$Rows,
    4
  )

  expect_equal(
    result$summary$Variables,
    2
  )
})


test_that("audit_data detects missing values", {

  dat <- data.frame(
    x = c(1, NA, 3),
    y = c("A", "B", NA)
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Missing,
    c(1L, 1L)
  )

  expect_equal(
    sum(result$variables$Missing),
    2
  )
})


test_that("audit_data calculates missing percentage", {

  dat <- data.frame(
    x = c(1, NA, NA, 4)
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$MissingPercent,
    50
  )
})


test_that("audit_data detects blank character values", {

  dat <- data.frame(
    x = c(
      "Dog",
      "",
      "   ",
      "Cat"
    )
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Blank,
    2L
  )
})


test_that("audit_data detects blanks in factors", {

  dat <- data.frame(
    x = factor(
      c(
        "Dog",
        "",
        "Cat"
      )
    )
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Blank,
    1L
  )
})


test_that("numeric variables have no blanks", {

  dat <- data.frame(
    x = c(1, 2, 3)
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Blank,
    0L
  )
})


test_that("audit_data detects infinite values", {

  dat <- data.frame(
    x = c(
      1,
      Inf,
      -Inf,
      NA
    )
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Infinite,
    2L
  )
})


test_that("non-numeric variables have no infinite values", {

  dat <- data.frame(
    x = c(
      "A",
      "B",
      "C"
    )
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Infinite,
    0L
  )
})


test_that("audit_data counts unique values", {

  dat <- data.frame(
    x = c(
      1,
      1,
      2,
      NA
    )
  )

  result <- audit_data(dat)

  expect_equal(
    result$variables$Unique,
    2L
  )
})


test_that("audit_data detects constant variables", {

  dat <- data.frame(
    x = rep(1, 10),
    y = 1:10
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_equal(
    result$variables$Constant,
    c(TRUE, FALSE)
  )
})


test_that("all missing variable is constant", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      5
    )
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_true(
    result$variables$Constant
  )
})


test_that("audit_data detects duplicate rows", {

  dat <- data.frame(
    x = c(1, 2, 2, 3),
    y = c("A", "B", "B", "C")
  )

  result <- audit_data(dat)

  expect_equal(
    result$duplicates$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$summary$DuplicateRows,
    2
  )
})


test_that("audit_data detects near-zero variance", {

  dat <- data.frame(
    x = c(
      rep("A", 99),
      "B"
    )
  )

  result <- audit_data(
    dat,
    include_nzv = TRUE
  )

  expect_true(
    result$variables$NearZeroVariance
  )
})


test_that("balanced variable is not near-zero variance", {

  dat <- data.frame(
    x = rep(
      c("A", "B"),
      50
    )
  )

  result <- audit_data(
    dat,
    include_nzv = TRUE
  )

  expect_false(
    result$variables$NearZeroVariance
  )
})


test_that("constant variable is near-zero variance", {

  dat <- data.frame(
    x = rep("A", 20)
  )

  result <- audit_data(
    dat,
    include_nzv = TRUE
  )

  expect_true(
    result$variables$NearZeroVariance
  )
})


test_that("all missing variable is near-zero variance", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      10
    )
  )

  result <- audit_data(
    dat,
    include_nzv = TRUE
  )

  expect_true(
    result$variables$NearZeroVariance
  )
})


test_that("near-zero variance can be disabled", {

  dat <- data.frame(
    x = rep("A", 20)
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_false(
    result$variables$NearZeroVariance
  )
})


test_that("variable issue is identified", {

  dat <- data.frame(
    clean = 1:5,
    missing = c(
      1,
      NA,
      3,
      4,
      5
    ),
    constant = rep(
      1,
      5
    )
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_equal(
    result$variables$Issue,
    c(
      FALSE,
      TRUE,
      TRUE
    )
  )
})


test_that("overview contains expected metrics", {

  dat <- data.frame(
    x = c(1, NA, 3)
  )

  result <- audit_data(dat)

  expect_equal(
    result$overview$Metric,
    c(
      "Rows",
      "Variables",
      "Missing values",
      "Blank values",
      "Infinite values",
      "Duplicated rows",
      "Constant variables",
      "Near-zero variance variables",
      "Variables with issues"
    )
  )
})


test_that("summary reports total issues", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      Inf
    )
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_true(
    result$summary$TotalIssues >= 2
  )
})


test_that("include_nzv must be logical scalar", {

  dat <- data.frame(
    x = 1:3
  )

  expect_error(
    audit_data(
      dat,
      include_nzv = 1
    ),
    "'include_nzv' must be TRUE or FALSE"
  )

  expect_error(
    audit_data(
      dat,
      include_nzv = NA
    ),
    "'include_nzv' must be TRUE or FALSE"
  )

  expect_error(
    audit_data(
      dat,
      include_nzv = c(
        TRUE,
        FALSE
      )
    ),
    "'include_nzv' must be TRUE or FALSE"
  )
})


test_that("audit_data rejects invalid data", {

  expect_error(
    audit_data(NULL)
  )

  expect_error(
    audit_data(1:5)
  )

  expect_error(
    audit_data(
      list(x = 1:5)
    )
  )

  expect_error(
    audit_data(
      data.frame()
    )
  )
})
test_that("audit_data accepts custom audit rules", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002"
    ),
    Age = c(
      5,
      40,
      -2
    ),
    Species = c(
      "Dog",
      "Cat",
      ""
    ),
    stringsAsFactors = FALSE
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    ),
    unique = "ID",
    required = "Species"
  )

  result <- audit_data(
    dat,
    rules = rules
  )

  expect_s3_class(
    result,
    "DataAuditReport"
  )

  expect_s3_class(
    result$rules,
    "DataAuditRuleResult"
  )

  expect_equal(
    result$rules$total_violations,
    5L
  )

  expect_equal(
    result$summary$RuleViolations,
    5L
  )
})


test_that("audit_data reports zero rule violations without rules", {

  dat <- data.frame(
    x = 1:5
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE
  )

  expect_null(
    result$rules
  )

  expect_equal(
    result$summary$RuleViolations,
    0L
  )

  expect_equal(
    result$summary$OverallIssues,
    result$summary$TotalIssues
  )
})


test_that("OverallIssues combines automatic and rule issues", {

  dat <- data.frame(
    Age = c(
      5,
      NA,
      40
    )
  )

  rules <- audit_rules(
    range = list(
      Age = c(0, 30)
    )
  )

  result <- audit_data(
    dat,
    include_nzv = FALSE,
    rules = rules
  )

  expect_equal(
    result$summary$OverallIssues,
    result$summary$TotalIssues +
      result$summary$RuleViolations
  )
})


test_that("audit_data rejects invalid rules", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    audit_data(
      dat,
      rules = list()
    ),
    "'rules' must be NULL or an object created by audit_rules"
  )
})


test_that("audit_data propagates rule variable errors", {

  dat <- data.frame(
    Age = 1:5
  )

  rules <- audit_rules(
    required = "MissingVariable"
  )

  expect_error(
    audit_data(
      dat,
      rules = rules
    ),
    "referenced by rules not found"
  )
})

#==========================================================
# Data Overview
#==========================================================

#' Data Overview
#'
#' Provides a concise overview of the size, structure,
#' completeness, and variable types of a dataset.
#'
#' @param data A data.frame.
#'
#' @return An object of class \code{DataOverview} containing
#'   dataset-level summary information.
#'
#' @details
#' The function reports the number of rows and variables,
#' total number of cells, missing and complete cells,
#' completeness percentage, complete cases, duplicate rows,
#' and counts of common variable types.
#'
#' @examples
#' data_overview(iris)
#'
#' dat <- iris
#' dat$Sepal.Length[1:5] <- NA
#' data_overview(dat)
#'
#' @export

data_overview <- function(data) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Dimensions
  #---------------------------------------------
  n_rows <- nrow(data)
  n_variables <- ncol(data)

  total_cells <- n_rows * n_variables

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_cells <- sum(is.na(data))

  complete_cells <- total_cells - missing_cells

  completeness <- if (total_cells == 0L) {
    NA_real_
  } else {
    100 * complete_cells / total_cells
  }

  #---------------------------------------------
  # Complete cases
  #---------------------------------------------
  complete_cases <- sum(
    stats::complete.cases(data)
  )

  #---------------------------------------------
  # Duplicate rows
  #---------------------------------------------
  duplicate_rows <- sum(
    duplicated(data)
  )

  #---------------------------------------------
  # Variable types
  #---------------------------------------------
  numeric_variables <- sum(
    vapply(
      data,
      is.numeric,
      logical(1)
    )
  )

  factor_variables <- sum(
    vapply(
      data,
      is.factor,
      logical(1)
    )
  )

  character_variables <- sum(
    vapply(
      data,
      is.character,
      logical(1)
    )
  )

  logical_variables <- sum(
    vapply(
      data,
      is.logical,
      logical(1)
    )
  )

  date_variables <- sum(
    vapply(
      data,
      inherits,
      logical(1),
      what = "Date"
    )
  )

  datetime_variables <- sum(
    vapply(
      data,
      inherits,
      logical(1),
      what = "POSIXct"
    )
  )

  #---------------------------------------------
  # Result
  #---------------------------------------------
  result <- list(

    Rows = n_rows,

    Variables = n_variables,

    TotalCells = total_cells,

    MissingCells = missing_cells,

    CompleteCells = complete_cells,

    Completeness = round(
      completeness,
      2
    ),

    CompleteCases = complete_cases,

    DuplicateRows = duplicate_rows,

    NumericVariables = numeric_variables,

    FactorVariables = factor_variables,

    CharacterVariables = character_variables,

    LogicalVariables = logical_variables,

    DateVariables = date_variables,

    DateTimeVariables = datetime_variables

  )

  class(result) <- "DataOverview"

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("data_overview returns DataOverview object", {

  result <- data_overview(iris)

  expect_s3_class(
    result,
    "DataOverview"
  )

})

test_that("data_overview returns list", {

  result <- data_overview(iris)

  expect_true(
    is.list(result)
  )

})

test_that("data_overview contains expected components", {

  result <- data_overview(iris)

  expected <- c(
    "Rows",
    "Variables",
    "TotalCells",
    "MissingCells",
    "CompleteCells",
    "Completeness",
    "CompleteCases",
    "DuplicateRows",
    "NumericVariables",
    "FactorVariables",
    "CharacterVariables",
    "LogicalVariables",
    "DateVariables",
    "DateTimeVariables"
  )

  expect_true(
    all(expected %in% names(result))
  )

})


#==========================================================
# iris dataset
#==========================================================

test_that("data_overview reports correct iris dimensions", {

  result <- data_overview(iris)

  expect_equal(
    result$Rows,
    150L
  )

  expect_equal(
    result$Variables,
    5L
  )

})

test_that("data_overview reports correct total cells", {

  result <- data_overview(iris)

  expect_equal(
    result$TotalCells,
    750L
  )

})

test_that("data_overview reports no missing cells for iris", {

  result <- data_overview(iris)

  expect_equal(
    result$MissingCells,
    0L
  )

  expect_equal(
    result$CompleteCells,
    750L
  )

  expect_equal(
    result$Completeness,
    100
  )

})

test_that("data_overview reports complete cases correctly", {

  result <- data_overview(iris)

  expect_equal(
    result$CompleteCases,
    150L
  )

})

test_that("data_overview reports iris variable types", {

  result <- data_overview(iris)

  expect_equal(
    result$NumericVariables,
    4L
  )

  expect_equal(
    result$FactorVariables,
    1L
  )

  expect_equal(
    result$CharacterVariables,
    0L
  )

  expect_equal(
    result$LogicalVariables,
    0L
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("data_overview calculates missing values correctly", {

  dat <- iris

  dat$Sepal.Length[1:10] <- NA
  dat$Species[1:5] <- NA

  result <- data_overview(dat)

  expect_equal(
    result$MissingCells,
    15L
  )

  expect_equal(
    result$CompleteCells,
    735L
  )

})

test_that("data_overview calculates completeness correctly", {

  dat <- data.frame(
    x = c(1, NA),
    y = c(1, 2)
  )

  result <- data_overview(dat)

  expect_equal(
    result$TotalCells,
    4L
  )

  expect_equal(
    result$MissingCells,
    1L
  )

  expect_equal(
    result$Completeness,
    75
  )

})

test_that("data_overview calculates complete cases with missing data", {

  dat <- data.frame(
    x = c(1, NA, 3),
    y = c(1, 2, NA)
  )

  result <- data_overview(dat)

  expect_equal(
    result$CompleteCases,
    1L
  )

})


#==========================================================
# Duplicate rows
#==========================================================

test_that("data_overview detects duplicate rows", {

  dat <- data.frame(
    x = c(1, 2, 2, 3),
    y = c("A", "B", "B", "C")
  )

  result <- data_overview(dat)

  expect_equal(
    result$DuplicateRows,
    1L
  )

})

test_that("data_overview reports no duplicates when absent", {

  dat <- data.frame(
    x = 1:5,
    y = letters[1:5]
  )

  result <- data_overview(dat)

  expect_equal(
    result$DuplicateRows,
    0L
  )

})


#==========================================================
# Variable types
#==========================================================

test_that("data_overview identifies character variables", {

  dat <- data.frame(
    x = c("A", "B", "C"),
    stringsAsFactors = FALSE
  )

  result <- data_overview(dat)

  expect_equal(
    result$CharacterVariables,
    1L
  )

})

test_that("data_overview identifies factor variables", {

  dat <- data.frame(
    x = factor(c("A", "B", "C"))
  )

  result <- data_overview(dat)

  expect_equal(
    result$FactorVariables,
    1L
  )

})

test_that("data_overview identifies logical variables", {

  dat <- data.frame(
    x = c(TRUE, FALSE, TRUE)
  )

  result <- data_overview(dat)

  expect_equal(
    result$LogicalVariables,
    1L
  )

})

test_that("data_overview identifies Date variables", {

  dat <- data.frame(
    date = as.Date("2026-01-01") + 0:4
  )

  result <- data_overview(dat)

  expect_equal(
    result$DateVariables,
    1L
  )

})

test_that("data_overview identifies POSIXct variables", {

  dat <- data.frame(
    time = as.POSIXct(
      "2026-01-01 00:00:00",
      tz = "UTC"
    ) + 0:4
  )

  result <- data_overview(dat)

  expect_equal(
    result$DateTimeVariables,
    1L
  )

})

test_that("data_overview handles mixed variable types", {

  dat <- data.frame(
    numeric = 1:5,
    character = letters[1:5],
    factor = factor(LETTERS[1:5]),
    logical = c(TRUE, FALSE, TRUE, FALSE, TRUE),
    date = as.Date("2026-01-01") + 0:4,
    stringsAsFactors = FALSE
  )

  result <- data_overview(dat)

  expect_equal(result$NumericVariables, 1L)
  expect_equal(result$CharacterVariables, 1L)
  expect_equal(result$FactorVariables, 1L)
  expect_equal(result$LogicalVariables, 1L)
  expect_equal(result$DateVariables, 1L)

})


#==========================================================
# One-column data
#==========================================================

test_that("data_overview works with one-column data", {

  dat <- data.frame(
    x = 1:10
  )

  result <- data_overview(dat)

  expect_equal(
    result$Rows,
    10L
  )

  expect_equal(
    result$Variables,
    1L
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("data_overview works with zero-row data.frame", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- data_overview(dat)

  expect_equal(
    result$Rows,
    0L
  )

  expect_equal(
    result$Variables,
    2L
  )

  expect_equal(
    result$TotalCells,
    0L
  )

  expect_true(
    is.na(result$Completeness)
  )

  expect_equal(
    result$CompleteCases,
    0L
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("data_overview works with tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::as_tibble(iris)

  result <- data_overview(dat)

  expect_s3_class(
    result,
    "DataOverview"
  )

  expect_equal(
    result$Rows,
    150L
  )

})


#==========================================================
# Invalid input
#==========================================================

test_that("data_overview rejects NULL", {

  expect_error(
    data_overview(NULL),
    "'data' cannot be NULL"
  )

})

test_that("data_overview rejects numeric vector", {

  expect_error(
    data_overview(1:10),
    "'data' must be a data.frame"
  )

})

test_that("data_overview rejects character vector", {

  expect_error(
    data_overview(letters),
    "'data' must be a data.frame"
  )

})

test_that("data_overview rejects list", {

  expect_error(
    data_overview(
      list(
        x = 1:10,
        y = 11:20
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("data_overview rejects empty data.frame", {

  expect_error(
    data_overview(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Variable Type Check
#==========================================================

#' Check Variable Types
#'
#' Checks whether variables in a dataset match user-defined
#' expected data types.
#'
#' @param data A data.frame.
#' @param expected A named character vector specifying the
#'   expected type for each variable.
#'
#' @return A data.frame of class \code{TypeCheck}
#'   containing variable-level type diagnostics.
#'
#' @details
#' Supported expected types are:
#'
#' \itemize{
#'   \item \code{"numeric"}
#'   \item \code{"integer"}
#'   \item \code{"character"}
#'   \item \code{"factor"}
#'   \item \code{"logical"}
#'   \item \code{"Date"}
#'   \item \code{"POSIXct"}
#' }
#'
#' The \code{expected} argument must be a named character
#' vector. Names correspond to variables in \code{data}.
#'
#' For example:
#'
#' \code{c(Age = "numeric", Species = "factor")}
#'
#' @examples
#' dat <- data.frame(
#'   Age = c(2, 4, 6),
#'   Species = factor(c("Dog", "Cat", "Dog")),
#'   Positive = c(TRUE, FALSE, TRUE)
#' )
#'
#' type_check(
#'   dat,
#'   c(
#'     Age = "numeric",
#'     Species = "factor",
#'     Positive = "logical"
#'   )
#' )
#'
#' @export

type_check <- function(data, expected) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate expected
  #---------------------------------------------
  if (missing(expected) ||
      !is.character(expected) ||
      length(expected) == 0L) {

    stop(
      "'expected' must be a non-empty named character vector.",
      call. = FALSE
    )
  }

  expected_names <- names(expected)

  if (is.null(expected_names) ||
      any(is.na(expected_names)) ||
      any(!nzchar(expected_names))) {

    stop(
      "'expected' must have variable names.",
      call. = FALSE
    )
  }

  if (anyDuplicated(expected_names)) {

    stop(
      "Variable names in 'expected' must be unique.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Supported types
  #---------------------------------------------
  supported_types <- c(
    "numeric",
    "integer",
    "character",
    "factor",
    "logical",
    "Date",
    "POSIXct"
  )

  if (any(!expected %in% supported_types)) {

    invalid <- unique(
      expected[
        !expected %in% supported_types
      ]
    )

    stop(
      sprintf(
        "Unsupported expected type(s): %s.",
        paste(
          invalid,
          collapse = ", "
        )
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check requested variables exist
  #---------------------------------------------
  missing_variables <- setdiff(
    expected_names,
    names(data)
  )

  if (length(missing_variables) > 0L) {

    stop(
      sprintf(
        "Variable(s) not found in 'data': %s.",
        paste(
          missing_variables,
          collapse = ", "
        )
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Determine actual type
  #---------------------------------------------
  actual_type <- function(x) {

    if (inherits(x, "Date")) {
      return("Date")
    }

    if (inherits(x, "POSIXct")) {
      return("POSIXct")
    }

    if (is.factor(x)) {
      return("factor")
    }

    if (is.logical(x)) {
      return("logical")
    }

    if (is.integer(x)) {
      return("integer")
    }

    if (is.numeric(x)) {
      return("numeric")
    }

    if (is.character(x)) {
      return("character")
    }

    class(x)[1L]
  }

  #---------------------------------------------
  # Type matching
  #---------------------------------------------
  matches_type <- function(x, expected_type) {

    switch(
      expected_type,

      numeric = is.numeric(x),

      integer = is.integer(x),

      character = is.character(x),

      factor = is.factor(x),

      logical = is.logical(x),

      Date = inherits(
        x,
        "Date"
      ),

      POSIXct = inherits(
        x,
        "POSIXct"
      )
    )
  }

  #---------------------------------------------
  # Extract requested variables
  #---------------------------------------------
  actual <- vapply(
    expected_names,
    function(variable) {

      actual_type(
        data[[variable]]
      )

    },
    character(1)
  )

  match <- vapply(
    seq_along(expected),
    function(i) {

      matches_type(
        data[[expected_names[i]]],
        expected[i]
      )

    },
    logical(1)
  )

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Variable = unname(expected_names),
    Expected = unname(expected),
    Actual = unname(actual),
    Match = unname(match),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "TypeCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("type_check returns TypeCheck", {

  dat <- data.frame(
    x = 1:5
  )

  result <- type_check(
    dat,
    c(x = "integer")
  )

  expect_s3_class(
    result,
    "TypeCheck"
  )

})

test_that("type_check returns data.frame", {

  dat <- data.frame(
    x = 1:5
  )

  result <- type_check(
    dat,
    c(x = "integer")
  )

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("type_check returns expected columns", {

  dat <- data.frame(
    x = 1:5
  )

  result <- type_check(
    dat,
    c(x = "integer")
  )

  expect_equal(
    names(result),
    c(
      "Variable",
      "Expected",
      "Actual",
      "Match"
    )
  )

})


#==========================================================
# Integer
#==========================================================

test_that("type_check identifies integer", {

  dat <- data.frame(
    x = 1:5
  )

  result <- type_check(
    dat,
    c(x = "integer")
  )

  expect_equal(
    result$Actual,
    "integer"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Numeric
#==========================================================

test_that("type_check identifies numeric", {

  dat <- data.frame(
    x = c(
      1.2,
      2.4,
      3.6
    )
  )

  result <- type_check(
    dat,
    c(x = "numeric")
  )

  expect_equal(
    result$Actual,
    "numeric"
  )

  expect_true(
    result$Match
  )

})

test_that("integer satisfies numeric expectation", {

  dat <- data.frame(
    x = 1:5
  )

  result <- type_check(
    dat,
    c(x = "numeric")
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Character
#==========================================================

test_that("type_check identifies character", {

  dat <- data.frame(
    x = c(
      "Dog",
      "Cat"
    ),
    stringsAsFactors = FALSE
  )

  result <- type_check(
    dat,
    c(x = "character")
  )

  expect_equal(
    result$Actual,
    "character"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Factor
#==========================================================

test_that("type_check identifies factor", {

  dat <- data.frame(
    x = factor(
      c(
        "Dog",
        "Cat"
      )
    )
  )

  result <- type_check(
    dat,
    c(x = "factor")
  )

  expect_equal(
    result$Actual,
    "factor"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Logical
#==========================================================

test_that("type_check identifies logical", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE,
      TRUE
    )
  )

  result <- type_check(
    dat,
    c(x = "logical")
  )

  expect_equal(
    result$Actual,
    "logical"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Date
#==========================================================

test_that("type_check identifies Date", {

  dat <- data.frame(
    x = as.Date(
      "2026-01-01"
    ) + 0:4
  )

  result <- type_check(
    dat,
    c(x = "Date")
  )

  expect_equal(
    result$Actual,
    "Date"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# POSIXct
#==========================================================

test_that("type_check identifies POSIXct", {

  dat <- data.frame(
    x = as.POSIXct(
      "2026-01-01",
      tz = "UTC"
    ) + 0:4
  )

  result <- type_check(
    dat,
    c(x = "POSIXct")
  )

  expect_equal(
    result$Actual,
    "POSIXct"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Type mismatch
#==========================================================

test_that("type_check identifies mismatch", {

  dat <- data.frame(
    x = c(
      "1",
      "2",
      "3"
    ),
    stringsAsFactors = FALSE
  )

  result <- type_check(
    dat,
    c(x = "numeric")
  )

  expect_equal(
    result$Actual,
    "character"
  )

  expect_false(
    result$Match
  )

})


#==========================================================
# Multiple variables
#==========================================================

test_that("type_check checks multiple variables", {

  dat <- data.frame(
    ID = 1:3,
    Weight = c(
      20.5,
      25.2,
      30.1
    ),
    Species = factor(
      c(
        "Dog",
        "Cat",
        "Dog"
      )
    ),
    Positive = c(
      TRUE,
      FALSE,
      TRUE
    )
  )

  result <- type_check(
    dat,
    c(
      ID = "integer",
      Weight = "numeric",
      Species = "factor",
      Positive = "logical"
    )
  )

  expect_equal(
    result$Variable,
    c(
      "ID",
      "Weight",
      "Species",
      "Positive"
    )
  )

  expect_true(
    all(result$Match)
  )

})


#==========================================================
# Mixed matches
#==========================================================

test_that("type_check returns mixed matches correctly", {

  dat <- data.frame(
    Age = c(
      1,
      2,
      3
    ),
    Species = c(
      "Dog",
      "Cat",
      "Dog"
    ),
    stringsAsFactors = FALSE
  )

  result <- type_check(
    dat,
    c(
      Age = "character",
      Species = "character"
    )
  )

  expect_equal(
    result$Match,
    c(
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Missing values do not alter type
#==========================================================

test_that("type_check handles missing values", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      3
    )
  )

  result <- type_check(
    dat,
    c(x = "numeric")
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# All missing typed vectors
#==========================================================

test_that("type_check handles all NA numeric", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      5
    )
  )

  result <- type_check(
    dat,
    c(x = "numeric")
  )

  expect_equal(
    result$Actual,
    "numeric"
  )

  expect_true(
    result$Match
  )

})

test_that("type_check handles all NA character", {

  dat <- data.frame(
    x = rep(
      NA_character_,
      5
    )
  )

  result <- type_check(
    dat,
    c(x = "character")
  )

  expect_equal(
    result$Actual,
    "character"
  )

  expect_true(
    result$Match
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("type_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- type_check(
    dat,
    c(
      x = "numeric",
      y = "character"
    )
  )

  expect_equal(
    nrow(result),
    2L
  )

  expect_true(
    all(result$Match)
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("type_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = 1:5,
    y = c(
      "A",
      "B",
      "C",
      "D",
      "E"
    )
  )

  result <- type_check(
    dat,
    c(
      x = "integer",
      y = "character"
    )
  )

  expect_s3_class(
    result,
    "TypeCheck"
  )

  expect_true(
    all(result$Match)
  )

})


#==========================================================
# expected validation
#==========================================================

test_that("type_check rejects missing expected", {

  expect_error(
    type_check(iris),
    "'expected' must be a non-empty named character vector"
  )

})

test_that("type_check rejects NULL expected", {

  expect_error(
    type_check(
      iris,
      NULL
    ),
    "'expected' must be a non-empty named character vector"
  )

})

test_that("type_check rejects numeric expected", {

  expect_error(
    type_check(
      iris,
      c(Species = 1)
    ),
    "'expected' must be a non-empty named character vector"
  )

})

test_that("type_check rejects empty expected", {

  expect_error(
    type_check(
      iris,
      character(0)
    ),
    "'expected' must be a non-empty named character vector"
  )

})

test_that("type_check rejects unnamed expected", {

  expect_error(
    type_check(
      iris,
      "numeric"
    ),
    "'expected' must have variable names"
  )

})

test_that("type_check rejects blank expected name", {

  expected <- c(
    "numeric"
  )

  names(expected) <- ""

  expect_error(
    type_check(
      iris,
      expected
    ),
    "'expected' must have variable names"
  )

})

test_that("type_check rejects NA expected name", {

  expected <- c(
    "numeric"
  )

  names(expected) <- NA_character_

  expect_error(
    type_check(
      iris,
      expected
    ),
    "'expected' must have variable names"
  )

})

test_that("type_check rejects duplicate variable names", {

  expected <- c(
    "numeric",
    "numeric"
  )

  names(expected) <- c(
    "Sepal.Length",
    "Sepal.Length"
  )

  expect_error(
    type_check(
      iris,
      expected
    ),
    "Variable names in 'expected' must be unique"
  )

})


#==========================================================
# Unsupported types
#==========================================================

test_that("type_check rejects unsupported type", {

  expect_error(
    type_check(
      iris,
      c(
        Sepal.Length = "doubleXYZ"
      )
    ),
    "Unsupported expected type"
  )

})

test_that("type_check reports multiple unsupported types", {

  expect_error(
    type_check(
      iris,
      c(
        Sepal.Length = "doubleXYZ",
        Sepal.Width = "unknownXYZ"
      )
    ),
    "Unsupported expected type"
  )

})


#==========================================================
# Unknown variables
#==========================================================

test_that("type_check rejects unknown variable", {

  expect_error(
    type_check(
      iris,
      c(
        Unknown = "numeric"
      )
    ),
    "Variable\\(s\\) not found in 'data': Unknown"
  )

})

test_that("type_check reports multiple unknown variables", {

  expect_error(
    type_check(
      iris,
      c(
        Unknown1 = "numeric",
        Unknown2 = "character"
      )
    ),
    "Variable\\(s\\) not found in 'data'"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("type_check rejects NULL data", {

  expect_error(
    type_check(
      NULL,
      c(x = "numeric")
    ),
    "'data' cannot be NULL"
  )

})

test_that("type_check rejects vector data", {

  expect_error(
    type_check(
      1:10,
      c(x = "numeric")
    ),
    "'data' must be a data.frame"
  )

})

test_that("type_check rejects list data", {

  expect_error(
    type_check(
      list(x = 1:5),
      c(x = "numeric")
    ),
    "'data' must be a data.frame"
  )

})

test_that("type_check rejects empty data.frame", {

  expect_error(
    type_check(
      data.frame(),
      c(x = "numeric")
    ),
    "at least one variable"
  )

})

#==========================================================
# Other actual variable classes
#==========================================================

test_that("type_check reports unsupported actual class", {

  dat <- data.frame(
    z = c(
      1 + 2i,
      3 + 4i,
      5 + 6i
    )
  )

  result <- type_check(
    dat,
    expected = c(
      z = "numeric"
    )
  )

  expect_true(
    is.data.frame(result)
  )

  expect_equal(
    result$Actual,
    "complex"
  )

  expect_false(
    result$Match
  )

})

#==========================================================
# Unique Identifier Check
#==========================================================

#' Check Unique Identifier
#'
#' Evaluates whether a variable can serve as a unique
#' identifier by checking for missing and duplicated values.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   variable to check.
#'
#' @return A data.frame of class \code{UniqueCheck}
#'   containing row-level uniqueness diagnostics.
#'
#' @details
#' Each observation is evaluated for missing and duplicated
#' identifier values.
#'
#' A value is considered duplicated when it occurs more than
#' once among non-missing observations. All occurrences of a
#' repeated value are flagged.
#'
#' Missing values are flagged separately and are not treated
#' as duplicated values.
#'
#' The output contains:
#'
#' \itemize{
#'   \item \code{Row}: row number.
#'   \item \code{Value}: identifier value.
#'   \item \code{Missing}: whether the value is missing.
#'   \item \code{Duplicate}: whether the non-missing value
#'     occurs more than once.
#'   \item \code{ValidUnique}: whether the value is both
#'     non-missing and unique.
#' }
#'
#' @examples
#' dat <- data.frame(
#'   Animal_ID = c(101, 102, 102, 103, NA)
#' )
#'
#' unique_check(
#'   dat,
#'   "Animal_ID"
#' )
#'
#' @export

unique_check <- function(data, variable) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (missing(variable) ||
      !is.character(variable) ||
      length(variable) != 1L ||
      is.na(variable) ||
      !nzchar(variable)) {

    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {

    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_value <- is.na(x)

  #---------------------------------------------
  # Duplicated values
  #
  # duplicated(x) detects later duplicates.
  # duplicated(x, fromLast = TRUE) detects
  # earlier occurrences.
  #
  # Combining both flags every occurrence of a
  # repeated non-missing value.
  #---------------------------------------------
  duplicate_value <-
    !missing_value &
    (
      duplicated(x) |
        duplicated(
          x,
          fromLast = TRUE
        )
    )

  #---------------------------------------------
  # Valid unique identifier
  #---------------------------------------------
  valid_unique <-
    !missing_value &
    !duplicate_value

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = x,
    Missing = missing_value,
    Duplicate = duplicate_value,
    ValidUnique = valid_unique,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "UniqueCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("unique_check returns UniqueCheck", {

  dat <- data.frame(
    ID = 1:5
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_s3_class(
    result,
    "UniqueCheck"
  )

})

test_that("unique_check returns data.frame", {

  dat <- data.frame(
    ID = 1:5
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("unique_check contains expected columns", {

  dat <- data.frame(
    ID = 1:5
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "Duplicate",
      "ValidUnique"
    )
  )

})

test_that("unique_check returns one row per observation", {

  dat <- data.frame(
    ID = 1:10
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    nrow(result),
    10L
  )

})


#==========================================================
# Unique identifiers
#==========================================================

test_that("unique_check identifies unique values", {

  dat <- data.frame(
    ID = 101:105
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_false(
    any(result$Missing)
  )

  expect_false(
    any(result$Duplicate)
  )

  expect_true(
    all(result$ValidUnique)
  )

})


#==========================================================
# Duplicate identifiers
#==========================================================

test_that("unique_check detects duplicated values", {

  dat <- data.frame(
    ID = c(
      101,
      102,
      102,
      103
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

})

test_that("unique_check flags all occurrences of duplicate value", {

  dat <- data.frame(
    ID = c(
      1,
      2,
      2,
      2,
      3
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      TRUE,
      FALSE
    )
  )

})

test_that("unique_check handles multiple duplicate groups", {

  dat <- data.frame(
    ID = c(
      1,
      1,
      2,
      3,
      3,
      4
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      TRUE,
      TRUE,
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Missing identifiers
#==========================================================

test_that("unique_check detects missing values", {

  dat <- data.frame(
    ID = c(
      101,
      NA,
      103
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

})

test_that("missing identifiers are not classified as duplicates", {

  dat <- data.frame(
    ID = c(
      NA,
      NA,
      101
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Missing,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )

  expect_false(
    any(result$Duplicate)
  )

})

test_that("missing identifiers are not valid unique identifiers", {

  dat <- data.frame(
    ID = c(
      101,
      NA,
      102
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$ValidUnique,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Combined missing and duplicate
#==========================================================

test_that("unique_check handles missing and duplicated IDs", {

  dat <- data.frame(
    ID = c(
      101,
      102,
      102,
      NA,
      103
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$ValidUnique,
    c(
      TRUE,
      FALSE,
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Character identifiers
#==========================================================

test_that("unique_check handles character IDs", {

  dat <- data.frame(
    ID = c(
      "A001",
      "A002",
      "A002",
      "A003"
    ),
    stringsAsFactors = FALSE
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Factor identifiers
#==========================================================

test_that("unique_check handles factor IDs", {

  dat <- data.frame(
    ID = factor(
      c(
        "A",
        "B",
        "B",
        "C"
      )
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE,
      FALSE
    )
  )

  expect_true(
    is.factor(result$Value)
  )

})


#==========================================================
# Logical identifiers
#==========================================================

test_that("unique_check handles logical values", {

  dat <- data.frame(
    ID = c(
      TRUE,
      FALSE,
      TRUE
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Date identifiers
#==========================================================

test_that("unique_check handles Date values", {

  dat <- data.frame(
    ID = as.Date(
      c(
        "2026-01-01",
        "2026-01-02",
        "2026-01-02"
      )
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE
    )
  )

  expect_s3_class(
    result$Value,
    "Date"
  )

})


#==========================================================
# POSIXct identifiers
#==========================================================

test_that("unique_check handles POSIXct values", {

  dat <- data.frame(
    ID = as.POSIXct(
      c(
        "2026-01-01 10:00:00",
        "2026-01-01 11:00:00",
        "2026-01-01 11:00:00"
      ),
      tz = "UTC"
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_equal(
    result$Duplicate,
    c(
      FALSE,
      TRUE,
      TRUE
    )
  )

  expect_s3_class(
    result$Value,
    "POSIXct"
  )

})


#==========================================================
# Single observation
#==========================================================

test_that("unique_check handles one observation", {

  dat <- data.frame(
    ID = 101
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_false(
    result$Missing
  )

  expect_false(
    result$Duplicate
  )

  expect_true(
    result$ValidUnique
  )

})


#==========================================================
# All duplicated
#==========================================================

test_that("unique_check handles all duplicated values", {

  dat <- data.frame(
    ID = rep(
      101,
      5
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_true(
    all(result$Duplicate)
  )

  expect_false(
    any(result$ValidUnique)
  )

})


#==========================================================
# All missing
#==========================================================

test_that("unique_check handles all missing values", {

  dat <- data.frame(
    ID = rep(
      NA_integer_,
      5
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$Duplicate)
  )

  expect_false(
    any(result$ValidUnique)
  )

})


#==========================================================
# Zero rows
#==========================================================

test_that("unique_check handles zero-row data", {

  dat <- data.frame(
    ID = integer(0)
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_s3_class(
    result,
    "UniqueCheck"
  )

  expect_equal(
    nrow(result),
    0L
  )

  expect_equal(
    length(result$Duplicate),
    0L
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("unique_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    ID = c(
      1,
      2,
      2,
      3
    )
  )

  result <- unique_check(
    dat,
    "ID"
  )

  expect_s3_class(
    result,
    "UniqueCheck"
  )

  expect_equal(
    sum(result$Duplicate),
    2L
  )

})


#==========================================================
# Variable validation
#==========================================================

test_that("unique_check rejects missing variable", {

  dat <- data.frame(
    ID = 1:5
  )

  expect_error(
    unique_check(dat),
    "'variable' must be a single non-empty character string"
  )

})

test_that("unique_check rejects NULL variable", {

  expect_error(
    unique_check(
      iris,
      NULL
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("unique_check rejects numeric variable argument", {

  expect_error(
    unique_check(
      iris,
      1
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("unique_check rejects multiple variable names", {

  expect_error(
    unique_check(
      iris,
      c(
        "Sepal.Length",
        "Sepal.Width"
      )
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("unique_check rejects NA variable name", {

  expect_error(
    unique_check(
      iris,
      NA_character_
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("unique_check rejects empty variable name", {

  expect_error(
    unique_check(
      iris,
      ""
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("unique_check rejects unknown variable", {

  expect_error(
    unique_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("unique_check rejects NULL data", {

  expect_error(
    unique_check(
      NULL,
      "ID"
    ),
    "'data' cannot be NULL"
  )

})

test_that("unique_check rejects numeric vector", {

  expect_error(
    unique_check(
      1:10,
      "ID"
    ),
    "'data' must be a data.frame"
  )

})

test_that("unique_check rejects character vector", {

  expect_error(
    unique_check(
      letters,
      "ID"
    ),
    "'data' must be a data.frame"
  )

})

test_that("unique_check rejects list", {

  expect_error(
    unique_check(
      list(ID = 1:5),
      "ID"
    ),
    "'data' must be a data.frame"
  )

})

test_that("unique_check rejects empty data.frame", {

  expect_error(
    unique_check(
      data.frame(),
      "ID"
    ),
    "at least one variable"
  )

})

#==========================================================
# Consistency Check
#==========================================================

#' Check Consistency Between Two Variables
#'
#' Checks whether values in two variables satisfy a specified
#' logical relationship.
#'
#' @param data A data.frame.
#' @param variable1 A single character string specifying the
#'   first variable.
#' @param variable2 A single character string specifying the
#'   second variable.
#' @param relationship A character string specifying the
#'   expected relationship between the variables. Supported
#'   relationships are \code{"<"}, \code{"<="}, \code{">"},
#'   \code{">="}, \code{"=="}, and \code{"!="}.
#'
#' @return A data.frame of class \code{ConsistencyCheck}
#'   containing row-level consistency diagnostics.
#'
#' @details
#' The function evaluates the relationship as
#' \code{variable1 relationship variable2}.
#'
#' For example, when checking whether an end date occurs on
#' or after a start date, use:
#'
#' \code{variable1 = "EndDate"},
#' \code{variable2 = "StartDate"}, and
#' \code{relationship = ">="}.
#'
#' Rows containing a missing value in either variable are
#' reported separately and are not classified as consistent.
#'
#' The two variables must have compatible types for the
#' requested comparison.
#'
#' @examples
#' dat <- data.frame(
#'   StartDate = as.Date(c(
#'     "2026-01-01",
#'     "2026-02-01",
#'     "2026-03-01"
#'   )),
#'   EndDate = as.Date(c(
#'     "2026-01-10",
#'     "2026-01-20",
#'     "2026-03-15"
#'   ))
#' )
#'
#' consistency_check(
#'   dat,
#'   variable1 = "EndDate",
#'   variable2 = "StartDate",
#'   relationship = ">="
#' )
#'
#' @export

consistency_check <- function(
    data,
    variable1,
    variable2,
    relationship
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Helper: validate variable name
  #---------------------------------------------
  validate_variable <- function(variable, argument) {

    if (
      !is.character(variable) ||
      length(variable) != 1L ||
      is.na(variable) ||
      !nzchar(variable)
    ) {
      stop(
        sprintf(
          "'%s' must be a single non-empty character string.",
          argument
        ),
        call. = FALSE
      )
    }

    if (!variable %in% names(data)) {
      stop(
        sprintf(
          "Variable '%s' was not found in 'data'.",
          variable
        ),
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  #---------------------------------------------
  # Validate required variables
  #---------------------------------------------
  if (missing(variable1)) {
    stop(
      "'variable1' must be supplied.",
      call. = FALSE
    )
  }

  if (missing(variable2)) {
    stop(
      "'variable2' must be supplied.",
      call. = FALSE
    )
  }

  validate_variable(
    variable1,
    "variable1"
  )

  validate_variable(
    variable2,
    "variable2"
  )

  #---------------------------------------------
  # Variables must differ
  #---------------------------------------------
  if (identical(variable1, variable2)) {
    stop(
      "'variable1' and 'variable2' must refer to different variables.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate relationship
  #---------------------------------------------
  supported_relationships <- c(
    "<",
    "<=",
    ">",
    ">=",
    "==",
    "!="
  )

  if (
    missing(relationship) ||
    !is.character(relationship) ||
    length(relationship) != 1L ||
    is.na(relationship) ||
    !relationship %in% supported_relationships
  ) {
    stop(
      paste0(
        "'relationship' must be one of: ",
        "'<', '<=', '>', '>=', '==', or '!='."
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variables
  #---------------------------------------------
  x <- data[[variable1]]
  y <- data[[variable2]]

  #---------------------------------------------
  # Missing observations
  #---------------------------------------------
  missing_value <- is.na(x) | is.na(y)

  #---------------------------------------------
  # Evaluate relationship
  #---------------------------------------------
  consistent <- rep(
    FALSE,
    length(x)
  )

  valid_index <- which(!missing_value)

  if (length(valid_index) > 0L) {

    comparison <- tryCatch(
      {
        switch(
          relationship,
          "<" = x[valid_index] < y[valid_index],
          "<=" = x[valid_index] <= y[valid_index],
          ">" = x[valid_index] > y[valid_index],
          ">=" = x[valid_index] >= y[valid_index],
          "==" = x[valid_index] == y[valid_index],
          "!=" = x[valid_index] != y[valid_index]
        )
      },
      error = function(e) {
        stop(
          paste0(
            "The selected variables cannot be compared ",
            "using relationship '",
            relationship,
            "'."
          ),
          call. = FALSE
        )
      }
    )

    #-------------------------------------------
    # Protect against undefined comparisons
    #-------------------------------------------
    comparison[is.na(comparison)] <- FALSE

    consistent[valid_index] <- comparison
  }

  #---------------------------------------------
  # Inconsistency
  #---------------------------------------------
  inconsistent <-
    !missing_value &
    !consistent

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value1 = x,
    Value2 = y,
    Missing = missing_value,
    Consistent = consistent,
    Inconsistent = inconsistent,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "ConsistencyCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("consistency_check returns correct class", {

  dat <- data.frame(
    x = c(1, 2, 3),
    y = c(2, 2, 1)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<="
  )

  expect_s3_class(
    result,
    "ConsistencyCheck"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("consistency_check returns expected columns", {

  dat <- data.frame(
    x = 1:3,
    y = 3:1
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value1",
      "Value2",
      "Missing",
      "Consistent",
      "Inconsistent"
    )
  )
})

#==========================================================
# Less than
#==========================================================

test_that("consistency_check evaluates less than", {

  dat <- data.frame(
    x = c(1, 3, 5),
    y = c(2, 3, 4)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<"
  )

  expect_equal(
    result$Consistent,
    c(TRUE, FALSE, FALSE)
  )
})

#==========================================================
# Less than or equal
#==========================================================

test_that("consistency_check evaluates less than or equal", {

  dat <- data.frame(
    x = c(1, 3, 5),
    y = c(2, 3, 4)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<="
  )

  expect_equal(
    result$Consistent,
    c(TRUE, TRUE, FALSE)
  )
})

#==========================================================
# Greater than
#==========================================================

test_that("consistency_check evaluates greater than", {

  dat <- data.frame(
    x = c(3, 2, 1),
    y = c(2, 2, 2)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    ">"
  )

  expect_equal(
    result$Consistent,
    c(TRUE, FALSE, FALSE)
  )
})

#==========================================================
# Greater than or equal
#==========================================================

test_that("consistency_check evaluates greater than or equal", {

  dat <- data.frame(
    x = c(3, 2, 1),
    y = c(2, 2, 2)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    ">="
  )

  expect_equal(
    result$Consistent,
    c(TRUE, TRUE, FALSE)
  )
})

#==========================================================
# Equal
#==========================================================

test_that("consistency_check evaluates equality", {

  dat <- data.frame(
    x = c(1, 2, 3),
    y = c(1, 5, 3)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "=="
  )

  expect_equal(
    result$Consistent,
    c(TRUE, FALSE, TRUE)
  )
})

#==========================================================
# Not equal
#==========================================================

test_that("consistency_check evaluates inequality", {

  dat <- data.frame(
    x = c(1, 2, 3),
    y = c(1, 5, 3)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "!="
  )

  expect_equal(
    result$Consistent,
    c(FALSE, TRUE, FALSE)
  )
})

#==========================================================
# Inconsistent
#==========================================================

test_that("inconsistent is inverse for non-missing rows", {

  dat <- data.frame(
    x = c(1, 5, 2),
    y = c(2, 3, 2)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<="
  )

  expect_equal(
    result$Inconsistent,
    c(FALSE, TRUE, FALSE)
  )
})

#==========================================================
# Missing values
#==========================================================

test_that("consistency_check handles missing values", {

  dat <- data.frame(
    x = c(1, NA, 3, 4),
    y = c(2, 2, NA, 1)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<"
  )

  expect_equal(
    result$Missing,
    c(FALSE, TRUE, TRUE, FALSE)
  )

  expect_false(
    result$Consistent[2]
  )

  expect_false(
    result$Consistent[3]
  )

  expect_false(
    result$Inconsistent[2]
  )

  expect_false(
    result$Inconsistent[3]
  )
})

test_that("consistency_check handles all missing comparisons", {

  dat <- data.frame(
    x = c(NA_real_, NA_real_),
    y = c(NA_real_, NA_real_)
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "<="
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$Consistent)
  )

  expect_false(
    any(result$Inconsistent)
  )
})

#==========================================================
# Date comparisons
#==========================================================

test_that("consistency_check compares Date variables", {

  dat <- data.frame(
    Start = as.Date(
      c(
        "2026-01-01",
        "2026-02-10",
        "2026-03-01"
      )
    ),
    End = as.Date(
      c(
        "2026-01-05",
        "2026-02-01",
        "2026-03-01"
      )
    )
  )

  result <- consistency_check(
    dat,
    "End",
    "Start",
    ">="
  )

  expect_equal(
    result$Consistent,
    c(TRUE, FALSE, TRUE)
  )
})

#==========================================================
# Character comparisons
#==========================================================

test_that("consistency_check evaluates character equality", {

  dat <- data.frame(
    x = c("Dog", "Cat", "Cattle"),
    y = c("Dog", "Dog", "Cattle")
  )

  result <- consistency_check(
    dat,
    "x",
    "y",
    "=="
  )

  expect_equal(
    result$Consistent,
    c(TRUE, FALSE, TRUE)
  )
})

#==========================================================
# Variable validation
#==========================================================

test_that("consistency_check requires variable1", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      variable2 = "y",
      relationship = "=="
    ),
    "'variable1' must be supplied"
  )
})

test_that("consistency_check requires variable2", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      variable1 = "x",
      relationship = "=="
    ),
    "'variable2' must be supplied"
  )
})

test_that("consistency_check validates variable names", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      NULL,
      "y",
      "=="
    ),
    "'variable1' must be a single non-empty character string"
  )

  expect_error(
    consistency_check(
      dat,
      "",
      "y",
      "=="
    ),
    "'variable1' must be a single non-empty character string"
  )

  expect_error(
    consistency_check(
      dat,
      1,
      "y",
      "=="
    ),
    "'variable1' must be a single non-empty character string"
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      NA_character_,
      "=="
    ),
    "'variable2' must be a single non-empty character string"
  )

  expect_error(
    consistency_check(
      dat,
      c("x", "y"),
      "y",
      "=="
    ),
    "'variable1' must be a single non-empty character string"
  )
})

test_that("consistency_check rejects unknown variables", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      "unknown",
      "y",
      "=="
    ),
    "Variable 'unknown' was not found in 'data'"
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "unknown",
      "=="
    ),
    "Variable 'unknown' was not found in 'data'"
  )
})

test_that("consistency_check rejects same variable", {

  dat <- data.frame(
    x = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "x",
      "=="
    ),
    "'variable1' and 'variable2' must refer to different variables"
  )
})

#==========================================================
# Relationship validation
#==========================================================

test_that("consistency_check requires relationship", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y"
    ),
    "'relationship' must be one of"
  )
})

test_that("consistency_check rejects unsupported relationship", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      "="
    ),
    "'relationship' must be one of"
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      "approximately"
    ),
    "'relationship' must be one of"
  )
})

test_that("consistency_check rejects malformed relationship", {

  dat <- data.frame(
    x = 1:3,
    y = 1:3
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      NULL
    ),
    "'relationship' must be one of"
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      NA_character_
    ),
    "'relationship' must be one of"
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      c("<", ">")
    ),
    "'relationship' must be one of"
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      1
    ),
    "'relationship' must be one of"
  )
})

#==========================================================
# Incompatible variables
#==========================================================

test_that("consistency_check rejects incompatible comparison", {

  dat <- data.frame(
    x = as.Date(
      c(
        "2026-01-01",
        "2026-01-02"
      )
    ),
    y = c(
      "Dog",
      "Cat"
    )
  )

  expect_error(
    consistency_check(
      dat,
      "x",
      "y",
      "<"
    ),
    "cannot be compared"
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("consistency_check rejects invalid data", {

  expect_error(
    consistency_check(
      NULL,
      "x",
      "y",
      "=="
    )
  )

  expect_error(
    consistency_check(
      1:5,
      "x",
      "y",
      "=="
    )
  )

  expect_error(
    consistency_check(
      list(x = 1, y = 1),
      "x",
      "y",
      "=="
    )
  )

  expect_error(
    consistency_check(
      data.frame(),
      "x",
      "y",
      "=="
    )
  )
})

#==========================================================
# Infinite Value Check
#==========================================================

#' Check Infinite Values
#'
#' Identifies positive and negative infinite values in
#' numeric variables of a dataset.
#'
#' @param data A data.frame.
#'
#' @return A data.frame of class \code{InfiniteCheck}
#'   containing variable-level information about infinite
#'   values.
#'
#' @details
#' Infinite values are detected only in numeric variables.
#' Positive infinity (\code{Inf}) and negative infinity
#' (\code{-Inf}) are reported separately.
#'
#' Missing values (\code{NA}) are not classified as
#' infinite values.
#'
#' Non-numeric variables are retained in the output and
#' reported with zero infinite values.
#'
#' @examples
#' dat <- data.frame(
#'   x = c(1, 2, Inf, 4),
#'   y = c(1, -Inf, 3, 4),
#'   group = c("A", "B", "C", "D")
#' )
#'
#' infinite_check(dat)
#'
#' @export

infinite_check <- function(data) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Analyze one variable
  #---------------------------------------------
  analyse_variable <- function(x) {

    # Non-numeric variables cannot contain
    # numeric infinite values
    if (!is.numeric(x)) {

      return(
        c(
          PositiveInf = 0L,
          NegativeInf = 0L,
          Infinite = 0L
        )
      )
    }

    #-------------------------------------------
    # Positive infinity
    #-------------------------------------------
    positive_inf <- sum(
      is.infinite(x) & x > 0,
      na.rm = TRUE
    )

    #-------------------------------------------
    # Negative infinity
    #-------------------------------------------
    negative_inf <- sum(
      is.infinite(x) & x < 0,
      na.rm = TRUE
    )

    #-------------------------------------------
    # Total infinity
    #-------------------------------------------
    infinite <- positive_inf + negative_inf

    c(
      PositiveInf = as.integer(positive_inf),
      NegativeInf = as.integer(negative_inf),
      Infinite = as.integer(infinite)
    )
  }

  #---------------------------------------------
  # Apply check
  #---------------------------------------------
  diagnostics <- lapply(
    data,
    analyse_variable
  )

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(

    Variable = names(data),

    Numeric = vapply(
      data,
      is.numeric,
      logical(1)
    ),

    PositiveInf = vapply(
      diagnostics,
      function(x) x[["PositiveInf"]],
      integer(1)
    ),

    NegativeInf = vapply(
      diagnostics,
      function(x) x[["NegativeInf"]],
      integer(1)
    ),

    Infinite = vapply(
      diagnostics,
      function(x) x[["Infinite"]],
      integer(1)
    ),

    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Flag variables containing infinity
  #---------------------------------------------
  result$HasInfinite <- result$Infinite > 0L

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "InfiniteCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("infinite_check returns InfiniteCheck object", {

  result <- infinite_check(iris)

  expect_s3_class(
    result,
    "InfiniteCheck"
  )

})

test_that("infinite_check returns data.frame", {

  result <- infinite_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("infinite_check contains expected columns", {

  result <- infinite_check(iris)

  expect_equal(
    names(result),
    c(
      "Variable",
      "Numeric",
      "PositiveInf",
      "NegativeInf",
      "Infinite",
      "HasInfinite"
    )
  )

})

test_that("infinite_check returns one row per variable", {

  result <- infinite_check(iris)

  expect_equal(
    nrow(result),
    ncol(iris)
  )

})


#==========================================================
# Data without infinite values
#==========================================================

test_that("infinite_check reports no infinity for iris", {

  result <- infinite_check(iris)

  expect_true(
    all(result$Infinite == 0L)
  )

  expect_false(
    any(result$HasInfinite)
  )

})


#==========================================================
# Positive infinity
#==========================================================

test_that("infinite_check detects positive infinity", {

  dat <- data.frame(
    x = c(
      1,
      2,
      Inf,
      4
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    result$PositiveInf,
    1L
  )

  expect_equal(
    result$NegativeInf,
    0L
  )

  expect_equal(
    result$Infinite,
    1L
  )

  expect_true(
    result$HasInfinite
  )

})


#==========================================================
# Negative infinity
#==========================================================

test_that("infinite_check detects negative infinity", {

  dat <- data.frame(
    x = c(
      1,
      -Inf,
      3,
      4
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    result$PositiveInf,
    0L
  )

  expect_equal(
    result$NegativeInf,
    1L
  )

  expect_equal(
    result$Infinite,
    1L
  )

  expect_true(
    result$HasInfinite
  )

})


#==========================================================
# Positive and negative infinity
#==========================================================

test_that("infinite_check detects both infinity types", {

  dat <- data.frame(
    x = c(
      Inf,
      -Inf,
      1,
      Inf,
      -Inf
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    result$PositiveInf,
    2L
  )

  expect_equal(
    result$NegativeInf,
    2L
  )

  expect_equal(
    result$Infinite,
    4L
  )

  expect_true(
    result$HasInfinite
  )

})


#==========================================================
# Multiple variables
#==========================================================

test_that("infinite_check handles multiple numeric variables", {

  dat <- data.frame(
    x = c(1, Inf, 3),
    y = c(-Inf, 2, 3),
    z = c(1, 2, 3)
  )

  result <- infinite_check(dat)

  expect_equal(
    result$PositiveInf,
    c(
      1L,
      0L,
      0L
    )
  )

  expect_equal(
    result$NegativeInf,
    c(
      0L,
      1L,
      0L
    )
  )

  expect_equal(
    result$Infinite,
    c(
      1L,
      1L,
      0L
    )
  )

  expect_equal(
    result$HasInfinite,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("infinite_check does not classify NA as infinite", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      3,
      4
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    result$Infinite,
    0L
  )

  expect_false(
    result$HasInfinite
  )

})

test_that("infinite_check handles NA and infinity together", {

  dat <- data.frame(
    x = c(
      NA,
      Inf,
      -Inf,
      NA
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    result$PositiveInf,
    1L
  )

  expect_equal(
    result$NegativeInf,
    1L
  )

  expect_equal(
    result$Infinite,
    2L
  )

})


#==========================================================
# Numeric classification
#==========================================================

test_that("infinite_check identifies numeric variables", {

  dat <- data.frame(
    x = 1:5,
    y = letters[1:5],
    stringsAsFactors = FALSE
  )

  result <- infinite_check(dat)

  expect_equal(
    result$Numeric,
    c(
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Character variables
#==========================================================

test_that("infinite_check handles character variables", {

  dat <- data.frame(
    x = c(
      "Inf",
      "-Inf",
      "A"
    ),
    stringsAsFactors = FALSE
  )

  result <- infinite_check(dat)

  expect_false(
    result$Numeric
  )

  expect_equal(
    result$PositiveInf,
    0L
  )

  expect_equal(
    result$NegativeInf,
    0L
  )

  expect_equal(
    result$Infinite,
    0L
  )

  expect_false(
    result$HasInfinite
  )

})


#==========================================================
# Factor variables
#==========================================================

test_that("infinite_check handles factors", {

  dat <- data.frame(
    x = factor(
      c(
        "Inf",
        "-Inf",
        "A"
      )
    )
  )

  result <- infinite_check(dat)

  expect_false(
    result$Numeric
  )

  expect_equal(
    result$Infinite,
    0L
  )

})


#==========================================================
# Logical variables
#==========================================================

test_that("infinite_check handles logical variables", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE,
      NA
    )
  )

  result <- infinite_check(dat)

  expect_false(
    result$Numeric
  )

  expect_equal(
    result$Infinite,
    0L
  )

})


#==========================================================
# Date variables
#==========================================================

test_that("infinite_check handles Date variables", {

  dat <- data.frame(
    x = as.Date("2026-01-01") + 0:4
  )

  result <- infinite_check(dat)

  expect_false(
    result$Numeric
  )

  expect_equal(
    result$Infinite,
    0L
  )

})


#==========================================================
# POSIXct variables
#==========================================================

test_that("infinite_check handles POSIXct variables", {

  dat <- data.frame(
    x = as.POSIXct(
      "2026-01-01 10:00:00",
      tz = "UTC"
    ) + 0:4
  )

  result <- infinite_check(dat)

  expect_false(
    result$Numeric
  )

  expect_equal(
    result$Infinite,
    0L
  )

})


#==========================================================
# Mixed data
#==========================================================

test_that("infinite_check handles mixed data", {

  dat <- data.frame(
    numeric = c(
      1,
      Inf,
      -Inf
    ),
    character = c(
      "A",
      "Inf",
      "B"
    ),
    factor = factor(
      c(
        "A",
        "B",
        "C"
      )
    ),
    logical = c(
      TRUE,
      FALSE,
      TRUE
    ),
    stringsAsFactors = FALSE
  )

  result <- infinite_check(dat)

  expect_equal(
    result$Numeric,
    c(
      TRUE,
      FALSE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$Infinite,
    c(
      2L,
      0L,
      0L,
      0L
    )
  )

})


#==========================================================
# All infinite
#==========================================================

test_that("infinite_check handles all-infinite variable", {

  dat <- data.frame(
    x = c(
      Inf,
      Inf,
      -Inf,
      -Inf
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    result$PositiveInf,
    2L
  )

  expect_equal(
    result$NegativeInf,
    2L
  )

  expect_equal(
    result$Infinite,
    4L
  )

  expect_true(
    result$HasInfinite
  )

})


#==========================================================
# One-column data
#==========================================================

test_that("infinite_check works with one-column data", {

  dat <- data.frame(
    x = c(
      1,
      Inf,
      3
    )
  )

  result <- infinite_check(dat)

  expect_equal(
    nrow(result),
    1L
  )

  expect_equal(
    result$Variable,
    "x"
  )

  expect_equal(
    result$Infinite,
    1L
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("infinite_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- infinite_check(dat)

  expect_s3_class(
    result,
    "InfiniteCheck"
  )

  expect_equal(
    nrow(result),
    2L
  )

  expect_equal(
    result$Numeric,
    c(
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$Infinite,
    c(
      0L,
      0L
    )
  )

  expect_false(
    any(result$HasInfinite)
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("infinite_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = c(
      1,
      Inf,
      -Inf
    ),
    y = c(
      "A",
      "B",
      "C"
    )
  )

  result <- infinite_check(dat)

  expect_s3_class(
    result,
    "InfiniteCheck"
  )

  expect_equal(
    result$Infinite,
    c(
      2L,
      0L
    )
  )

})


#==========================================================
# Variable names
#==========================================================

test_that("infinite_check preserves variable names", {

  dat <- data.frame(
    Weight = c(10, Inf, 20),
    Temperature = c(38, 39, 40)
  )

  result <- infinite_check(dat)

  expect_equal(
    result$Variable,
    c(
      "Weight",
      "Temperature"
    )
  )

})


#==========================================================
# Invalid input
#==========================================================

test_that("infinite_check rejects NULL", {

  expect_error(
    infinite_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("infinite_check rejects numeric vector", {

  expect_error(
    infinite_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("infinite_check rejects character vector", {

  expect_error(
    infinite_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("infinite_check rejects list", {

  expect_error(
    infinite_check(
      list(
        x = 1:5,
        y = 6:10
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("infinite_check rejects data.frame without variables", {

  expect_error(
    infinite_check(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Audit Rules
#==========================================================

#' Create Data Audit Rules
#'
#' Creates a structured collection of user-defined data
#' validation rules for use with DataAudit.
#'
#' @param range Named list. Each element specifies the
#'   acceptable numeric range for a variable as a numeric
#'   vector of length two.
#' @param category Named list. Each element contains the
#'   allowed values for a variable.
#' @param unique Character vector containing variables that
#'   should contain unique values.
#' @param required Character vector containing variables
#'   that should not contain missing values.
#' @param pattern Named list. Each element specifies a
#'   regular expression that values of a variable should
#'   satisfy.
#'
#' @return An object of class \code{DataAuditRules}.
#'
#' @details
#' \code{audit_rules()} defines validation requirements but
#' does not evaluate a dataset. The resulting object can be
#' supplied to functions that apply DataAudit rules.
#'
#' Variable names are not checked against a dataset when the
#' rule object is created because the same rule specification
#' may be reused across multiple datasets.
#'
#' @examples
#' rules <- audit_rules(
#'   range = list(
#'     Age = c(0, 30),
#'     Weight = c(0, 100)
#'   ),
#'   category = list(
#'     Species = c("Dog", "Cat", "Cattle")
#'   ),
#'   unique = "Animal_ID",
#'   required = c("Animal_ID", "Species"),
#'   pattern = list(
#'     Animal_ID = "^A[0-9]+$"
#'   )
#' )
#'
#' rules
#'
#' @export

audit_rules <- function(
    range = NULL,
    category = NULL,
    unique = NULL,
    required = NULL,
    pattern = NULL
) {

  #---------------------------------------------
  # Helper: validate named list
  #---------------------------------------------
  validate_named_list <- function(
    x,
    argument
  ) {

    if (is.null(x)) {
      return(invisible(TRUE))
    }

    if (!is.list(x)) {
      stop(
        sprintf(
          "'%s' must be a named list or NULL.",
          argument
        ),
        call. = FALSE
      )
    }

    if (length(x) == 0L) {
      return(invisible(TRUE))
    }

    nms <- names(x)

    if (
      is.null(nms) ||
      any(is.na(nms)) ||
      any(!nzchar(nms))
    ) {
      stop(
        sprintf(
          "'%s' must be a named list with non-empty names.",
          argument
        ),
        call. = FALSE
      )
    }

    if (anyDuplicated(nms)) {
      stop(
        sprintf(
          "'%s' contains duplicated variable names.",
          argument
        ),
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  #---------------------------------------------
  # Helper: validate variable vector
  #---------------------------------------------
  validate_variables <- function(
    x,
    argument
  ) {

    if (is.null(x)) {
      return(invisible(TRUE))
    }

    if (
      !is.character(x) ||
      anyNA(x) ||
      any(!nzchar(x))
    ) {
      stop(
        sprintf(
          "'%s' must be a character vector of non-empty variable names or NULL.",
          argument
        ),
        call. = FALSE
      )
    }

    if (anyDuplicated(x)) {
      stop(
        sprintf(
          "'%s' contains duplicated variable names.",
          argument
        ),
        call. = FALSE
      )
    }

    invisible(TRUE)
  }

  #---------------------------------------------
  # Validate range
  #---------------------------------------------
  validate_named_list(
    range,
    "range"
  )

  if (!is.null(range)) {

    for (variable in names(range)) {

      limits <- range[[variable]]

      if (
        !is.numeric(limits) ||
        length(limits) != 2L ||
        anyNA(limits) ||
        any(!is.finite(limits))
      ) {
        stop(
          sprintf(
            "Range rule for '%s' must contain two finite numeric values.",
            variable
          ),
          call. = FALSE
        )
      }

      if (limits[1L] > limits[2L]) {
        stop(
          sprintf(
            "Minimum range value cannot exceed maximum for '%s'.",
            variable
          ),
          call. = FALSE
        )
      }
    }
  }

  #---------------------------------------------
  # Validate category
  #---------------------------------------------
  validate_named_list(
    category,
    "category"
  )

  if (!is.null(category)) {

    for (variable in names(category)) {

      allowed <- category[[variable]]

      if (
        !is.atomic(allowed) ||
        length(allowed) == 0L
      ) {
        stop(
          sprintf(
            "Category rule for '%s' must contain a non-empty atomic vector.",
            variable
          ),
          call. = FALSE
        )
      }

      if (anyNA(allowed)) {
        stop(
          sprintf(
            "Category rule for '%s' cannot contain missing values.",
            variable
          ),
          call. = FALSE
        )
      }
    }
  }

  #---------------------------------------------
  # Validate unique and required
  #---------------------------------------------
  validate_variables(
    unique,
    "unique"
  )

  validate_variables(
    required,
    "required"
  )

  #---------------------------------------------
  # Validate pattern
  #---------------------------------------------
  validate_named_list(
    pattern,
    "pattern"
  )

  if (!is.null(pattern)) {

    for (variable in names(pattern)) {

      regex <- pattern[[variable]]

      if (
        !is.character(regex) ||
        length(regex) != 1L ||
        is.na(regex) ||
        !nzchar(regex)
      ) {
        stop(
          sprintf(
            "Pattern rule for '%s' must be a single non-empty character string.",
            variable
          ),
          call. = FALSE
        )
      }

      valid_regex <- tryCatch(
        {
          suppressWarnings(
            grepl(
              regex,
              "",
              perl = TRUE
            )
          )
          TRUE
        },
        error = function(e) {
          FALSE
        }
      )

      if (!valid_regex) {
        stop(
          sprintf(
            "Pattern rule for '%s' is not a valid regular expression.",
            variable
          ),
          call. = FALSE
        )
      }
    }
  }

  #---------------------------------------------
  # Construct object
  #---------------------------------------------
  result <- list(
    range = range,
    category = category,
    unique = unique,
    required = required,
    pattern = pattern
  )

  class(result) <- c(
    "DataAuditRules",
    "list"
  )

  result
}

library(testthat)
library(DataAudit)

test_that("audit_rules returns correct class", {

  rules <- audit_rules()

  expect_s3_class(
    rules,
    "DataAuditRules"
  )

  expect_type(
    rules,
    "list"
  )
})


test_that("audit_rules has expected components", {

  rules <- audit_rules()

  expect_equal(
    names(rules),
    c(
      "range",
      "category",
      "unique",
      "required",
      "pattern"
    )
  )
})


test_that("empty rules are allowed", {

  rules <- audit_rules()

  expect_true(
    all(
      vapply(
        rules,
        is.null,
        logical(1)
      )
    )
  )
})


test_that("valid range rules are accepted", {

  expect_no_error(
    audit_rules(
      range = list(
        Age = c(0, 30),
        Weight = c(1, 100)
      )
    )
  )
})


test_that("range must be named list", {

  expect_error(
    audit_rules(
      range = c(0, 10)
    ),
    "'range' must be a named list or NULL"
  )

  expect_error(
    audit_rules(
      range = list(
        c(0, 10)
      )
    ),
    "'range' must be a named list with non-empty names"
  )
})


test_that("range names cannot be duplicated", {

  x <- list(
    c(0, 10),
    c(0, 20)
  )

  names(x) <- c(
    "Age",
    "Age"
  )

  expect_error(
    audit_rules(
      range = x
    ),
    "'range' contains duplicated variable names"
  )
})


test_that("range limits must be numeric length two", {

  expect_error(
    audit_rules(
      range = list(
        Age = 10
      )
    ),
    "must contain two finite numeric values"
  )

  expect_error(
    audit_rules(
      range = list(
        Age = c(
          "0",
          "10"
        )
      )
    ),
    "must contain two finite numeric values"
  )

  expect_error(
    audit_rules(
      range = list(
        Age = c(
          0,
          NA
        )
      )
    ),
    "must contain two finite numeric values"
  )

  expect_error(
    audit_rules(
      range = list(
        Age = c(
          0,
          Inf
        )
      )
    ),
    "must contain two finite numeric values"
  )
})


test_that("range minimum cannot exceed maximum", {

  expect_error(
    audit_rules(
      range = list(
        Age = c(
          30,
          0
        )
      )
    ),
    "Minimum range value cannot exceed maximum"
  )
})


test_that("valid category rules are accepted", {

  expect_no_error(
    audit_rules(
      category = list(
        Species = c(
          "Dog",
          "Cat",
          "Cattle"
        )
      )
    )
  )
})


test_that("category must be named list", {

  expect_error(
    audit_rules(
      category = c(
        "Dog",
        "Cat"
      )
    ),
    "'category' must be a named list or NULL"
  )

  expect_error(
    audit_rules(
      category = list(
        c(
          "Dog",
          "Cat"
        )
      )
    ),
    "'category' must be a named list with non-empty names"
  )
})


test_that("category rules must contain atomic values", {

  expect_error(
    audit_rules(
      category = list(
        Species = list(
          "Dog",
          "Cat"
        )
      )
    ),
    "must contain a non-empty atomic vector"
  )

  expect_error(
    audit_rules(
      category = list(
        Species = character()
      )
    ),
    "must contain a non-empty atomic vector"
  )
})


test_that("category values cannot contain missing values", {

  expect_error(
    audit_rules(
      category = list(
        Species = c(
          "Dog",
          NA
        )
      )
    ),
    "cannot contain missing values"
  )
})


test_that("unique variables are accepted", {

  expect_no_error(
    audit_rules(
      unique = c(
        "Animal_ID",
        "Sample_ID"
      )
    )
  )
})


test_that("required variables are accepted", {

  expect_no_error(
    audit_rules(
      required = c(
        "Animal_ID",
        "Species"
      )
    )
  )
})


test_that("unique must contain valid variable names", {

  expect_error(
    audit_rules(
      unique = 1
    ),
    "'unique' must be a character vector"
  )

  expect_error(
    audit_rules(
      unique = c(
        "ID",
        NA
      )
    ),
    "'unique' must be a character vector"
  )

  expect_error(
    audit_rules(
      unique = c(
        "ID",
        ""
      )
    ),
    "'unique' must be a character vector"
  )
})


test_that("required must contain valid variable names", {

  expect_error(
    audit_rules(
      required = 1
    ),
    "'required' must be a character vector"
  )

  expect_error(
    audit_rules(
      required = NA_character_
    ),
    "'required' must be a character vector"
  )
})


test_that("duplicated unique variables are rejected", {

  expect_error(
    audit_rules(
      unique = c(
        "ID",
        "ID"
      )
    ),
    "'unique' contains duplicated variable names"
  )
})


test_that("duplicated required variables are rejected", {

  expect_error(
    audit_rules(
      required = c(
        "Species",
        "Species"
      )
    ),
    "'required' contains duplicated variable names"
  )
})


test_that("valid pattern rules are accepted", {

  expect_no_error(
    audit_rules(
      pattern = list(
        Animal_ID =
          "^A[0-9]+$"
      )
    )
  )
})


test_that("pattern must be named list", {

  expect_error(
    audit_rules(
      pattern = "^A"
    ),
    "'pattern' must be a named list or NULL"
  )

  expect_error(
    audit_rules(
      pattern = list(
        "^A"
      )
    ),
    "'pattern' must be a named list with non-empty names"
  )
})


test_that("pattern must contain scalar strings", {

  expect_error(
    audit_rules(
      pattern = list(
        ID = 1
      )
    ),
    "must be a single non-empty character string"
  )

  expect_error(
    audit_rules(
      pattern = list(
        ID = ""
      )
    ),
    "must be a single non-empty character string"
  )

  expect_error(
    audit_rules(
      pattern = list(
        ID = c(
          "^A",
          "^B"
        )
      )
    ),
    "must be a single non-empty character string"
  )
})


test_that("invalid regular expression is rejected without warning", {

  expect_error(
    audit_rules(
      pattern = list(
        ID = "["
      )
    ),
    "is not a valid regular expression"
  )
})


test_that("all rule types can be combined", {

  rules <- audit_rules(
    range = list(
      Age = c(
        0,
        30
      )
    ),
    category = list(
      Species = c(
        "Dog",
        "Cat"
      )
    ),
    unique = "Animal_ID",
    required = c(
      "Animal_ID",
      "Species"
    ),
    pattern = list(
      Animal_ID =
        "^A[0-9]+$"
    )
  )

  expect_equal(
    rules$range$Age,
    c(0, 30)
  )

  expect_equal(
    rules$category$Species,
    c(
      "Dog",
      "Cat"
    )
  )

  expect_equal(
    rules$unique,
    "Animal_ID"
  )

  expect_equal(
    rules$required,
    c(
      "Animal_ID",
      "Species"
    )
  )

  expect_equal(
    rules$pattern$Animal_ID,
    "^A[0-9]+$"
  )
})


test_that("empty named lists are accepted", {

  expect_no_error(
    audit_rules(
      range = list(),
      category = list(),
      pattern = list()
    )
  )
})

#==========================================================
# Missing Value Check
#==========================================================

#' Check Missing Values
#'
#' Examines each variable in a dataset and reports the
#' number and percentage of missing and complete values.
#'
#' @param data A data.frame.
#'
#' @return A data.frame of class \code{MissingCheck}
#'   containing variable-level missing-value information.
#'
#' @details
#' For each variable, the function reports the total number
#' of observations, number of missing values, number of
#' complete values, and their corresponding percentages.
#'
#' Missing values are identified using \code{is.na()}.
#'
#' @examples
#' missing_check(iris)
#'
#' dat <- iris
#' dat$Sepal.Length[1:10] <- NA
#' dat$Species[1:5] <- NA
#' missing_check(dat)
#'
#' @export

missing_check <- function(data) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Dataset size
  #---------------------------------------------
  n <- nrow(data)

  #---------------------------------------------
  # Missing counts
  #---------------------------------------------
  missing <- vapply(
    data,
    function(x) sum(is.na(x)),
    integer(1)
  )

  #---------------------------------------------
  # Complete counts
  #---------------------------------------------
  complete <- n - missing

  #---------------------------------------------
  # Percentages
  #---------------------------------------------
  if (n == 0L) {

    missing_percent <- rep(
      NA_real_,
      ncol(data)
    )

    complete_percent <- rep(
      NA_real_,
      ncol(data)
    )

  } else {

    missing_percent <- 100 * missing / n

    complete_percent <- 100 * complete / n

  }

  #---------------------------------------------
  # Result
  #---------------------------------------------
  result <- data.frame(

    Variable = names(data),

    Total = rep(
      n,
      ncol(data)
    ),

    Missing = unname(missing),

    Complete = unname(complete),

    MissingPercent = round(
      unname(missing_percent),
      2
    ),

    CompletePercent = round(
      unname(complete_percent),
      2
    ),

    stringsAsFactors = FALSE,

    check.names = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "MissingCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("missing_check returns MissingCheck object", {

  result <- missing_check(iris)

  expect_s3_class(
    result,
    "MissingCheck"
  )

})

test_that("missing_check returns data.frame", {

  result <- missing_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("missing_check contains expected columns", {

  result <- missing_check(iris)

  expect_equal(
    names(result),
    c(
      "Variable",
      "Total",
      "Missing",
      "Complete",
      "MissingPercent",
      "CompletePercent"
    )
  )

})

test_that("missing_check returns one row per variable", {

  result <- missing_check(iris)

  expect_equal(
    nrow(result),
    ncol(iris)
  )

})


#==========================================================
# Complete dataset
#==========================================================

test_that("missing_check reports zero missing values for iris", {

  result <- missing_check(iris)

  expect_true(
    all(result$Missing == 0L)
  )

})

test_that("missing_check reports correct totals for iris", {

  result <- missing_check(iris)

  expect_true(
    all(result$Total == 150L)
  )

  expect_true(
    all(result$Complete == 150L)
  )

})

test_that("missing_check reports correct percentages for complete data", {

  result <- missing_check(iris)

  expect_true(
    all(result$MissingPercent == 0)
  )

  expect_true(
    all(result$CompletePercent == 100)
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("missing_check counts missing values correctly", {

  dat <- iris

  dat$Sepal.Length[1:10] <- NA
  dat$Species[1:5] <- NA

  result <- missing_check(dat)

  expect_equal(
    result$Missing[
      result$Variable == "Sepal.Length"
    ],
    10L
  )

  expect_equal(
    result$Missing[
      result$Variable == "Species"
    ],
    5L
  )

})

test_that("missing_check counts complete values correctly", {

  dat <- iris

  dat$Sepal.Length[1:10] <- NA

  result <- missing_check(dat)

  expect_equal(
    result$Complete[
      result$Variable == "Sepal.Length"
    ],
    140L
  )

})

test_that("missing_check calculates missing percentage correctly", {

  dat <- data.frame(
    x = c(1, NA, 3, NA)
  )

  result <- missing_check(dat)

  expect_equal(
    result$MissingPercent,
    50
  )

  expect_equal(
    result$CompletePercent,
    50
  )

})

test_that("missing_check rounds percentages to two decimal places", {

  dat <- data.frame(
    x = c(1, 2, NA)
  )

  result <- missing_check(dat)

  expect_equal(
    result$MissingPercent,
    33.33
  )

  expect_equal(
    result$CompletePercent,
    66.67
  )

})


#==========================================================
# All missing
#==========================================================

test_that("missing_check handles completely missing variable", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      5
    )
  )

  result <- missing_check(dat)

  expect_equal(
    result$Missing,
    5L
  )

  expect_equal(
    result$Complete,
    0L
  )

  expect_equal(
    result$MissingPercent,
    100
  )

  expect_equal(
    result$CompletePercent,
    0
  )

})


#==========================================================
# Mixed variable types
#==========================================================

test_that("missing_check handles mixed variable types", {

  dat <- data.frame(
    numeric = c(1, NA, 3, 4),
    character = c("A", "B", NA, "D"),
    factor = factor(
      c("Low", NA, "High", "Low")
    ),
    logical = c(TRUE, FALSE, NA, TRUE),
    stringsAsFactors = FALSE
  )

  result <- missing_check(dat)

  expect_equal(
    result$Missing,
    rep(1L, 4)
  )

  expect_equal(
    result$Complete,
    rep(3L, 4)
  )

  expect_equal(
    result$MissingPercent,
    rep(25, 4)
  )

})


#==========================================================
# Character blanks
#==========================================================

test_that("missing_check does not treat blank strings as missing", {

  dat <- data.frame(
    x = c(
      "A",
      "",
      " ",
      NA
    ),
    stringsAsFactors = FALSE
  )

  result <- missing_check(dat)

  expect_equal(
    result$Missing,
    1L
  )

})


#==========================================================
# Date
#==========================================================

test_that("missing_check handles Date variables", {

  dat <- data.frame(
    date = as.Date(
      c(
        "2026-01-01",
        NA,
        "2026-01-03"
      )
    )
  )

  result <- missing_check(dat)

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Complete,
    2L
  )

})


#==========================================================
# POSIXct
#==========================================================

test_that("missing_check handles POSIXct variables", {

  x <- as.POSIXct(
    c(
      "2026-01-01 10:00:00",
      "2026-01-02 10:00:00",
      "2026-01-03 10:00:00"
    ),
    tz = "UTC"
  )

  x[2] <- NA

  dat <- data.frame(
    time = x
  )

  result <- missing_check(dat)

  expect_equal(
    result$Missing,
    1L
  )

  expect_equal(
    result$Complete,
    2L
  )

})


#==========================================================
# One-column data
#==========================================================

test_that("missing_check handles one-column data", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      3,
      4
    )
  )

  result <- missing_check(dat)

  expect_equal(
    nrow(result),
    1L
  )

  expect_equal(
    result$Variable,
    "x"
  )

  expect_equal(
    result$Total,
    4L
  )

  expect_equal(
    result$Missing,
    1L
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("missing_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- missing_check(dat)

  expect_equal(
    nrow(result),
    2L
  )

  expect_true(
    all(result$Total == 0L)
  )

  expect_true(
    all(result$Missing == 0L)
  )

  expect_true(
    all(result$Complete == 0L)
  )

  expect_true(
    all(is.na(result$MissingPercent))
  )

  expect_true(
    all(is.na(result$CompletePercent))
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("missing_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::as_tibble(iris)

  result <- missing_check(dat)

  expect_s3_class(
    result,
    "MissingCheck"
  )

  expect_equal(
    nrow(result),
    5L
  )

})


#==========================================================
# Column names
#==========================================================

test_that("missing_check preserves variable names", {

  dat <- data.frame(
    First = 1:3,
    Second = 4:6
  )

  result <- missing_check(dat)

  expect_equal(
    result$Variable,
    c(
      "First",
      "Second"
    )
  )

})


#==========================================================
# Invalid input
#==========================================================

test_that("missing_check rejects NULL", {

  expect_error(
    missing_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("missing_check rejects numeric vector", {

  expect_error(
    missing_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("missing_check rejects character vector", {

  expect_error(
    missing_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("missing_check rejects list", {

  expect_error(
    missing_check(
      list(
        x = 1:5,
        y = 6:10
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("missing_check rejects empty data.frame", {

  expect_error(
    missing_check(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Category Check
#==========================================================

#' Check Categorical Values
#'
#' Checks whether values in a variable belong to a
#' user-defined set of allowed categories.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   variable to check.
#' @param allowed A non-empty vector containing the allowed
#'   values.
#'
#' @return A data.frame of class \code{CategoryCheck}
#'   containing row-level category diagnostics.
#'
#' @details
#' Missing values are reported separately and are not
#' classified as invalid categories.
#'
#' The supplied \code{allowed} values are compared with the
#' observed values using their underlying values. Character,
#' factor, numeric, integer, and logical variables are
#' supported.
#'
#' @examples
#' dat <- data.frame(
#'   Species = c("Dog", "Cat", "Horse", NA),
#'   stringsAsFactors = FALSE
#' )
#'
#' category_check(
#'   dat,
#'   variable = "Species",
#'   allowed = c("Dog", "Cat", "Cattle")
#' )
#'
#' @export

category_check <- function(data, variable, allowed) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (missing(variable) ||
      !is.character(variable) ||
      length(variable) != 1L ||
      is.na(variable) ||
      !nzchar(variable)) {

    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {

    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate allowed
  #---------------------------------------------
  if (missing(allowed) ||
      is.null(allowed) ||
      length(allowed) == 0L) {

    stop(
      "'allowed' must contain at least one value.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Reject unsupported allowed objects
  #---------------------------------------------
  if (is.list(allowed) ||
      is.data.frame(allowed) ||
      is.matrix(allowed)) {

    stop(
      "'allowed' must be an atomic vector.",
      call. = FALSE
    )
  }



  #---------------------------------------------
  # NA is not an allowed category
  #---------------------------------------------
  if (anyNA(allowed)) {

    stop(
      "'allowed' must not contain missing values.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Supported variable types
  #---------------------------------------------
  supported <-
    is.character(x) ||
    is.factor(x) ||
    is.numeric(x) ||
    is.logical(x)

  if (!supported) {

    stop(
      "'variable' must be character, factor, numeric, integer, or logical.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Convert factors for comparison
  #---------------------------------------------
  if (is.factor(x)) {

    compare_x <- as.character(x)

  } else {

    compare_x <- x
  }

  if (is.factor(allowed)) {

    compare_allowed <- as.character(allowed)

  } else {

    compare_allowed <- allowed
  }

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_value <- is.na(compare_x)

  #---------------------------------------------
  # Valid categories
  #---------------------------------------------
  valid_category <-
    !missing_value &
    compare_x %in% compare_allowed

  #---------------------------------------------
  # Invalid categories
  #---------------------------------------------
  invalid_category <-
    !missing_value &
    !valid_category

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = x,
    Missing = missing_value,
    ValidCategory = valid_category,
    InvalidCategory = invalid_category,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "CategoryCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("category_check returns CategoryCheck", {

  dat <- data.frame(
    x = c("A", "B")
  )

  result <- category_check(
    dat,
    "x",
    c("A", "B")
  )

  expect_s3_class(
    result,
    "CategoryCheck"
  )

})

test_that("category_check returns data.frame", {

  dat <- data.frame(
    x = c("A", "B")
  )

  result <- category_check(
    dat,
    "x",
    c("A", "B")
  )

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("category_check returns expected columns", {

  dat <- data.frame(
    x = c("A", "B")
  )

  result <- category_check(
    dat,
    "x",
    c("A", "B")
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "ValidCategory",
      "InvalidCategory"
    )
  )

})


#==========================================================
# Character variables
#==========================================================

test_that("category_check identifies valid categories", {

  dat <- data.frame(
    Species = c(
      "Dog",
      "Cat",
      "Cattle"
    ),
    stringsAsFactors = FALSE
  )

  result <- category_check(
    dat,
    "Species",
    c(
      "Dog",
      "Cat",
      "Cattle"
    )
  )

  expect_true(
    all(result$ValidCategory)
  )

  expect_false(
    any(result$InvalidCategory)
  )

})

test_that("category_check identifies invalid categories", {

  dat <- data.frame(
    Species = c(
      "Dog",
      "Horse",
      "Cat",
      "Goat"
    ),
    stringsAsFactors = FALSE
  )

  result <- category_check(
    dat,
    "Species",
    c(
      "Dog",
      "Cat",
      "Cattle"
    )
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      TRUE,
      FALSE,
      TRUE
    )
  )

})

test_that("category_check is case sensitive", {

  dat <- data.frame(
    Species = c(
      "Dog",
      "dog"
    ),
    stringsAsFactors = FALSE
  )

  result <- category_check(
    dat,
    "Species",
    "Dog"
  )

  expect_equal(
    result$ValidCategory,
    c(
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("category_check identifies missing values", {

  dat <- data.frame(
    x = c(
      "A",
      NA,
      "B"
    ),
    stringsAsFactors = FALSE
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "B"
    )
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

})

test_that("missing values are not invalid categories", {

  dat <- data.frame(
    x = c(
      "A",
      NA,
      "C"
    ),
    stringsAsFactors = FALSE
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "B"
    )
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

  expect_equal(
    result$ValidCategory,
    c(
      TRUE,
      FALSE,
      FALSE
    )
  )

})


#==========================================================
# Factor variables
#==========================================================

test_that("category_check handles factors", {

  dat <- data.frame(
    x = factor(
      c(
        "Dog",
        "Cat",
        "Horse"
      )
    )
  )

  result <- category_check(
    dat,
    "x",
    c(
      "Dog",
      "Cat"
    )
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

  expect_true(
    is.factor(result$Value)
  )

})

test_that("category_check handles factor allowed values", {

  dat <- data.frame(
    x = factor(
      c(
        "A",
        "B",
        "C"
      )
    )
  )

  allowed <- factor(
    c(
      "A",
      "B"
    )
  )

  result <- category_check(
    dat,
    "x",
    allowed
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Numeric variables
#==========================================================

test_that("category_check handles numeric categories", {

  dat <- data.frame(
    Score = c(
      0,
      1,
      2,
      3
    )
  )

  result <- category_check(
    dat,
    "Score",
    c(
      0,
      1,
      2
    )
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Integer variables
#==========================================================

test_that("category_check handles integer categories", {

  dat <- data.frame(
    Code = c(
      1L,
      2L,
      3L,
      4L
    )
  )

  result <- category_check(
    dat,
    "Code",
    c(
      1L,
      2L,
      3L
    )
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Logical variables
#==========================================================

test_that("category_check handles logical categories", {

  dat <- data.frame(
    Positive = c(
      TRUE,
      FALSE,
      TRUE
    )
  )

  result <- category_check(
    dat,
    "Positive",
    TRUE
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# All invalid
#==========================================================

test_that("category_check handles all invalid categories", {

  dat <- data.frame(
    x = c(
      "X",
      "Y",
      "Z"
    )
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "B"
    )
  )

  expect_true(
    all(result$InvalidCategory)
  )

  expect_false(
    any(result$ValidCategory)
  )

})


#==========================================================
# All missing
#==========================================================

test_that("category_check handles all missing character values", {

  dat <- data.frame(
    x = rep(
      NA_character_,
      4
    )
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "B"
    )
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$ValidCategory)
  )

  expect_false(
    any(result$InvalidCategory)
  )

})


#==========================================================
# Single allowed category
#==========================================================

test_that("category_check handles one allowed value", {

  dat <- data.frame(
    x = c(
      "Yes",
      "No",
      "Yes"
    )
  )

  result <- category_check(
    dat,
    "x",
    "Yes"
  )

  expect_equal(
    result$ValidCategory,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Duplicate allowed values
#==========================================================

test_that("category_check handles duplicate allowed values", {

  dat <- data.frame(
    x = c(
      "A",
      "B",
      "C"
    )
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "A",
      "B"
    )
  )

  expect_equal(
    result$InvalidCategory,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Zero rows
#==========================================================

test_that("category_check handles zero-row data", {

  dat <- data.frame(
    x = character(0)
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "B"
    )
  )

  expect_s3_class(
    result,
    "CategoryCheck"
  )

  expect_equal(
    nrow(result),
    0L
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("category_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    Species = c(
      "Dog",
      "Cat",
      "Horse"
    )
  )

  result <- category_check(
    dat,
    "Species",
    c(
      "Dog",
      "Cat"
    )
  )

  expect_s3_class(
    result,
    "CategoryCheck"
  )

  expect_equal(
    sum(result$InvalidCategory),
    1L
  )

})


#==========================================================
# Row numbers
#==========================================================

test_that("category_check preserves row numbers", {

  dat <- data.frame(
    x = c(
      "A",
      "B",
      "C"
    )
  )

  result <- category_check(
    dat,
    "x",
    c(
      "A",
      "B"
    )
  )

  expect_equal(
    result$Row,
    1:3
  )

})


#==========================================================
# Variable validation
#==========================================================

test_that("category_check rejects missing variable", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      allowed = c("A", "B")
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("category_check rejects NULL variable", {

  expect_error(
    category_check(
      iris,
      NULL,
      c("A", "B")
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("category_check rejects numeric variable argument", {

  expect_error(
    category_check(
      iris,
      1,
      c("A", "B")
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("category_check rejects multiple variable names", {

  expect_error(
    category_check(
      iris,
      c(
        "Sepal.Length",
        "Sepal.Width"
      ),
      c(1, 2)
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("category_check rejects NA variable", {

  expect_error(
    category_check(
      iris,
      NA_character_,
      c(1, 2)
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("category_check rejects blank variable name", {

  expect_error(
    category_check(
      iris,
      "",
      c(1, 2)
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("category_check rejects unknown variable", {

  expect_error(
    category_check(
      iris,
      "Unknown",
      c(1, 2)
    ),
    "Variable 'Unknown' was not found in 'data'"
  )

})


#==========================================================
# allowed validation
#==========================================================

test_that("category_check rejects missing allowed argument", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      "x"
    ),
    "'allowed' must contain at least one value"
  )

})

test_that("category_check rejects NULL allowed", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      "x",
      NULL
    ),
    "'allowed' must contain at least one value"
  )

})

test_that("category_check rejects empty allowed vector", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      "x",
      character(0)
    ),
    "'allowed' must contain at least one value"
  )

})

test_that("category_check rejects allowed containing NA", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      "x",
      c(
        "A",
        NA_character_
      )
    ),
    "'allowed' must not contain missing values"
  )

})

test_that("category_check rejects list allowed", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      "x",
      list(
        "A",
        "B"
      )
    ),
    "'allowed' must be an atomic vector"
  )

})

test_that("category_check rejects data.frame allowed", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  expect_error(
    category_check(
      dat,
      "x",
      data.frame(
        allowed = c(
          "A",
          "B"
        )
      )
    ),
    "'allowed' must be an atomic vector"
  )

})

test_that("category_check rejects matrix allowed", {

  dat <- data.frame(
    x = c(
      "A",
      "B"
    )
  )

  allowed <- matrix(
    c(
      "A",
      "B"
    ),
    ncol = 1
  )

  expect_error(
    category_check(
      dat,
      "x",
      allowed
    ),
    "'allowed' must be an atomic vector"
  )

})


#==========================================================
# Unsupported variable classes
#==========================================================

test_that("category_check rejects Date variable", {

  dat <- data.frame(
    x = as.Date(
      "2026-01-01"
    ) + 0:2
  )

  expect_error(
    category_check(
      dat,
      "x",
      as.Date("2026-01-01")
    ),
    "'variable' must be character, factor, numeric, integer, or logical"
  )

})

test_that("category_check rejects POSIXct variable", {

  dat <- data.frame(
    x = as.POSIXct(
      "2026-01-01",
      tz = "UTC"
    ) + 0:2
  )

  expect_error(
    category_check(
      dat,
      "x",
      1:3
    ),
    "'variable' must be character, factor, numeric, integer, or logical"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("category_check rejects NULL data", {

  expect_error(
    category_check(
      NULL,
      "x",
      c("A", "B")
    ),
    "'data' cannot be NULL"
  )

})

test_that("category_check rejects numeric vector data", {

  expect_error(
    category_check(
      1:10,
      "x",
      1:10
    ),
    "'data' must be a data.frame"
  )

})

test_that("category_check rejects character vector data", {

  expect_error(
    category_check(
      letters,
      "x",
      letters
    ),
    "'data' must be a data.frame"
  )

})

test_that("category_check rejects list data", {

  expect_error(
    category_check(
      list(
        x = c("A", "B")
      ),
      "x",
      c("A", "B")
    ),
    "'data' must be a data.frame"
  )

})

test_that("category_check rejects empty data.frame", {

  expect_error(
    category_check(
      data.frame(),
      "x",
      c("A", "B")
    ),
    "at least one variable"
  )

})

#==========================================================
# Print DataAudit Rule Result
#==========================================================

#' Print a DataAudit Rule Result
#'
#' Prints a concise summary of custom audit-rule results
#' produced by \code{\link{apply_audit_rules}}.
#'
#' @param x An object of class \code{DataAuditRuleResult}.
#' @param ... Additional arguments, currently unused.
#'
#' @return The input object \code{x}, invisibly.
#'
#' @export
print.DataAuditRuleResult <- function(
    x,
    ...
) {

  #---------------------------------------------
  # Validate object
  #---------------------------------------------
  required_components <- c(
    "status",
    "total_violations",
    "summary",
    "range",
    "category",
    "unique",
    "required",
    "pattern"
  )

  if (
    !inherits(x, "DataAuditRuleResult") ||
    !all(required_components %in% names(x))
  ) {

    stop(
      "Invalid 'DataAuditRuleResult' object.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate summary
  #---------------------------------------------
  if (!is.data.frame(x$summary)) {

    stop(
      "Invalid rule summary in 'DataAuditRuleResult'.",
      call. = FALSE
    )
  }

  required_summary <- c(
    "Rule",
    "Violations"
  )

  if (
    !all(
      required_summary %in%
      names(x$summary)
    )
  ) {

    stop(
      "Invalid rule summary in 'DataAuditRuleResult'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Header
  #---------------------------------------------
  cat(
    "DataAudit Rule Validation\n"
  )

  cat(
    "=========================\n\n"
  )

  #---------------------------------------------
  # Overall result
  #---------------------------------------------
  cat(
    sprintf(
      "Status:           %s\n",
      x$status
    )
  )

  cat(
    sprintf(
      "Total violations: %d\n",
      x$total_violations
    )
  )

  #---------------------------------------------
  # Rule summary
  #---------------------------------------------
  cat(
    "\nRule summary\n"
  )

  cat(
    "------------\n"
  )

  print(
    x$summary,
    row.names = FALSE
  )

  #---------------------------------------------
  # Variables with violations
  #---------------------------------------------
  result_groups <- list(
    Range = x$range,
    Category = x$category,
    Unique = x$unique,
    Required = x$required,
    Pattern = x$pattern
  )

  violation_variables <- character()

  for (rule_name in names(result_groups)) {

    group <- result_groups[[rule_name]]

    if (length(group) == 0L) {
      next
    }

    for (variable in names(group)) {

      z <- group[[variable]]

      if (
        is.data.frame(z) &&
        "Violation" %in% names(z) &&
        any(z$Violation)
      ) {

        violation_variables <- c(
          violation_variables,
          sprintf(
            "%s: %s",
            rule_name,
            variable
          )
        )
      }
    }
  }

  #---------------------------------------------
  # Print affected variables
  #---------------------------------------------
  if (length(violation_variables) > 0L) {

    cat(
      "\nVariables with rule violations\n"
    )

    cat(
      "------------------------------\n"
    )

    for (item in violation_variables) {

      cat(
        sprintf(
          "- %s\n",
          item
        )
      )
    }

  } else {

    cat(
      "\nNo rule violations detected.\n"
    )
  }

  invisible(x)
}

#==========================================================
# Case Consistency Check
#==========================================================

#' Check Case Consistency
#'
#' Detects values in a character or factor variable that
#' differ only in letter case.
#'
#' @param data A data.frame.
#' @param variable A single character string specifying the
#'   variable to check.
#'
#' @return A data.frame of class \code{CaseCheck} containing
#'   row-level case-consistency diagnostics.
#'
#' @details
#' Values are compared after conversion to lower case.
#' Therefore, values such as \code{"Dog"}, \code{"dog"},
#' and \code{"DOG"} are treated as belonging to the same
#' normalized category.
#'
#' A value is flagged when more than one distinct original
#' representation exists for the same normalized value.
#'
#' Missing values are reported separately and are not
#' classified as case inconsistencies.
#'
#' The function does not modify the original data.
#'
#' @examples
#' dat <- data.frame(
#'   Species = c("Dog", "dog", "DOG", "Cat", "Cat", NA)
#' )
#'
#' case_check(
#'   dat,
#'   variable = "Species"
#' )
#'
#' @export

case_check <- function(data, variable) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (
    missing(variable) ||
    !is.character(variable) ||
    length(variable) != 1L ||
    is.na(variable) ||
    !nzchar(variable)
  ) {
    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Check variable exists
  #---------------------------------------------
  if (!variable %in% names(data)) {
    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Extract variable
  #---------------------------------------------
  x <- data[[variable]]

  #---------------------------------------------
  # Validate variable type
  #---------------------------------------------
  if (!is.character(x) && !is.factor(x)) {
    stop(
      "'variable' must refer to a character or factor variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Convert factor to character
  #---------------------------------------------
  value <- as.character(x)

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_value <- is.na(value)

  #---------------------------------------------
  # Normalize case
  #---------------------------------------------
  normalized <- rep(
    NA_character_,
    length(value)
  )

  normalized[!missing_value] <-
    tolower(value[!missing_value])

  #---------------------------------------------
  # Determine inconsistent normalized groups
  #---------------------------------------------
  case_issue <- rep(
    FALSE,
    length(value)
  )

  valid_index <- which(!missing_value)

  if (length(valid_index) > 0L) {

    normalized_groups <- split(
      value[valid_index],
      normalized[valid_index]
    )

    inconsistent_groups <- names(
      normalized_groups
    )[
      vapply(
        normalized_groups,
        function(z) {
          length(unique(z)) > 1L
        },
        logical(1)
      )
    ]

    case_issue[valid_index] <-
      normalized[valid_index] %in%
      inconsistent_groups
  }

  #---------------------------------------------
  # Overall validity
  #---------------------------------------------
  valid_case <-
    !missing_value &
    !case_issue

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = value,
    Normalized = normalized,
    Missing = missing_value,
    CaseIssue = case_issue,
    ValidCase = valid_case,
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Assign class
  #---------------------------------------------
  class(result) <- c(
    "CaseCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("case_check returns CaseCheck", {

  dat <- data.frame(
    x = c("Dog", "Cat")
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_s3_class(
    result,
    "CaseCheck"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("case_check returns expected columns", {

  dat <- data.frame(
    x = c("Dog", "Cat")
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Normalized",
      "Missing",
      "CaseIssue",
      "ValidCase"
    )
  )
})

#==========================================================
# Normalization
#==========================================================

test_that("case_check normalizes values to lower case", {

  dat <- data.frame(
    x = c(
      "Dog",
      "DOG",
      "Cat"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    result$Normalized,
    c(
      "dog",
      "dog",
      "cat"
    )
  )
})

#==========================================================
# Case inconsistency
#==========================================================

test_that("case_check detects inconsistent case", {

  dat <- data.frame(
    x = c(
      "Dog",
      "dog",
      "DOG",
      "Cat"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    result$CaseIssue,
    c(
      TRUE,
      TRUE,
      TRUE,
      FALSE
    )
  )
})

test_that("all representations in inconsistent group are flagged", {

  dat <- data.frame(
    x = c(
      "Cattle",
      "cattle",
      "CATTLE",
      "Goat",
      "Goat"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    result$CaseIssue,
    c(
      TRUE,
      TRUE,
      TRUE,
      FALSE,
      FALSE
    )
  )
})

test_that("consistent repeated categories are not flagged", {

  dat <- data.frame(
    x = c(
      "Dog",
      "Dog",
      "Cat",
      "Cat"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_false(
    any(result$CaseIssue)
  )

  expect_true(
    all(result$ValidCase)
  )
})

test_that("unique categories are not flagged", {

  dat <- data.frame(
    x = c(
      "Dog",
      "Cat",
      "Cattle"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_false(
    any(result$CaseIssue)
  )
})

#==========================================================
# Missing values
#==========================================================

test_that("case_check handles missing values", {

  dat <- data.frame(
    x = c(
      "Dog",
      NA,
      "dog"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    result$Missing,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

  expect_true(
    is.na(result$Normalized[2])
  )

  expect_false(
    result$CaseIssue[2]
  )

  expect_false(
    result$ValidCase[2]
  )
})

test_that("case_check handles all missing values", {

  dat <- data.frame(
    x = c(
      NA_character_,
      NA_character_,
      NA_character_
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_true(
    all(result$Missing)
  )

  expect_false(
    any(result$CaseIssue)
  )

  expect_false(
    any(result$ValidCase)
  )

  expect_true(
    all(is.na(result$Normalized))
  )
})

#==========================================================
# Factor variables
#==========================================================

test_that("case_check works with factors", {

  dat <- data.frame(
    x = factor(
      c(
        "Dog",
        "dog",
        "Cat"
      )
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    result$CaseIssue,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )
})

#==========================================================
# Empty strings
#==========================================================

test_that("case_check handles empty strings", {

  dat <- data.frame(
    x = c(
      "",
      "",
      "Dog"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_false(
    any(result$CaseIssue)
  )

  expect_equal(
    result$Normalized,
    c(
      "",
      "",
      "dog"
    )
  )
})

#==========================================================
# Internal spaces
#==========================================================

test_that("case_check preserves internal spaces", {

  dat <- data.frame(
    x = c(
      "Dairy Cattle",
      "dairy cattle",
      "Goat"
    )
  )

  result <- case_check(
    dat,
    "x"
  )

  expect_equal(
    result$CaseIssue,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )
})

#==========================================================
# Invalid variable
#==========================================================

test_that("case_check rejects missing variable", {

  expect_error(
    case_check(iris),
    "'variable' must be a single non-empty character string"
  )
})

test_that("case_check rejects NULL variable", {

  expect_error(
    case_check(
      iris,
      NULL
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("case_check rejects empty variable", {

  expect_error(
    case_check(
      iris,
      ""
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("case_check rejects numeric variable argument", {

  expect_error(
    case_check(
      iris,
      1
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("case_check rejects multiple variable names", {

  expect_error(
    case_check(
      iris,
      c(
        "Species",
        "Sepal.Length"
      )
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("case_check rejects NA variable", {

  expect_error(
    case_check(
      iris,
      NA_character_
    ),
    "'variable' must be a single non-empty character string"
  )
})

#==========================================================
# Unknown variable
#==========================================================

test_that("case_check rejects unknown variable", {

  expect_error(
    case_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

#==========================================================
# Invalid variable type
#==========================================================

test_that("case_check rejects numeric column", {

  expect_error(
    case_check(
      iris,
      "Sepal.Length"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

test_that("case_check rejects logical column", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE
    )
  )

  expect_error(
    case_check(
      dat,
      "x"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

test_that("case_check rejects Date column", {

  dat <- data.frame(
    x = as.Date(
      c(
        "2026-01-01",
        "2026-01-02"
      )
    )
  )

  expect_error(
    case_check(
      dat,
      "x"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("case_check rejects invalid data", {

  expect_error(
    case_check(
      NULL,
      "x"
    )
  )

  expect_error(
    case_check(
      1:5,
      "x"
    )
  )

  expect_error(
    case_check(
      list(x = "Dog"),
      "x"
    )
  )

  expect_error(
    case_check(
      data.frame(),
      "x"
    )
  )
})

#==========================================================
# Range Check
#==========================================================

#' Check Values Against a Valid Range
#'
#' Identifies values in a numeric variable that fall below
#' or above user-defined valid limits.
#'
#' @param data A data.frame.
#' @param variable Character string giving the name of the
#'   numeric variable to check.
#' @param min Optional numeric lower limit. Default is
#'   \code{NULL}.
#' @param max Optional numeric upper limit. Default is
#'   \code{NULL}.
#'
#' @return A data.frame of class \code{RangeCheck}
#'   containing row-level range diagnostics.
#'
#' @details
#' At least one of \code{min} or \code{max} must be
#' supplied.
#'
#' Missing values are not classified as out-of-range.
#' Infinite values are evaluated against the specified
#' limits and will normally be flagged when an appropriate
#' finite limit is supplied.
#'
#' @examples
#' dat <- data.frame(
#'   Temperature = c(38.5, 39.2, 44.0, 34.0, NA)
#' )
#'
#' range_check(
#'   dat,
#'   variable = "Temperature",
#'   min = 35,
#'   max = 43
#' )
#'
#' @export

range_check <- function(
    data,
    variable,
    min = NULL,
    max = NULL
) {

  #---------------------------------------------
  # Validate data
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Validate variable
  #---------------------------------------------
  if (missing(variable) ||
      !is.character(variable) ||
      length(variable) != 1L ||
      is.na(variable) ||
      !nzchar(variable)) {

    stop(
      "'variable' must be a single non-empty character string.",
      call. = FALSE
    )
  }

  if (!variable %in% names(data)) {

    stop(
      sprintf(
        "Variable '%s' was not found in 'data'.",
        variable
      ),
      call. = FALSE
    )
  }

  x <- data[[variable]]

  if (!is.numeric(x)) {

    stop(
      "'variable' must refer to a numeric variable.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate limits
  #---------------------------------------------
  if (is.null(min) && is.null(max)) {

    stop(
      "At least one of 'min' or 'max' must be supplied.",
      call. = FALSE
    )
  }

  if (!is.null(min)) {

    if (!is.numeric(min) ||
        length(min) != 1L ||
        is.na(min) ||
        !is.finite(min)) {

      stop(
        "'min' must be NULL or a single finite numeric value.",
        call. = FALSE
      )
    }
  }

  if (!is.null(max)) {

    if (!is.numeric(max) ||
        length(max) != 1L ||
        is.na(max) ||
        !is.finite(max)) {

      stop(
        "'max' must be NULL or a single finite numeric value.",
        call. = FALSE
      )
    }
  }

  if (!is.null(min) &&
      !is.null(max) &&
      min > max) {

    stop(
      "'min' cannot be greater than 'max'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Detect violations
  #---------------------------------------------
  below <- rep(
    FALSE,
    length(x)
  )

  above <- rep(
    FALSE,
    length(x)
  )

  if (!is.null(min)) {

    below <-
      !is.na(x) &
      x < min
  }

  if (!is.null(max)) {

    above <-
      !is.na(x) &
      x > max
  }

  out_of_range <- below | above

  #---------------------------------------------
  # Construct output
  #---------------------------------------------
  result <- data.frame(
    Row = seq_along(x),
    Value = x,
    BelowMinimum = below,
    AboveMaximum = above,
    OutOfRange = out_of_range,
    stringsAsFactors = FALSE
  )

  class(result) <- c(
    "RangeCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("range_check returns RangeCheck", {

  dat <- data.frame(x = 1:10)

  result <- range_check(
    dat,
    "x",
    min = 1,
    max = 10
  )

  expect_s3_class(
    result,
    "RangeCheck"
  )

})

test_that("range_check returns data.frame", {

  dat <- data.frame(x = 1:10)

  result <- range_check(
    dat,
    "x",
    min = 1,
    max = 10
  )

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("range_check contains expected columns", {

  dat <- data.frame(x = 1:5)

  result <- range_check(
    dat,
    "x",
    min = 1,
    max = 5
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "BelowMinimum",
      "AboveMaximum",
      "OutOfRange"
    )
  )

})


#==========================================================
# Valid values
#==========================================================

test_that("range_check identifies valid values", {

  dat <- data.frame(
    x = 1:5
  )

  result <- range_check(
    dat,
    "x",
    min = 1,
    max = 5
  )

  expect_false(
    any(result$OutOfRange)
  )

})


#==========================================================
# Lower violation
#==========================================================

test_that("range_check detects values below minimum", {

  dat <- data.frame(
    x = c(
      -1,
      1,
      2,
      3
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 5
  )

  expect_equal(
    result$BelowMinimum,
    c(
      TRUE,
      FALSE,
      FALSE,
      FALSE
    )
  )

  expect_equal(
    result$AboveMaximum,
    rep(FALSE, 4)
  )

  expect_equal(
    result$OutOfRange,
    c(
      TRUE,
      FALSE,
      FALSE,
      FALSE
    )
  )

})


#==========================================================
# Upper violation
#==========================================================

test_that("range_check detects values above maximum", {

  dat <- data.frame(
    x = c(
      1,
      2,
      10
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 5
  )

  expect_equal(
    result$AboveMaximum,
    c(
      FALSE,
      FALSE,
      TRUE
    )
  )

  expect_true(
    result$OutOfRange[3]
  )

})


#==========================================================
# Both limits
#==========================================================

test_that("range_check detects lower and upper violations", {

  dat <- data.frame(
    x = c(
      -5,
      5,
      15
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 10
  )

  expect_equal(
    result$OutOfRange,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Boundary values
#==========================================================

test_that("range_check treats limits as valid", {

  dat <- data.frame(
    x = c(
      0,
      5,
      10
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 10
  )

  expect_false(
    any(result$OutOfRange)
  )

})


#==========================================================
# Minimum only
#==========================================================

test_that("range_check works with minimum only", {

  dat <- data.frame(
    x = c(
      -1,
      0,
      10,
      100
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0
  )

  expect_equal(
    result$BelowMinimum,
    c(
      TRUE,
      FALSE,
      FALSE,
      FALSE
    )
  )

  expect_false(
    any(result$AboveMaximum)
  )

})


#==========================================================
# Maximum only
#==========================================================

test_that("range_check works with maximum only", {

  dat <- data.frame(
    x = c(
      -100,
      0,
      10,
      11
    )
  )

  result <- range_check(
    dat,
    "x",
    max = 10
  )

  expect_false(
    any(result$BelowMinimum)
  )

  expect_equal(
    result$AboveMaximum,
    c(
      FALSE,
      FALSE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("range_check does not flag missing values", {

  dat <- data.frame(
    x = c(
      1,
      NA,
      20
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 10
  )

  expect_false(
    result$OutOfRange[2]
  )

  expect_true(
    is.na(result$Value[2])
  )

})


#==========================================================
# Infinite values
#==========================================================

test_that("range_check handles infinite values", {

  dat <- data.frame(
    x = c(
      -Inf,
      5,
      Inf
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 10
  )

  expect_equal(
    result$OutOfRange,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Row numbers
#==========================================================

test_that("range_check returns row numbers", {

  dat <- data.frame(
    x = c(10, 20, 30)
  )

  result <- range_check(
    dat,
    "x",
    min = 0
  )

  expect_equal(
    result$Row,
    1:3
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("range_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0)
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 10
  )

  expect_s3_class(
    result,
    "RangeCheck"
  )

  expect_equal(
    nrow(result),
    0L
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("range_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = c(
      -1,
      5,
      20
    )
  )

  result <- range_check(
    dat,
    "x",
    min = 0,
    max = 10
  )

  expect_equal(
    result$OutOfRange,
    c(
      TRUE,
      FALSE,
      TRUE
    )
  )

})


#==========================================================
# Variable validation
#==========================================================

test_that("range_check rejects missing variable argument", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      min = 0
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("range_check rejects NULL variable", {

  expect_error(
    range_check(
      iris,
      variable = NULL,
      min = 0
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("range_check rejects numeric variable argument", {

  expect_error(
    range_check(
      iris,
      variable = 1,
      min = 0
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("range_check rejects vector variable argument", {

  expect_error(
    range_check(
      iris,
      variable = c(
        "Sepal.Length",
        "Sepal.Width"
      ),
      min = 0
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("range_check rejects NA variable", {

  expect_error(
    range_check(
      iris,
      variable = NA_character_,
      min = 0
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("range_check rejects empty variable name", {

  expect_error(
    range_check(
      iris,
      variable = "",
      min = 0
    ),
    "'variable' must be a single non-empty character string"
  )

})

test_that("range_check rejects unknown variable", {

  expect_error(
    range_check(
      iris,
      variable = "Unknown",
      min = 0
    ),
    "Variable 'Unknown' was not found"
  )

})

test_that("range_check rejects non-numeric variable", {

  expect_error(
    range_check(
      iris,
      variable = "Species",
      min = 0
    ),
    "'variable' must refer to a numeric variable"
  )

})


#==========================================================
# Limit validation
#==========================================================

test_that("range_check requires at least one limit", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x"
    ),
    "At least one of 'min' or 'max' must be supplied"
  )

})

test_that("range_check rejects invalid minimum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      min = "0"
    ),
    "'min' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects vector minimum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      min = c(0, 1)
    ),
    "'min' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects NA minimum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      min = NA_real_
    ),
    "'min' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects infinite minimum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      min = Inf
    ),
    "'min' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects invalid maximum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      max = "10"
    ),
    "'max' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects vector maximum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      max = c(5, 10)
    ),
    "'max' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects NA maximum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      max = NA_real_
    ),
    "'max' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects infinite maximum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      max = Inf
    ),
    "'max' must be NULL or a single finite numeric value"
  )

})

test_that("range_check rejects minimum above maximum", {

  dat <- data.frame(
    x = 1:5
  )

  expect_error(
    range_check(
      dat,
      "x",
      min = 10,
      max = 5
    ),
    "'min' cannot be greater than 'max'"
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("range_check rejects NULL data", {

  expect_error(
    range_check(
      NULL,
      "x",
      min = 0
    ),
    "'data' cannot be NULL"
  )

})

test_that("range_check rejects numeric vector", {

  expect_error(
    range_check(
      1:10,
      "x",
      min = 0
    ),
    "'data' must be a data.frame"
  )

})

test_that("range_check rejects list", {

  expect_error(
    range_check(
      list(x = 1:5),
      "x",
      min = 0
    ),
    "'data' must be a data.frame"
  )

})

test_that("range_check rejects empty data.frame", {

  expect_error(
    range_check(
      data.frame(),
      "x",
      min = 0
    ),
    "at least one variable"
  )

})

#==========================================================
# Summary DataAudit Report
#==========================================================

#' Summarize a DataAudit Report
#'
#' Produces a compact summary of a data audit generated by
#' \code{\link{audit_data}}.
#'
#' @param object An object of class \code{DataAuditReport}.
#' @param ... Additional arguments, currently unused.
#'
#' @return An object of class \code{summary.DataAuditReport}
#'   containing dataset-level, variable-level, and optional
#'   custom rule summaries.
#'
#' @export
summary.DataAuditReport <- function(
    object,
    ...
) {

  #---------------------------------------------
  # Validate object
  #---------------------------------------------
  required_components <- c(
    "overview",
    "variables",
    "duplicates",
    "summary",
    "rules"
  )

  if (
    !inherits(object, "DataAuditReport") ||
    !all(required_components %in% names(object))
  ) {

    stop(
      "Invalid 'DataAuditReport' object.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate dataset summary
  #---------------------------------------------
  if (!is.data.frame(object$summary)) {

    stop(
      "Invalid summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  required_summary <- c(
    "Rows",
    "Variables",
    "VariablesWithIssues",
    "DuplicateRows",
    "TotalIssues",
    "RuleViolations",
    "OverallIssues"
  )

  if (
    !all(
      required_summary %in%
      names(object$summary)
    )
  ) {

    stop(
      "Invalid summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate variable summary
  #---------------------------------------------
  if (
    !is.data.frame(object$variables) ||
    !"Issue" %in% names(object$variables)
  ) {

    stop(
      "Invalid variable summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Validate duplicate summary
  #---------------------------------------------
  if (
    !is.data.frame(object$duplicates) ||
    !"Duplicate" %in% names(object$duplicates)
  ) {

    stop(
      "Invalid duplicate summary in 'DataAuditReport'.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Variables with issues
  #---------------------------------------------
  variables_with_issues <-
    object$variables[
      object$variables$Issue,
      ,
      drop = FALSE
    ]

  #---------------------------------------------
  # Duplicated rows
  #---------------------------------------------
  duplicate_rows <-
    object$duplicates[
      object$duplicates$Duplicate,
      ,
      drop = FALSE
    ]

  #---------------------------------------------
  # Custom rule summary
  #---------------------------------------------
  rule_summary <- NULL

  if (!is.null(object$rules)) {

    if (
      !inherits(
        object$rules,
        "DataAuditRuleResult"
      )
    ) {

      stop(
        "Invalid rule results in 'DataAuditReport'.",
        call. = FALSE
      )
    }

    if (!is.data.frame(object$rules$summary)) {

      stop(
        "Invalid rule summary in 'DataAuditReport'.",
        call. = FALSE
      )
    }

    required_rule_summary <- c(
      "Rule",
      "Violations"
    )

    if (
      !all(
        required_rule_summary %in%
        names(object$rules$summary)
      )
    ) {

      stop(
        "Invalid rule summary in 'DataAuditReport'.",
        call. = FALSE
      )
    }

    rule_summary <-
      object$rules$summary
  }

  #---------------------------------------------
  # Construct summary
  #---------------------------------------------
  result <- list(
    dataset = object$summary,
    variables_with_issues =
      variables_with_issues,
    duplicate_rows =
      duplicate_rows,
    rules = rule_summary
  )

  class(result) <- c(
    "summary.DataAuditReport",
    "list"
  )

  result
}

#==========================================================
# Constant Variable Check
#==========================================================

#' Check Constant Variables
#'
#' Identifies variables that contain no meaningful
#' variation.
#'
#' @param data A data.frame.
#'
#' @return A data.frame of class \code{ConstantCheck}
#'   containing the number of unique non-missing values,
#'   missing values, and constant-variable status for each
#'   variable.
#'
#' @details
#' A variable is classified as constant when it contains
#' one or fewer unique non-missing values.
#'
#' Missing values are excluded when counting unique values.
#' Consequently, a variable containing only missing values
#' is also classified as constant.
#'
#' @examples
#' dat <- data.frame(
#'   id = 1:5,
#'   group = rep("A", 5),
#'   value = c(10, 10, 10, NA, 10)
#' )
#'
#' constant_check(dat)
#'
#' @export

constant_check <- function(data) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  #---------------------------------------------
  # Number of unique non-missing values
  #---------------------------------------------
  unique_values <- vapply(
    data,
    function(x) {

      length(
        unique(
          x[!is.na(x)]
        )
      )

    },
    integer(1)
  )

  #---------------------------------------------
  # Missing values
  #---------------------------------------------
  missing_values <- vapply(
    data,
    function(x) {
      sum(is.na(x))
    },
    integer(1)
  )

  #---------------------------------------------
  # Constant status
  #---------------------------------------------
  constant <- unique_values <= 1L

  #---------------------------------------------
  # Result
  #---------------------------------------------
  result <- data.frame(
    Variable = names(data),
    UniqueValues = unname(unique_values),
    Missing = unname(missing_values),
    Constant = unname(constant),
    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "ConstantCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("constant_check returns ConstantCheck object", {

  result <- constant_check(iris)

  expect_s3_class(
    result,
    "ConstantCheck"
  )

})

test_that("constant_check returns data.frame", {

  result <- constant_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("constant_check contains expected columns", {

  result <- constant_check(iris)

  expect_equal(
    names(result),
    c(
      "Variable",
      "UniqueValues",
      "Missing",
      "Constant"
    )
  )

})

test_that("constant_check returns one row per variable", {

  result <- constant_check(iris)

  expect_equal(
    nrow(result),
    ncol(iris)
  )

})


#==========================================================
# Non-constant variables
#==========================================================

test_that("constant_check identifies non-constant variables", {

  dat <- data.frame(
    x = 1:5,
    y = letters[1:5]
  )

  result <- constant_check(dat)

  expect_false(
    any(result$Constant)
  )

})

test_that("constant_check counts unique values correctly", {

  dat <- data.frame(
    x = c(1, 2, 2, 3, 3)
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    3L
  )

  expect_false(
    result$Constant
  )

})


#==========================================================
# Numeric constants
#==========================================================

test_that("constant_check detects numeric constant", {

  dat <- data.frame(
    x = rep(10, 5)
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    result$Constant
  )

})


#==========================================================
# Character constants
#==========================================================

test_that("constant_check detects character constant", {

  dat <- data.frame(
    x = rep("Dog", 5),
    stringsAsFactors = FALSE
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    result$Constant
  )

})


#==========================================================
# Factor constants
#==========================================================

test_that("constant_check detects factor constant", {

  dat <- data.frame(
    x = factor(
      rep("A", 5)
    )
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    result$Constant
  )

})


#==========================================================
# Logical constants
#==========================================================

test_that("constant_check detects logical constant", {

  dat <- data.frame(
    x = rep(TRUE, 5)
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    result$Constant
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("constant_check ignores missing values when counting unique values", {

  dat <- data.frame(
    x = c(
      5,
      5,
      NA,
      5,
      NA
    )
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_equal(
    result$Missing,
    2L
  )

  expect_true(
    result$Constant
  )

})

test_that("constant_check does not mark varying data with NA as constant", {

  dat <- data.frame(
    x = c(
      1,
      2,
      NA,
      2
    )
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_equal(
    result$Missing,
    1L
  )

  expect_false(
    result$Constant
  )

})


#==========================================================
# All missing
#==========================================================

test_that("constant_check identifies all-NA variable as constant", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      5
    )
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    0L
  )

  expect_equal(
    result$Missing,
    5L
  )

  expect_true(
    result$Constant
  )

})


#==========================================================
# Mixed variables
#==========================================================

test_that("constant_check handles mixed variable types", {

  dat <- data.frame(
    id = 1:5,
    species = rep("Dog", 5),
    farm = factor(
      rep("A", 5)
    ),
    positive = rep(TRUE, 5),
    stringsAsFactors = FALSE
  )

  result <- constant_check(dat)

  expect_equal(
    result$Variable,
    c(
      "id",
      "species",
      "farm",
      "positive"
    )
  )

  expect_equal(
    result$UniqueValues,
    c(
      5L,
      1L,
      1L,
      1L
    )
  )

  expect_equal(
    result$Constant,
    c(
      FALSE,
      TRUE,
      TRUE,
      TRUE
    )
  )

})


#==========================================================
# Date variables
#==========================================================

test_that("constant_check detects constant Date variable", {

  dat <- data.frame(
    date = rep(
      as.Date("2026-01-01"),
      5
    )
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    result$Constant
  )

})

test_that("constant_check identifies varying Date variable", {

  dat <- data.frame(
    date = as.Date("2026-01-01") + 0:4
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    5L
  )

  expect_false(
    result$Constant
  )

})


#==========================================================
# POSIXct variables
#==========================================================

test_that("constant_check detects constant POSIXct variable", {

  dat <- data.frame(
    time = rep(
      as.POSIXct(
        "2026-01-01 10:00:00",
        tz = "UTC"
      ),
      5
    )
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_true(
    result$Constant
  )

})

test_that("constant_check identifies varying POSIXct variable", {

  dat <- data.frame(
    time = as.POSIXct(
      "2026-01-01 10:00:00",
      tz = "UTC"
    ) + 0:4
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    5L
  )

  expect_false(
    result$Constant
  )

})


#==========================================================
# Blank strings
#==========================================================

test_that("constant_check treats blank strings as values", {

  dat <- data.frame(
    x = c(
      "",
      "",
      "",
      ""
    ),
    stringsAsFactors = FALSE
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    1L
  )

  expect_equal(
    result$Missing,
    0L
  )

  expect_true(
    result$Constant
  )

})

test_that("constant_check distinguishes blank and non-blank values", {

  dat <- data.frame(
    x = c(
      "",
      "A",
      "",
      "A"
    ),
    stringsAsFactors = FALSE
  )

  result <- constant_check(dat)

  expect_equal(
    result$UniqueValues,
    2L
  )

  expect_false(
    result$Constant
  )

})


#==========================================================
# One-row data
#==========================================================

test_that("constant_check identifies variables in one-row data as constant", {

  dat <- data.frame(
    x = 1,
    y = "A",
    stringsAsFactors = FALSE
  )

  result <- constant_check(dat)

  expect_true(
    all(result$Constant)
  )

  expect_equal(
    result$UniqueValues,
    c(1L, 1L)
  )

})


#==========================================================
# One-column data
#==========================================================

test_that("constant_check works with one-column data", {

  dat <- data.frame(
    x = c(
      1,
      1,
      1
    )
  )

  result <- constant_check(dat)

  expect_equal(
    nrow(result),
    1L
  )

  expect_equal(
    result$Variable,
    "x"
  )

  expect_true(
    result$Constant
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("constant_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- constant_check(dat)

  expect_s3_class(
    result,
    "ConstantCheck"
  )

  expect_equal(
    nrow(result),
    2L
  )

  expect_equal(
    result$UniqueValues,
    c(0L, 0L)
  )

  expect_equal(
    result$Missing,
    c(0L, 0L)
  )

  expect_true(
    all(result$Constant)
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("constant_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = c(1, 1, 1),
    y = c("A", "B", "C")
  )

  result <- constant_check(dat)

  expect_s3_class(
    result,
    "ConstantCheck"
  )

  expect_equal(
    result$Constant,
    c(TRUE, FALSE)
  )

})


#==========================================================
# Variable names
#==========================================================

test_that("constant_check preserves variable names", {

  dat <- data.frame(
    Animal_ID = 1:3,
    Species = rep("Dog", 3)
  )

  result <- constant_check(dat)

  expect_equal(
    result$Variable,
    c(
      "Animal_ID",
      "Species"
    )
  )

})


#==========================================================
# Invalid input
#==========================================================

test_that("constant_check rejects NULL", {

  expect_error(
    constant_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("constant_check rejects numeric vector", {

  expect_error(
    constant_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("constant_check rejects character vector", {

  expect_error(
    constant_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("constant_check rejects list", {

  expect_error(
    constant_check(
      list(
        x = 1:5,
        y = 6:10
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("constant_check rejects data.frame without variables", {

  expect_error(
    constant_check(
      data.frame()
    ),
    "at least one variable"
  )

})

#==========================================================
# Outlier Check
#==========================================================

#' Check Numeric Outliers
#'
#' Identifies potential outliers in numeric variables using
#' the interquartile range (IQR) method.
#'
#' @param data A data.frame.
#' @param k A single positive numeric value controlling the
#'   IQR multiplier. Default is 1.5.
#'
#' @return A data.frame of class \code{OutlierCheck}
#'   containing variable-level outlier diagnostics.
#'
#' @details
#' For each numeric variable, the lower and upper limits are:
#'
#' \deqn{Q1 - k * IQR}
#'
#' and
#'
#' \deqn{Q3 + k * IQR}
#'
#' Values strictly below the lower limit or strictly above
#' the upper limit are classified as potential outliers.
#'
#' Missing and infinite values are excluded from the
#' calculation of quartiles and outliers. They are reported
#' separately.
#'
#' Non-numeric variables are retained in the output but are
#' not evaluated for outliers.
#'
#' @examples
#' dat <- data.frame(
#'   x = c(10, 11, 12, 13, 100),
#'   group = c("A", "A", "B", "B", "B")
#' )
#'
#' outlier_check(dat)
#'
#' outlier_check(dat, k = 3)
#'
#' @export

outlier_check <- function(data, k = 1.5) {

  #---------------------------------------------
  # Validation
  #---------------------------------------------
  validate_data(data)

  if (!is.numeric(k) ||
      length(k) != 1L ||
      is.na(k) ||
      !is.finite(k)) {

    stop(
      "'k' must be a single finite numeric value.",
      call. = FALSE
    )
  }

  if (k <= 0) {

    stop(
      "'k' must be greater than zero.",
      call. = FALSE
    )
  }

  #---------------------------------------------
  # Analyze one variable
  #---------------------------------------------
  analyse_variable <- function(x) {

    numeric_variable <- is.numeric(x)

    missing <- sum(is.na(x))

    if (!numeric_variable) {

      return(
        list(
          Numeric = FALSE,
          Missing = missing,
          Infinite = 0L,
          Observed = NA_integer_,
          Q1 = NA_real_,
          Q3 = NA_real_,
          IQR = NA_real_,
          LowerBound = NA_real_,
          UpperBound = NA_real_,
          LowerOutliers = NA_integer_,
          UpperOutliers = NA_integer_,
          Outliers = NA_integer_,
          OutlierPercent = NA_real_
        )
      )
    }

    #-------------------------------------------
    # Infinite values
    #-------------------------------------------
    infinite <- sum(
      is.infinite(x),
      na.rm = TRUE
    )

    #-------------------------------------------
    # Keep only finite observed values
    #-------------------------------------------
    observed <- x[
      !is.na(x) &
        is.finite(x)
    ]

    n_observed <- length(observed)

    #-------------------------------------------
    # No finite observations
    #-------------------------------------------
    if (n_observed == 0L) {

      return(
        list(
          Numeric = TRUE,
          Missing = missing,
          Infinite = as.integer(infinite),
          Observed = 0L,
          Q1 = NA_real_,
          Q3 = NA_real_,
          IQR = NA_real_,
          LowerBound = NA_real_,
          UpperBound = NA_real_,
          LowerOutliers = 0L,
          UpperOutliers = 0L,
          Outliers = 0L,
          OutlierPercent = NA_real_
        )
      )
    }

    #-------------------------------------------
    # Quartiles
    #-------------------------------------------
    quartiles <- stats::quantile(
      observed,
      probs = c(0.25, 0.75),
      names = FALSE,
      type = 7
    )

    q1 <- as.numeric(quartiles[1L])
    q3 <- as.numeric(quartiles[2L])

    #-------------------------------------------
    # IQR
    #-------------------------------------------
    iqr <- q3 - q1

    #-------------------------------------------
    # Bounds
    #-------------------------------------------
    lower_bound <- q1 - k * iqr
    upper_bound <- q3 + k * iqr

    #-------------------------------------------
    # Outliers
    #-------------------------------------------
    lower_outliers <- sum(
      observed < lower_bound
    )

    upper_outliers <- sum(
      observed > upper_bound
    )

    outliers <-
      lower_outliers +
      upper_outliers

    #-------------------------------------------
    # Percentage
    #-------------------------------------------
    outlier_percent <-
      100 * outliers / n_observed

    list(
      Numeric = TRUE,
      Missing = missing,
      Infinite = as.integer(infinite),
      Observed = as.integer(n_observed),
      Q1 = q1,
      Q3 = q3,
      IQR = iqr,
      LowerBound = lower_bound,
      UpperBound = upper_bound,
      LowerOutliers = as.integer(lower_outliers),
      UpperOutliers = as.integer(upper_outliers),
      Outliers = as.integer(outliers),
      OutlierPercent = outlier_percent
    )
  }

  #---------------------------------------------
  # Apply analysis
  #---------------------------------------------
  diagnostics <- lapply(
    data,
    analyse_variable
  )

  #---------------------------------------------
  # Construct result
  #---------------------------------------------
  result <- data.frame(

    Variable = names(data),

    Numeric = vapply(
      diagnostics,
      function(x) x$Numeric,
      logical(1)
    ),

    Missing = vapply(
      diagnostics,
      function(x) x$Missing,
      integer(1)
    ),

    Infinite = vapply(
      diagnostics,
      function(x) x$Infinite,
      integer(1)
    ),

    Observed = vapply(
      diagnostics,
      function(x) x$Observed,
      integer(1)
    ),

    Q1 = vapply(
      diagnostics,
      function(x) x$Q1,
      numeric(1)
    ),

    Q3 = vapply(
      diagnostics,
      function(x) x$Q3,
      numeric(1)
    ),

    IQR = vapply(
      diagnostics,
      function(x) x$IQR,
      numeric(1)
    ),

    LowerBound = vapply(
      diagnostics,
      function(x) x$LowerBound,
      numeric(1)
    ),

    UpperBound = vapply(
      diagnostics,
      function(x) x$UpperBound,
      numeric(1)
    ),

    LowerOutliers = vapply(
      diagnostics,
      function(x) x$LowerOutliers,
      integer(1)
    ),

    UpperOutliers = vapply(
      diagnostics,
      function(x) x$UpperOutliers,
      integer(1)
    ),

    Outliers = vapply(
      diagnostics,
      function(x) x$Outliers,
      integer(1)
    ),

    OutlierPercent = vapply(
      diagnostics,
      function(x) x$OutlierPercent,
      numeric(1)
    ),

    stringsAsFactors = FALSE,
    check.names = FALSE
  )

  #---------------------------------------------
  # Flag variables
  #---------------------------------------------
  result$HasOutlier <-
    !is.na(result$Outliers) &
    result$Outliers > 0L

  #---------------------------------------------
  # Round reporting values
  #---------------------------------------------
  numeric_columns <- c(
    "Q1",
    "Q3",
    "IQR",
    "LowerBound",
    "UpperBound",
    "OutlierPercent"
  )

  result[numeric_columns] <- lapply(
    result[numeric_columns],
    function(x) round(x, 2)
  )

  #---------------------------------------------
  # Class
  #---------------------------------------------
  class(result) <- c(
    "OutlierCheck",
    "data.frame"
  )

  result
}

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("outlier_check returns OutlierCheck", {

  result <- outlier_check(iris)

  expect_s3_class(
    result,
    "OutlierCheck"
  )

})

test_that("outlier_check returns data.frame", {

  result <- outlier_check(iris)

  expect_s3_class(
    result,
    "data.frame"
  )

})

test_that("outlier_check returns one row per variable", {

  result <- outlier_check(iris)

  expect_equal(
    nrow(result),
    ncol(iris)
  )

})

test_that("outlier_check has expected columns", {

  result <- outlier_check(iris)

  expect_equal(
    names(result),
    c(
      "Variable",
      "Numeric",
      "Missing",
      "Infinite",
      "Observed",
      "Q1",
      "Q3",
      "IQR",
      "LowerBound",
      "UpperBound",
      "LowerOutliers",
      "UpperOutliers",
      "Outliers",
      "OutlierPercent",
      "HasOutlier"
    )
  )

})


#==========================================================
# No outliers
#==========================================================

test_that("outlier_check identifies data without outliers", {

  dat <- data.frame(
    x = 1:10
  )

  result <- outlier_check(dat)

  expect_equal(
    result$Outliers,
    0L
  )

  expect_false(
    result$HasOutlier
  )

})


#==========================================================
# Upper outlier
#==========================================================

test_that("outlier_check detects upper outlier", {

  dat <- data.frame(
    x = c(
      rep(10, 20),
      100
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$UpperOutliers,
    1L
  )

  expect_equal(
    result$LowerOutliers,
    0L
  )

  expect_equal(
    result$Outliers,
    1L
  )

  expect_true(
    result$HasOutlier
  )

})


#==========================================================
# Lower outlier
#==========================================================

test_that("outlier_check detects lower outlier", {

  dat <- data.frame(
    x = c(
      -100,
      rep(10, 20)
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$LowerOutliers,
    1L
  )

  expect_equal(
    result$UpperOutliers,
    0L
  )

  expect_equal(
    result$Outliers,
    1L
  )

})


#==========================================================
# Both directions
#==========================================================

test_that("outlier_check detects lower and upper outliers", {

  dat <- data.frame(
    x = c(
      -100,
      rep(10, 20),
      100
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$LowerOutliers,
    1L
  )

  expect_equal(
    result$UpperOutliers,
    1L
  )

  expect_equal(
    result$Outliers,
    2L
  )

})


#==========================================================
# Constant variable
#==========================================================

test_that("outlier_check handles constant variable", {

  dat <- data.frame(
    x = rep(5, 10)
  )

  result <- outlier_check(dat)

  expect_equal(
    result$IQR,
    0
  )

  expect_equal(
    result$LowerBound,
    5
  )

  expect_equal(
    result$UpperBound,
    5
  )

  expect_equal(
    result$Outliers,
    0L
  )

  expect_false(
    result$HasOutlier
  )

})


#==========================================================
# Missing values
#==========================================================

test_that("outlier_check excludes missing values", {

  dat <- data.frame(
    x = c(
      rep(10, 20),
      100,
      NA,
      NA
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$Missing,
    2L
  )

  expect_equal(
    result$Observed,
    21L
  )

  expect_equal(
    result$Outliers,
    1L
  )

})


#==========================================================
# Infinite values
#==========================================================

test_that("outlier_check excludes infinite values", {

  dat <- data.frame(
    x = c(
      1:10,
      Inf,
      -Inf
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$Infinite,
    2L
  )

  expect_equal(
    result$Observed,
    10L
  )

  expect_equal(
    result$Outliers,
    0L
  )

})


#==========================================================
# No finite observations
#==========================================================

test_that("outlier_check handles all missing numeric data", {

  dat <- data.frame(
    x = rep(
      NA_real_,
      5
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$Observed,
    0L
  )

  expect_equal(
    result$Missing,
    5L
  )

  expect_true(
    is.na(result$Q1)
  )

  expect_true(
    is.na(result$Q3)
  )

  expect_true(
    is.na(result$IQR)
  )

  expect_equal(
    result$Outliers,
    0L
  )

  expect_false(
    result$HasOutlier
  )

})

test_that("outlier_check handles only infinite values", {

  dat <- data.frame(
    x = c(
      Inf,
      -Inf
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$Observed,
    0L
  )

  expect_equal(
    result$Infinite,
    2L
  )

  expect_true(
    is.na(result$LowerBound)
  )

  expect_true(
    is.na(result$UpperBound)
  )

  expect_true(
    is.na(result$OutlierPercent)
  )

})


#==========================================================
# Non-numeric variables
#==========================================================

test_that("outlier_check handles character variables", {

  dat <- data.frame(
    x = c(
      "A",
      "B",
      "C"
    ),
    stringsAsFactors = FALSE
  )

  result <- outlier_check(dat)

  expect_false(
    result$Numeric
  )

  expect_true(
    is.na(result$Outliers)
  )

  expect_false(
    result$HasOutlier
  )

})

test_that("outlier_check handles factor variables", {

  dat <- data.frame(
    x = factor(
      c(
        "A",
        "B",
        "C"
      )
    )
  )

  result <- outlier_check(dat)

  expect_false(
    result$Numeric
  )

  expect_true(
    is.na(result$IQR)
  )

})

test_that("outlier_check handles logical variables", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE,
      TRUE
    )
  )

  result <- outlier_check(dat)

  expect_false(
    result$Numeric
  )

  expect_true(
    is.na(result$Outliers)
  )

})


#==========================================================
# Date variables
#==========================================================

test_that("outlier_check does not treat Date as numeric", {

  dat <- data.frame(
    x = as.Date("2026-01-01") + 0:4
  )

  result <- outlier_check(dat)

  expect_false(
    result$Numeric
  )

  expect_true(
    is.na(result$Outliers)
  )

})


#==========================================================
# POSIXct
#==========================================================

test_that("outlier_check does not treat POSIXct as numeric", {

  dat <- data.frame(
    x = as.POSIXct(
      "2026-01-01",
      tz = "UTC"
    ) + 0:4
  )

  result <- outlier_check(dat)

  expect_false(
    result$Numeric
  )

})


#==========================================================
# Multiple variables
#==========================================================

test_that("outlier_check handles mixed data", {

  dat <- data.frame(
    normal = 1:21,
    upper = c(
      rep(10, 20),
      100
    ),
    group = rep(
      "A",
      21
    ),
    stringsAsFactors = FALSE
  )

  result <- outlier_check(dat)

  expect_equal(
    result$Numeric,
    c(
      TRUE,
      TRUE,
      FALSE
    )
  )

  expect_equal(
    result$Outliers,
    c(
      0L,
      1L,
      NA_integer_
    )
  )

  expect_equal(
    result$HasOutlier,
    c(
      FALSE,
      TRUE,
      FALSE
    )
  )

})


#==========================================================
# Percentage
#==========================================================

test_that("outlier_check calculates percentage correctly", {

  dat <- data.frame(
    x = c(
      rep(10, 20),
      100
    )
  )

  result <- outlier_check(dat)

  expect_equal(
    result$OutlierPercent,
    4.76
  )

})


#==========================================================
# Custom k
#==========================================================

test_that("outlier_check accepts custom k", {

  dat <- data.frame(
    x = c(
      1:10,
      20
    )
  )

  result1 <- outlier_check(
    dat,
    k = 1.5
  )

  result2 <- outlier_check(
    dat,
    k = 3
  )

  expect_true(
    result1$Outliers >= result2$Outliers
  )

})

test_that("smaller k can identify more outliers", {

  dat <- data.frame(
    x = c(
      1:10,
      15
    )
  )

  result1 <- outlier_check(
    dat,
    k = 0.5
  )

  result2 <- outlier_check(
    dat,
    k = 3
  )

  expect_true(
    result1$Outliers >= result2$Outliers
  )

})


#==========================================================
# k validation
#==========================================================

test_that("outlier_check rejects character k", {

  expect_error(
    outlier_check(
      iris,
      k = "1.5"
    ),
    "'k' must be a single finite numeric value"
  )

})

test_that("outlier_check rejects vector k", {

  expect_error(
    outlier_check(
      iris,
      k = c(1.5, 3)
    ),
    "'k' must be a single finite numeric value"
  )

})

test_that("outlier_check rejects NA k", {

  expect_error(
    outlier_check(
      iris,
      k = NA_real_
    ),
    "'k' must be a single finite numeric value"
  )

})

test_that("outlier_check rejects infinite k", {

  expect_error(
    outlier_check(
      iris,
      k = Inf
    ),
    "'k' must be a single finite numeric value"
  )

})

test_that("outlier_check rejects zero k", {

  expect_error(
    outlier_check(
      iris,
      k = 0
    ),
    "'k' must be greater than zero"
  )

})

test_that("outlier_check rejects negative k", {

  expect_error(
    outlier_check(
      iris,
      k = -1
    ),
    "'k' must be greater than zero"
  )

})


#==========================================================
# Zero-row data
#==========================================================

test_that("outlier_check handles zero-row data", {

  dat <- data.frame(
    x = numeric(0),
    y = character(0)
  )

  result <- outlier_check(dat)

  expect_s3_class(
    result,
    "OutlierCheck"
  )

  expect_equal(
    nrow(result),
    2L
  )

  expect_equal(
    result$Observed[1],
    0L
  )

  expect_true(
    is.na(result$Observed[2])
  )

  expect_false(
    any(result$HasOutlier)
  )

})


#==========================================================
# Tibble
#==========================================================

test_that("outlier_check handles tibble", {

  skip_if_not_installed("tibble")

  dat <- tibble::tibble(
    x = c(
      rep(10, 20),
      100
    ),
    group = rep(
      "A",
      21
    )
  )

  result <- outlier_check(dat)

  expect_s3_class(
    result,
    "OutlierCheck"
  )

  expect_equal(
    result$Outliers[1],
    1L
  )

})


#==========================================================
# Invalid data
#==========================================================

test_that("outlier_check rejects NULL", {

  expect_error(
    outlier_check(NULL),
    "'data' cannot be NULL"
  )

})

test_that("outlier_check rejects numeric vector", {

  expect_error(
    outlier_check(1:10),
    "'data' must be a data.frame"
  )

})

test_that("outlier_check rejects character vector", {

  expect_error(
    outlier_check(letters),
    "'data' must be a data.frame"
  )

})

test_that("outlier_check rejects list", {

  expect_error(
    outlier_check(
      list(
        x = 1:10
      )
    ),
    "'data' must be a data.frame"
  )

})

test_that("outlier_check rejects empty data.frame", {

  expect_error(
    outlier_check(
      data.frame()
    ),
    "at least one variable"
  )

})

library(testthat)
library(DataAudit)

#==========================================================
# Basic output
#==========================================================

test_that("whitespace_check returns WhitespaceCheck", {

  dat <- data.frame(
    x = c("Dog", "Cat")
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_s3_class(
    result,
    "WhitespaceCheck"
  )

  expect_s3_class(
    result,
    "data.frame"
  )
})

test_that("whitespace_check returns expected columns", {

  dat <- data.frame(
    x = c("Dog", "Cat")
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    names(result),
    c(
      "Row",
      "Value",
      "Missing",
      "LeadingWhitespace",
      "TrailingWhitespace",
      "WhitespaceOnly",
      "WhitespaceIssue"
    )
  )
})

#==========================================================
# Detection
#==========================================================

test_that("whitespace_check detects leading whitespace", {

  dat <- data.frame(
    x = c("Dog", " Cat", "  Cattle")
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$LeadingWhitespace,
    c(FALSE, TRUE, TRUE)
  )
})

test_that("whitespace_check detects trailing whitespace", {

  dat <- data.frame(
    x = c("Dog", "Cat ", "Cattle  ")
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$TrailingWhitespace,
    c(FALSE, TRUE, TRUE)
  )
})

test_that("whitespace_check detects both leading and trailing whitespace", {

  dat <- data.frame(
    x = c(
      "Dog",
      " Cat ",
      " Cattle"
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$WhitespaceIssue,
    c(FALSE, TRUE, TRUE)
  )

  expect_true(
    result$LeadingWhitespace[2]
  )

  expect_true(
    result$TrailingWhitespace[2]
  )
})

test_that("whitespace_check detects whitespace-only values", {

  dat <- data.frame(
    x = c(
      "Dog",
      " ",
      "   ",
      ""
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$WhitespaceOnly,
    c(FALSE, TRUE, TRUE, FALSE)
  )
})

test_that("empty string is not whitespace", {

  dat <- data.frame(
    x = c("", " ")
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_false(
    result$WhitespaceIssue[1]
  )

  expect_true(
    result$WhitespaceIssue[2]
  )
})

#==========================================================
# Missing values
#==========================================================

test_that("whitespace_check handles missing values", {

  dat <- data.frame(
    x = c(
      "Dog",
      NA,
      " Cat "
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$Missing,
    c(FALSE, TRUE, FALSE)
  )

  expect_false(
    result$WhitespaceIssue[2]
  )

  expect_true(
    result$WhitespaceIssue[3]
  )
})

#==========================================================
# Factors
#==========================================================

test_that("whitespace_check works with factors", {

  dat <- data.frame(
    x = factor(
      c(
        "Dog",
        " Cat",
        "Cattle "
      )
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$WhitespaceIssue,
    c(FALSE, TRUE, TRUE)
  )
})

#==========================================================
# Tabs and newlines
#==========================================================

test_that("whitespace_check detects tabs", {

  dat <- data.frame(
    x = c(
      "Dog",
      "\tCat",
      "Cattle\t"
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$WhitespaceIssue,
    c(FALSE, TRUE, TRUE)
  )
})

test_that("whitespace_check detects newline whitespace", {

  dat <- data.frame(
    x = c(
      "Dog",
      "\nCat",
      "Cattle\n"
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_equal(
    result$WhitespaceIssue,
    c(FALSE, TRUE, TRUE)
  )
})

#==========================================================
# Valid data
#==========================================================

test_that("clean values are not flagged", {

  dat <- data.frame(
    x = c(
      "Dog",
      "Domestic Cat",
      "Dairy Cattle"
    )
  )

  result <- whitespace_check(
    dat,
    "x"
  )

  expect_false(
    any(result$WhitespaceIssue)
  )
})

#==========================================================
# Invalid variable argument
#==========================================================

test_that("whitespace_check rejects missing variable", {

  expect_error(
    whitespace_check(iris),
    "'variable' must be a single non-empty character string"
  )
})

test_that("whitespace_check rejects NULL variable", {

  expect_error(
    whitespace_check(
      iris,
      NULL
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("whitespace_check rejects empty variable name", {

  expect_error(
    whitespace_check(
      iris,
      ""
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("whitespace_check rejects numeric variable argument", {

  expect_error(
    whitespace_check(
      iris,
      1
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("whitespace_check rejects multiple variable names", {

  expect_error(
    whitespace_check(
      iris,
      c(
        "Species",
        "Sepal.Length"
      )
    ),
    "'variable' must be a single non-empty character string"
  )
})

test_that("whitespace_check rejects NA variable", {

  expect_error(
    whitespace_check(
      iris,
      NA_character_
    ),
    "'variable' must be a single non-empty character string"
  )
})

#==========================================================
# Unknown variable
#==========================================================

test_that("whitespace_check rejects unknown variable", {

  expect_error(
    whitespace_check(
      iris,
      "Unknown"
    ),
    "Variable 'Unknown' was not found in 'data'"
  )
})

#==========================================================
# Invalid variable types
#==========================================================

test_that("whitespace_check rejects numeric column", {

  expect_error(
    whitespace_check(
      iris,
      "Sepal.Length"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

test_that("whitespace_check rejects logical column", {

  dat <- data.frame(
    x = c(
      TRUE,
      FALSE
    )
  )

  expect_error(
    whitespace_check(
      dat,
      "x"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

test_that("whitespace_check rejects Date column", {

  dat <- data.frame(
    x = as.Date(
      c(
        "2026-01-01",
        "2026-01-02"
      )
    )
  )

  expect_error(
    whitespace_check(
      dat,
      "x"
    ),
    "'variable' must refer to a character or factor variable"
  )
})

#==========================================================
# Invalid data
#==========================================================

test_that("whitespace_check rejects invalid data", {

  expect_error(
    whitespace_check(
      NULL,
      "x"
    )
  )

  expect_error(
    whitespace_check(
      1:5,
      "x"
    )
  )

  expect_error(
    whitespace_check(
      list(x = "Dog"),
      "x"
    )
  )

  expect_error(
    whitespace_check(
      data.frame(),
      "x"
    )
  )
})

---
title: "Getting Started with DataAudit"
author: "Vinodh Kumar Obli Rajendran and Keerthi Aaradhana"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with DataAudit}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

Data quality is an important prerequisite for reliable statistical analysis,
visualization, modelling, and interpretation.

`DataAudit` provides a structured framework for identifying common
data-quality problems in R. The package combines general data auditing,
individual validation functions, reusable validation rules, structured audit
reports, and data-quality scoring.

A typical workflow is:

1. inspect the dataset;
2. perform a general audit;
3. conduct focused checks when required;
4. define study-specific validation rules;
5. integrate the rules into the audit;
6. inspect the resulting report; and
7. calculate an overall data-quality score.

# Load DataAudit

```{r load-package}
library(DataAudit)
```

# Example dataset

Consider a small dataset containing several intentional data-quality problems.

```{r example-data}
dat <- data.frame(
  ID = c(
    "A001",
    "A002",
    "A002",
    "BAD",
    "A005"
  ),
  Species = c(
    "Dog",
    "Horse",
    "",
    "Cat",
    "Dog"
  ),
  Age = c(
    5,
    40,
    -2,
    10,
    NA
  ),
  stringsAsFactors = FALSE
)

dat
```

The dataset contains examples of:

- a duplicated identifier;
- a value outside the intended species categories;
- a blank value;
- ages outside an intended range;
- a missing age; and
- an identifier that does not follow the intended pattern.

# General data audit

The primary high-level function is `audit_data()`.

```{r general-audit}
report <- audit_data(
  dat,
  include_nzv = FALSE
)

report
```

The resulting object inherits from the `DataAuditReport` class.

```{r report-class}
class(report)
```

The report contains several components.

```{r report-components}
names(report)
```

These provide information about:

- the overall dataset;
- variable-level issues;
- duplicated rows;
- overall issue counts; and
- custom validation-rule results when rules are supplied.

# Dataset overview

The overview component provides high-level audit statistics.

```{r overview}
report$overview
```

# Variable-level diagnostics

Variable-level diagnostics are stored in the `variables` component.

```{r variable-results}
report$variables
```

This provides information such as missing values, blank values, infinite
values, numbers of unique values, constant variables, near-zero variance
variables, and whether a variable contains a detected issue.

# Duplicate observations

Duplicated observations can be inspected through:

```{r duplicates}
report$duplicates
```

This checks duplicated rows. Identifier uniqueness can be assessed separately
using `unique_check()` or through validation rules.

# Individual checks

DataAudit functions can also be used independently.

## Missing values

```{r missing-check}
missing_check(dat)
```

## Blank values

```{r blank-check}
blank_check(dat)
```

## Range validation

Suppose valid ages are expected to range from 0 to 30.

```{r range-check}
range_check(
  dat,
  variable = "Age",
  min = 0,
  max = 30
)
```

## Category validation

Suppose only `"Dog"` and `"Cat"` are accepted categories.

```{r category-check}
category_check(
  dat,
  variable = "Species",
  allowed = c(
    "Dog",
    "Cat"
  )
)
```

## Identifier uniqueness

```{r unique-check}
unique_check(
  dat,
  variable = "ID"
)
```

## Pattern validation

Suppose identifiers must begin with `A` followed by three digits.

```{r pattern-check}
pattern_check(
  dat,
  variable = "ID",
  pattern = "^A[0-9]{3}$"
)
```

# Reusable validation rules

For repeated or study-specific validation, multiple requirements can be
defined together using `audit_rules()`.

```{r define-rules}
rules <- audit_rules(
  range = list(
    Age = c(0, 30)
  ),
  category = list(
    Species = c(
      "Dog",
      "Cat"
    )
  ),
  unique = "ID",
  required = "Species",
  pattern = list(
    ID = "^A[0-9]{3}$"
  )
)

rules
```

The rules object can be reused with datasets having the same expected
structure.

# Apply validation rules

Rules can be evaluated directly using `apply_audit_rules()`.

```{r apply-rules}
rule_results <- apply_audit_rules(
  dat,
  rules
)

rule_results
```

The returned object contains the rule-specific results and total number of
violations.

```{r rule-summary}
rule_results$summary
rule_results$total_violations
```

A compact summary can also be obtained using:

```{r summarize-rules}
summary(rule_results)
```

# Integrated audit

The validation rules can be incorporated directly into `audit_data()`.

```{r integrated-audit}
report <- audit_data(
  dat,
  include_nzv = FALSE,
  rules = rules
)

report
```

The resulting report now combines automatic data-quality assessment with
user-defined validation rules.

```{r integrated-components}
report$summary
report$rules
```

# Summarizing an audit

The report can be summarized using the standard `summary()` generic.

```{r report-summary}
audit_summary <- summary(report)

audit_summary
```

The summary object provides compact access to dataset-level information,
variables with issues, duplicated rows, and rule results.

```{r summary-components}
audit_summary$dataset
audit_summary$variables_with_issues
audit_summary$duplicate_rows
audit_summary$rules
```

# Data-quality score

An overall data-quality score can be calculated from a `DataAuditReport`.

```{r score}
score <- audit_score(report)

score
```

The score ranges from 0 to 100, with higher values indicating fewer detected
issues relative to dataset size.

The score is accompanied by one of five qualitative categories:

- Excellent;
- Good;
- Moderate;
- Poor; or
- Critical.

The individual components can be inspected directly.

```{r score-components}
score$score
score$quality
score$automatic_issues
score$rule_violations
score$overall_issues
```

The data-quality score is intended as a compact audit summary. It should not
replace domain-specific assessment of whether a dataset is suitable for a
particular scientific or analytical purpose.

# Additional checks

DataAudit provides additional focused functions for different aspects of data
quality, including:

- `constant_check()`;
- `nzv_check()`;
- `infinite_check()`;
- `outlier_check()`;
- `type_check()`;
- `date_check()`;
- `id_check()`;
- `length_check()`;
- `case_check()`;
- `whitespace_check()`;
- `sequence_check()`;
- `group_sequence_check()`;
- `dependency_check()`; and
- `consistency_check()`.

These functions can be combined with the general audit when more detailed
validation is required.

# Recommended workflow

A practical workflow is:

```text
Raw data
   |
   v
audit_data()
   |
   +---- General data-quality assessment
   |
   +---- Individual checks
   |
   v
audit_rules()
   |
   v
apply_audit_rules()
   |
   v
Integrated DataAuditReport
   |
   +---- print()
   |
   +---- summary()
   |
   v
audit_score()
```

This separates general data-quality screening from domain-specific validation
while allowing both to be incorporated into a single structured report.

# Conclusion

`DataAudit` provides a reproducible framework for detecting, organizing, and
summarizing common data-quality problems in R.

The package can be used for rapid preliminary screening with `audit_data()`,
focused validation through individual check functions, reusable validation
with `audit_rules()`, and compact reporting through `summary()` and
`audit_score()`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outlier_check.R
\name{outlier_check}
\alias{outlier_check}
\title{Check Numeric Outliers}
\usage{
outlier_check(data, k = 1.5)
}
\arguments{
\item{data}{A data.frame.}

\item{k}{A single positive numeric value controlling the
IQR multiplier. Default is 1.5.}
}
\value{
A data.frame of class \code{OutlierCheck}
containing variable-level outlier diagnostics.
}
\description{
Identifies potential outliers in numeric variables using
the interquartile range (IQR) method.
}
\details{
For each numeric variable, the lower and upper limits are:

\deqn{Q1 - k * IQR}

and

\deqn{Q3 + k * IQR}

Values strictly below the lower limit or strictly above
the upper limit are classified as potential outliers.

Missing and infinite values are excluded from the
calculation of quartiles and outliers. They are reported
separately.

Non-numeric variables are retained in the output but are
not evaluated for outliers.
}
\examples{
dat <- data.frame(
  x = c(10, 11, 12, 13, 100),
  group = c("A", "A", "B", "B", "B")
)

outlier_check(dat)

outlier_check(dat, k = 3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pattern_check.R
\name{pattern_check}
\alias{pattern_check}
\title{Check Character Patterns}
\usage{
pattern_check(data, variable, pattern, ignore_case = FALSE)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
variable to check.}

\item{pattern}{A single non-empty character string
containing a regular expression.}

\item{ignore_case}{Logical. If TRUE, matching ignores
letter case. Default is FALSE.}
}
\value{
A data.frame of class \code{PatternCheck}
containing row-level pattern diagnostics.
}
\description{
Checks whether values in a character or factor variable
match a user-specified regular expression.
}
\details{
Missing values are reported separately and are not
classified as matching or invalid.

Factor variables are converted to character internally
for pattern matching. The original values are retained
in the returned data frame.

Pattern matching is performed using \code{grepl()}.
}
\examples{
dat <- data.frame(
  Sample_ID = c(
    "SMP-001",
    "SMP-002",
    "BAD-003",
    NA
  ),
  stringsAsFactors = FALSE
)

pattern_check(
  dat,
  variable = "Sample_ID",
  pattern = "^SMP-[0-9]{3}$"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/range_check.R
\name{range_check}
\alias{range_check}
\title{Check Values Against a Valid Range}
\usage{
range_check(data, variable, min = NULL, max = NULL)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{Character string giving the name of the
numeric variable to check.}

\item{min}{Optional numeric lower limit. Default is
\code{NULL}.}

\item{max}{Optional numeric upper limit. Default is
\code{NULL}.}
}
\value{
A data.frame of class \code{RangeCheck}
containing row-level range diagnostics.
}
\description{
Identifies values in a numeric variable that fall below
or above user-defined valid limits.
}
\details{
At least one of \code{min} or \code{max} must be
supplied.

Missing values are not classified as out-of-range.
Infinite values are evaluated against the specified
limits and will normally be flagged when an appropriate
finite limit is supplied.
}
\examples{
dat <- data.frame(
  Temperature = c(38.5, 39.2, 44.0, 34.0, NA)
)

range_check(
  dat,
  variable = "Temperature",
  min = 35,
  max = 43
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constant_check.R
\name{constant_check}
\alias{constant_check}
\title{Check Constant Variables}
\usage{
constant_check(data)
}
\arguments{
\item{data}{A data.frame.}
}
\value{
A data.frame of class \code{ConstantCheck}
containing the number of unique non-missing values,
missing values, and constant-variable status for each
variable.
}
\description{
Identifies variables that contain no meaningful
variation.
}
\details{
A variable is classified as constant when it contains
one or fewer unique non-missing values.

Missing values are excluded when counting unique values.
Consequently, a variable containing only missing values
is also classified as constant.
}
\examples{
dat <- data.frame(
  id = 1:5,
  group = rep("A", 5),
  value = c(10, 10, 10, NA, 10)
)

constant_check(dat)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infinite_check.R
\name{infinite_check}
\alias{infinite_check}
\title{Check Infinite Values}
\usage{
infinite_check(data)
}
\arguments{
\item{data}{A data.frame.}
}
\value{
A data.frame of class \code{InfiniteCheck}
containing variable-level information about infinite
values.
}
\description{
Identifies positive and negative infinite values in
numeric variables of a dataset.
}
\details{
Infinite values are detected only in numeric variables.
Positive infinity (\code{Inf}) and negative infinity
(\code{-Inf}) are reported separately.

Missing values (\code{NA}) are not classified as
infinite values.

Non-numeric variables are retained in the output and
reported with zero infinite values.
}
\examples{
dat <- data.frame(
  x = c(1, 2, Inf, 4),
  y = c(1, -Inf, 3, 4),
  group = c("A", "B", "C", "D")
)

infinite_check(dat)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/date_check.R
\name{date_check}
\alias{date_check}
\title{Check Date Values}
\usage{
date_check(
  data,
  variable,
  min_date = NULL,
  max_date = NULL,
  allow_future = TRUE
)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
Date or POSIXct variable to check.}

\item{min_date}{Optional minimum acceptable date.}

\item{max_date}{Optional maximum acceptable date.}

\item{allow_future}{Logical. If FALSE, dates later than
the current date are flagged. Default is TRUE.}
}
\value{
A data.frame of class \code{DateCheck} containing
row-level date diagnostics.
}
\description{
Checks a Date or POSIXct variable for missing values,
dates outside user-defined limits, and optionally future
dates.
}
\details{
The selected variable must inherit from \code{Date} or
\code{POSIXct}.

Missing values are reported separately and are not
classified as out-of-range dates.

When \code{allow_future = FALSE}, observations later than
the current date are flagged in \code{Future}.

The \code{min_date} and \code{max_date} arguments may be
supplied as Date, POSIXct, or single character strings
in ISO format (\code{YYYY-MM-DD}).
}
\examples{
dat <- data.frame(
  Sample_Date = as.Date(c(
    "2025-01-01",
    "2025-06-01",
    "2026-01-01",
    NA
  ))
)

date_check(
  dat,
  variable = "Sample_Date",
  min_date = "2025-01-01",
  max_date = "2025-12-31"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nzv_check.R
\name{nzv_check}
\alias{nzv_check}
\title{Check Near-Zero Variance Variables}
\usage{
nzv_check(data, unique_cut = 10, freq_cut = 19)
}
\arguments{
\item{data}{A data.frame.}

\item{unique_cut}{Numeric. Maximum percentage of unique
values used to identify near-zero variance variables.
Default is 10.}

\item{freq_cut}{Numeric. Minimum frequency ratio used to
identify near-zero variance variables. Default is 19.}
}
\value{
A data.frame of class \code{NZVCheck} containing
variable-level near-zero variance diagnostics.
}
\description{
Identifies variables with zero or near-zero variance
based on the percentage of unique values and the
frequency ratio of the two most common values.
}
\details{
Missing values are excluded when calculating the number
of unique values, percentage of unique values, and
frequency ratio.

A variable is classified as having zero variance when it
contains one or fewer unique non-missing values.

A non-constant variable is classified as near-zero
variance when both:

\itemize{
\item PercentUnique is less than or equal to
\code{unique_cut}.
\item FrequencyRatio is greater than \code{freq_cut}.
}

Variables with zero variance are also flagged in the
\code{NearZeroVariance} column.
}
\examples{
dat <- data.frame(
  x = 1:100,
  y = c(rep(0, 99), 1),
  z = rep(5, 100)
)

nzv_check(dat)

nzv_check(
  dat,
  unique_cut = 20,
  freq_cut = 10
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/category_check.R
\name{category_check}
\alias{category_check}
\title{Check Categorical Values}
\usage{
category_check(data, variable, allowed)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
variable to check.}

\item{allowed}{A non-empty vector containing the allowed
values.}
}
\value{
A data.frame of class \code{CategoryCheck}
containing row-level category diagnostics.
}
\description{
Checks whether values in a variable belong to a
user-defined set of allowed categories.
}
\details{
Missing values are reported separately and are not
classified as invalid categories.

The supplied \code{allowed} values are compared with the
observed values using their underlying values. Character,
factor, numeric, integer, and logical variables are
supported.
}
\examples{
dat <- data.frame(
  Species = c("Dog", "Cat", "Horse", NA),
  stringsAsFactors = FALSE
)

category_check(
  dat,
  variable = "Species",
  allowed = c("Dog", "Cat", "Cattle")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_overview.R
\name{data_overview}
\alias{data_overview}
\title{Data Overview}
\usage{
data_overview(data)
}
\arguments{
\item{data}{A data.frame.}
}
\value{
An object of class \code{DataOverview} containing
dataset-level summary information.
}
\description{
Provides a concise overview of the size, structure,
completeness, and variable types of a dataset.
}
\details{
The function reports the number of rows and variables,
total number of cells, missing and complete cells,
completeness percentage, complete cases, duplicate rows,
and counts of common variable types.
}
\examples{
data_overview(iris)

dat <- iris
dat$Sepal.Length[1:5] <- NA
data_overview(dat)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_DataAuditRuleResult.R
\name{print.DataAuditRuleResult}
\alias{print.DataAuditRuleResult}
\title{Print a DataAudit Rule Result}
\usage{
\method{print}{DataAuditRuleResult}(x, ...)
}
\arguments{
\item{x}{An object of class \code{DataAuditRuleResult}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
The input object \code{x}, invisibly.
}
\description{
Prints a concise summary of custom audit-rule results
produced by \code{\link{apply_audit_rules}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/duplicate_check.R
\name{duplicate_check}
\alias{duplicate_check}
\title{Check Duplicate Rows}
\usage{
duplicate_check(data)
}
\arguments{
\item{data}{A data.frame.}
}
\value{
An object of class \code{DuplicateCheck}. The
returned data.frame contains the row number and a logical
indicator showing whether each row is a duplicate of an
earlier row.
}
\description{
Identifies duplicate rows in a dataset and reports their
row positions.
}
\details{
Duplicate rows are identified using \code{duplicated()}.
The first occurrence of a repeated row is not classified
as a duplicate. Only subsequent occurrences are marked as
duplicates.
}
\examples{
duplicate_check(iris)

dat <- data.frame(
  x = c(1, 2, 2, 3),
  y = c("A", "B", "B", "C")
)

duplicate_check(dat)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sequence_check.R
\name{sequence_check}
\alias{sequence_check}
\title{Check Numeric Sequence Integrity}
\usage{
sequence_check(data, variable, step = 1)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
numeric or integer variable to check.}

\item{step}{A positive numeric value specifying the
expected increment between consecutive values.
Default is 1.}
}
\value{
A data.frame of class \code{SequenceCheck}
containing row-level sequence diagnostics.
}
\description{
Checks a numeric or integer variable for missing values,
duplicated values, ordering problems, and gaps in an
expected sequence.
}
\details{
The sequence is evaluated in the original row order.

A duplicate is a non-missing value appearing more than
once in the variable. All occurrences of duplicated
values are flagged.

For each non-missing observation after the first available
observation, the difference from the previous non-missing
value is calculated.

An ordering problem occurs when this difference is less
than or equal to zero. Duplicate values are therefore also
ordering problems.

A gap occurs when the difference from the previous
non-missing value is greater than \code{step}.

Differences smaller than \code{step} but greater than zero
are classified as step mismatches.

Missing observations are reported separately.
}
\examples{
dat <- data.frame(
  Visit = c(1, 2, 4, 4, 3, NA, 5)
)

sequence_check(
  dat,
  variable = "Visit"
)

sequence_check(
  data.frame(Time = c(0, 5, 10, 20)),
  variable = "Time",
  step = 5
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit_data.R
\name{audit_data}
\alias{audit_data}
\title{Perform a General Data Quality Audit}
\usage{
audit_data(data, include_nzv = TRUE, rules = NULL)
}
\arguments{
\item{data}{A data.frame.}

\item{include_nzv}{Logical. If TRUE, near-zero variance
variables are included in the audit. Default is TRUE.}

\item{rules}{Optional object of class \code{DataAuditRules}
created by \code{\link{audit_rules}}. When supplied,
the rules are evaluated and included in the audit
report.}
}
\value{
An object of class \code{DataAuditReport}.
}
\description{
Performs a general audit of a data frame and summarizes
common data-quality problems including missing values,
blank values, infinite values, constant variables,
near-zero variance variables, and duplicated rows.
}
\details{
\code{audit_data()} is intended as a general first-pass
audit requiring minimal user configuration.

More specialized checks such as ranges, categories,
patterns, identifiers, dates, dependencies, and
cross-variable consistency can be performed using the
corresponding DataAudit functions.

Custom validation rules created with
\code{\link{audit_rules}} can be supplied through
\code{rules}. These rules are evaluated using
\code{\link{apply_audit_rules}}.

The returned object contains:

\itemize{
\item \code{overview}: basic dataset information.
\item \code{variables}: variable-level audit summary.
\item \code{duplicates}: duplicated-row diagnostics.
\item \code{summary}: overall audit statistics.
\item \code{rules}: custom rule results, or NULL when
no rules are supplied.
}
}
\examples{
dat <- data.frame(
  ID = c(1, 2, 2, 4),
  Species = c("Dog", "", "", "Cat"),
  Value = c(10, NA, Inf, 20),
  Constant = 1
)

audit_data(dat)

rules <- audit_rules(
  range = list(
    Value = c(0, 100)
  ),
  required = "Species"
)

audit_data(
  dat,
  rules = rules
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_DataAuditReport.R
\name{print.DataAuditReport}
\alias{print.DataAuditReport}
\title{Print a DataAudit Report}
\usage{
\method{print}{DataAuditReport}(x, ...)
}
\arguments{
\item{x}{An object of class \code{DataAuditReport}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
The input object \code{x}, invisibly.
}
\description{
Prints a concise summary of a data audit produced by
\code{\link{audit_data}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DataAudit-package.R
\docType{package}
\name{DataAudit-package}
\alias{DataAudit}
\alias{DataAudit-package}
\title{DataAudit: Comprehensive Data Quality Auditing and Validation}
\description{
DataAudit provides tools for systematic assessment,
validation, and reporting of data quality in R.
}
\details{
The package supports both general-purpose data auditing
and user-defined validation rules.
}
\section{General data auditing}{


The main function \code{\link{audit_data}} performs a
general assessment of a data frame, including detection
of missing values, blank values, infinite values,
duplicated observations, constant variables, and
near-zero variance variables.
}

\section{Validation rules}{


Reusable validation rules can be defined using
\code{\link{audit_rules}} and evaluated using
\code{\link{apply_audit_rules}}.

Rules can be incorporated directly into
\code{\link{audit_data}} to combine automatic auditing
with study-specific validation requirements.
}

\section{Individual checks}{


DataAudit provides focused functions for checking
ranges, categories, data types, patterns, dates,
identifiers, lengths, character case, whitespace,
sequences, dependencies, uniqueness, duplicates,
outliers, and cross-variable consistency.
}

\section{Audit reports}{


\code{\link{audit_data}} returns a structured
\code{DataAuditReport} object containing dataset-level,
variable-level, duplicate-row, and optional rule-based
diagnostics.

Print and summary methods are provided for convenient
inspection of audit results.
}

\section{Data-quality scoring}{


\code{\link{audit_score}} calculates a standardized
data-quality score from a \code{DataAuditReport}.

Scores range from 0 to 100, with higher values
indicating fewer detected data-quality issues relative
to dataset size.
}

\section{Typical workflow}{


A typical workflow consists of:

\enumerate{
\item Inspect the dataset using \code{\link{audit_data}}.
\item Define study-specific rules using
\code{\link{audit_rules}}.
\item Integrate the rules into the audit.
\item Inspect the resulting \code{DataAuditReport}.
\item Calculate an overall score using
\code{\link{audit_score}} when appropriate.
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/vinodhpmd/DataAudit}
  \item Report bugs at \url{https://github.com/vinodhpmd/DataAudit/issues}
}

}
\author{
\strong{Maintainer}: Vinodh Kumar Obli Rajendran \email{vinodhkumar.rajendran@gmail.com} (\href{https://orcid.org/0000-0002-7232-4122}{ORCID})

Authors:
\itemize{
  \item Vinodh Kumar Obli Rajendran \email{vinodhkumar.rajendran@gmail.com} (\href{https://orcid.org/0000-0002-7232-4122}{ORCID})
  \item Keerthi Aaradhana \email{vkeerthi1817@gmail.com}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_check.R
\name{type_check}
\alias{type_check}
\title{Check Variable Types}
\usage{
type_check(data, expected)
}
\arguments{
\item{data}{A data.frame.}

\item{expected}{A named character vector specifying the
expected type for each variable.}
}
\value{
A data.frame of class \code{TypeCheck}
containing variable-level type diagnostics.
}
\description{
Checks whether variables in a dataset match user-defined
expected data types.
}
\details{
Supported expected types are:

\itemize{
\item \code{"numeric"}
\item \code{"integer"}
\item \code{"character"}
\item \code{"factor"}
\item \code{"logical"}
\item \code{"Date"}
\item \code{"POSIXct"}
}

The \code{expected} argument must be a named character
vector. Names correspond to variables in \code{data}.

For example:

\code{c(Age = "numeric", Species = "factor")}
}
\examples{
dat <- data.frame(
  Age = c(2, 4, 6),
  Species = factor(c("Dog", "Cat", "Dog")),
  Positive = c(TRUE, FALSE, TRUE)
)

type_check(
  dat,
  c(
    Age = "numeric",
    Species = "factor",
    Positive = "logical"
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit_rules.R
\name{audit_rules}
\alias{audit_rules}
\title{Create Data Audit Rules}
\usage{
audit_rules(
  range = NULL,
  category = NULL,
  unique = NULL,
  required = NULL,
  pattern = NULL
)
}
\arguments{
\item{range}{Named list. Each element specifies the
acceptable numeric range for a variable as a numeric
vector of length two.}

\item{category}{Named list. Each element contains the
allowed values for a variable.}

\item{unique}{Character vector containing variables that
should contain unique values.}

\item{required}{Character vector containing variables
that should not contain missing values.}

\item{pattern}{Named list. Each element specifies a
regular expression that values of a variable should
satisfy.}
}
\value{
An object of class \code{DataAuditRules}.
}
\description{
Creates a structured collection of user-defined data
validation rules for use with DataAudit.
}
\details{
\code{audit_rules()} defines validation requirements but
does not evaluate a dataset. The resulting object can be
supplied to functions that apply DataAudit rules.

Variable names are not checked against a dataset when the
rule object is created because the same rule specification
may be reused across multiple datasets.
}
\examples{
rules <- audit_rules(
  range = list(
    Age = c(0, 30),
    Weight = c(0, 100)
  ),
  category = list(
    Species = c("Dog", "Cat", "Cattle")
  ),
  unique = "Animal_ID",
  required = c("Animal_ID", "Species"),
  pattern = list(
    Animal_ID = "^A[0-9]+$"
  )
)

rules

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whitespace_check.R
\name{whitespace_check}
\alias{whitespace_check}
\title{Check Leading and Trailing Whitespace}
\usage{
whitespace_check(data, variable)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
variable to check.}
}
\value{
A data.frame of class \code{WhitespaceCheck}
containing row-level whitespace diagnostics.
}
\description{
Detects leading and trailing whitespace in a character
or factor variable.
}
\details{
Missing values are reported separately and are not
classified as whitespace problems.

The function detects leading whitespace, trailing
whitespace, and values consisting entirely of whitespace.
It does not modify the original data.
}
\examples{
dat <- data.frame(
  Species = c("Dog", " Cat", "Cattle ", " Goat ", NA)
)

whitespace_check(
  dat,
  variable = "Species"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/apply_audit_rules.R
\name{apply_audit_rules}
\alias{apply_audit_rules}
\title{Apply Data Audit Rules}
\usage{
apply_audit_rules(data, rules)
}
\arguments{
\item{data}{A data.frame.}

\item{rules}{An object of class \code{DataAuditRules}
created by \code{\link{audit_rules}}.}
}
\value{
An object of class \code{DataAuditRuleResult}.
}
\description{
Applies a set of validation rules created by
\code{\link{audit_rules}} to a data frame.
}
\details{
The function evaluates range, category, uniqueness,
required-value, and regular-expression rules.

Rules referring to variables that are not present in
\code{data} produce an error.
}
\examples{
dat <- data.frame(
  Animal_ID = c("A001", "A002", "A002", NA),
  Species = c("Dog", "Cat", "Horse", "Dog"),
  Age = c(5, 3, 40, 2),
  stringsAsFactors = FALSE
)

rules <- audit_rules(
  range = list(
    Age = c(0, 30)
  ),
  category = list(
    Species = c("Dog", "Cat", "Cattle")
  ),
  unique = "Animal_ID",
  required = "Animal_ID",
  pattern = list(
    Animal_ID = "^A[0-9]+$"
  )
)

apply_audit_rules(dat, rules)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/id_check.R
\name{id_check}
\alias{id_check}
\title{Check Identifier Integrity}
\usage{
id_check(data, variable, pattern = NULL)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
identifier variable to check.}

\item{pattern}{Optional regular expression that valid
identifiers must match.}
}
\value{
A data.frame of class \code{IDCheck} containing
row-level identifier diagnostics.
}
\description{
Checks an identifier variable for missing values, blank
values, duplicates, leading or trailing whitespace, and
optionally values that do not match a regular expression.
}
\details{
Character, factor, numeric, and integer identifier
variables are supported.

Duplicate detection excludes missing and blank values.

Whitespace checks are performed on the character
representation of the identifier.

If \code{pattern} is supplied, non-missing and non-blank
identifiers that do not match the pattern are flagged.

The function does not modify the original data.
}
\examples{
dat <- data.frame(
  Animal_ID = c(
    "A001",
    "A002",
    "A002",
    " A003",
    "",
    NA
  )
)

id_check(
  dat,
  variable = "Animal_ID",
  pattern = "^A[0-9]{3}$"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dependency_check.R
\name{dependency_check}
\alias{dependency_check}
\title{Check Conditional Variable Dependency}
\usage{
dependency_check(
  data,
  condition_variable,
  condition_value,
  target_variable,
  allow_blank = FALSE
)
}
\arguments{
\item{data}{A data.frame.}

\item{condition_variable}{A single character string
specifying the variable defining the condition.}

\item{condition_value}{The value of \code{condition_variable}
that activates the dependency rule.}

\item{target_variable}{A single character string specifying
the variable that is required when the condition is met.}

\item{allow_blank}{Logical. If FALSE, blank character values
in the target variable are treated as missing. Default is
FALSE.}
}
\value{
A data.frame of class \code{DependencyCheck}
containing row-level dependency diagnostics.
}
\description{
Checks whether a target variable satisfies a requirement
when a specified condition is met in another variable.
}
\details{
The dependency rule is interpreted as:

If \code{condition_variable == condition_value}, then
\code{target_variable} must contain an acceptable value.

Missing values in the condition variable do not activate
the dependency rule.

When \code{allow_blank = FALSE}, empty strings and strings
containing only whitespace are treated as unavailable target
values.

The function does not modify the original data.
}
\examples{
dat <- data.frame(
  Disease_Status = c(
    "Positive",
    "Negative",
    "Positive",
    "Positive"
  ),
  Test_Result = c(
    "PCR positive",
    NA,
    NA,
    "Detected"
  )
)

dependency_check(
  dat,
  condition_variable = "Disease_Status",
  condition_value = "Positive",
  target_variable = "Test_Result"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/audit_score.R
\name{audit_score}
\alias{audit_score}
\title{Calculate a Data Quality Score}
\usage{
audit_score(x)
}
\arguments{
\item{x}{An object of class \code{DataAuditReport}.}
}
\value{
An object of class \code{DataAuditScore}.
}
\description{
Calculates a standardized data-quality score from a
\code{DataAuditReport} produced by \code{\link{audit_data}}.
}
\details{
The score ranges from 0 to 100, where higher values
indicate better data quality.

The score is based on automatic audit issues and custom
rule violations relative to the number of observations
and variables in the audited dataset.
}
\examples{
dat <- data.frame(
  ID = c(1, 2, 2, 4),
  Value = c(10, NA, Inf, 20)
)

report <- audit_data(dat)
audit_score(report)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/case_check.R
\name{case_check}
\alias{case_check}
\title{Check Case Consistency}
\usage{
case_check(data, variable)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
variable to check.}
}
\value{
A data.frame of class \code{CaseCheck} containing
row-level case-consistency diagnostics.
}
\description{
Detects values in a character or factor variable that
differ only in letter case.
}
\details{
Values are compared after conversion to lower case.
Therefore, values such as \code{"Dog"}, \code{"dog"},
and \code{"DOG"} are treated as belonging to the same
normalized category.

A value is flagged when more than one distinct original
representation exists for the same normalized value.

Missing values are reported separately and are not
classified as case inconsistencies.

The function does not modify the original data.
}
\examples{
dat <- data.frame(
  Species = c("Dog", "dog", "DOG", "Cat", "Cat", NA)
)

case_check(
  dat,
  variable = "Species"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_DataAuditReport.R
\name{summary.DataAuditReport}
\alias{summary.DataAuditReport}
\title{Summarize a DataAudit Report}
\usage{
\method{summary}{DataAuditReport}(object, ...)
}
\arguments{
\item{object}{An object of class \code{DataAuditReport}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
An object of class \code{summary.DataAuditReport}
containing dataset-level, variable-level, and optional
custom rule summaries.
}
\description{
Produces a compact summary of a data audit generated by
\code{\link{audit_data}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/length_check.R
\name{length_check}
\alias{length_check}
\title{Check Character Length}
\usage{
length_check(
  data,
  variable,
  min_length = NULL,
  max_length = NULL,
  exact_length = NULL
)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
variable to check.}

\item{min_length}{Optional minimum acceptable number of
characters.}

\item{max_length}{Optional maximum acceptable number of
characters.}

\item{exact_length}{Optional exact required number of
characters.}
}
\value{
A data.frame of class \code{LengthCheck}
containing row-level length diagnostics.
}
\description{
Checks the number of characters in a character or factor
variable and identifies values outside user-defined
length limits.
}
\details{
The selected variable must be character or factor.

Missing values are reported separately and are not
classified as length violations.

If \code{exact_length} is supplied, \code{min_length}
and \code{max_length} must both be NULL.
}
\examples{
dat <- data.frame(
  Sample_ID = c("A01", "A002", "B12", NA)
)

length_check(
  dat,
  variable = "Sample_ID",
  exact_length = 3
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missing_check.R
\name{missing_check}
\alias{missing_check}
\title{Check Missing Values}
\usage{
missing_check(data)
}
\arguments{
\item{data}{A data.frame.}
}
\value{
A data.frame of class \code{MissingCheck}
containing variable-level missing-value information.
}
\description{
Examines each variable in a dataset and reports the
number and percentage of missing and complete values.
}
\details{
For each variable, the function reports the total number
of observations, number of missing values, number of
complete values, and their corresponding percentages.

Missing values are identified using \code{is.na()}.
}
\examples{
missing_check(iris)

dat <- iris
dat$Sepal.Length[1:10] <- NA
dat$Species[1:5] <- NA
missing_check(dat)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_DataAuditScore.R
\name{print.DataAuditScore}
\alias{print.DataAuditScore}
\title{Print a DataAudit Score}
\usage{
\method{print}{DataAuditScore}(x, ...)
}
\arguments{
\item{x}{An object of class \code{DataAuditScore}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
The input object \code{x}, invisibly.
}
\description{
Prints a concise summary of a data-quality score
generated by \code{\link{audit_score}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blank_check.R
\name{blank_check}
\alias{blank_check}
\title{Check Blank Values}
\usage{
blank_check(data, trim = TRUE)
}
\arguments{
\item{data}{A data.frame.}

\item{trim}{Logical. If TRUE, strings containing only
whitespace are treated as blank. Default is TRUE.}
}
\value{
A data.frame of class \code{BlankCheck}
containing variable-level blank-value diagnostics.
}
\description{
Identifies blank values in character and factor variables.
}
\details{
Blank values are evaluated only for character and factor
variables. Missing values (\code{NA}) are reported
separately and are not counted as blank values.

When \code{trim = TRUE}, leading and trailing whitespace
is removed before checking whether a value is blank.
Therefore values such as \code{" "}, \code{"   "}, and
\code{"\\t"} are classified as blank.
}
\examples{
dat <- data.frame(
  name = c("Dog", "", "Cat", "   ", NA),
  value = 1:5,
  stringsAsFactors = FALSE
)

blank_check(dat)

blank_check(dat, trim = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/consistency_check.R
\name{consistency_check}
\alias{consistency_check}
\title{Check Consistency Between Two Variables}
\usage{
consistency_check(data, variable1, variable2, relationship)
}
\arguments{
\item{data}{A data.frame.}

\item{variable1}{A single character string specifying the
first variable.}

\item{variable2}{A single character string specifying the
second variable.}

\item{relationship}{A character string specifying the
expected relationship between the variables. Supported
relationships are \code{"<"}, \code{"<="}, \code{">"},
\code{">="}, \code{"=="}, and \code{"!="}.}
}
\value{
A data.frame of class \code{ConsistencyCheck}
containing row-level consistency diagnostics.
}
\description{
Checks whether values in two variables satisfy a specified
logical relationship.
}
\details{
The function evaluates the relationship as
\code{variable1 relationship variable2}.

For example, when checking whether an end date occurs on
or after a start date, use:

\code{variable1 = "EndDate"},
\code{variable2 = "StartDate"}, and
\code{relationship = ">="}.

Rows containing a missing value in either variable are
reported separately and are not classified as consistent.

The two variables must have compatible types for the
requested comparison.
}
\examples{
dat <- data.frame(
  StartDate = as.Date(c(
    "2026-01-01",
    "2026-02-01",
    "2026-03-01"
  )),
  EndDate = as.Date(c(
    "2026-01-10",
    "2026-01-20",
    "2026-03-15"
  ))
)

consistency_check(
  dat,
  variable1 = "EndDate",
  variable2 = "StartDate",
  relationship = ">="
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_sequence_check.R
\name{group_sequence_check}
\alias{group_sequence_check}
\title{Check Sequence Integrity Within Groups}
\usage{
group_sequence_check(data, group, variable, step = 1)
}
\arguments{
\item{data}{A data.frame.}

\item{group}{A single character string specifying the
grouping variable.}

\item{variable}{A single character string specifying the
numeric or integer sequence variable.}

\item{step}{A positive finite numeric value specifying the
expected increment. Default is 1.}
}
\value{
A data.frame of class \code{GroupSequenceCheck}
containing row-level sequence diagnostics.
}
\description{
Checks a numeric or integer sequence independently within
groups for missing values, duplicates, ordering problems,
gaps, and step mismatches.
}
\details{
Sequence integrity is evaluated independently within each
group and in the original row order.

Missing values in the grouping variable are retained and
treated as a separate missing group for diagnostic
purposes.

Missing sequence values are reported separately and are
excluded when calculating previous values and differences.

All occurrences of duplicated non-missing sequence values
within a group are flagged.

A gap occurs when the difference from the previous
non-missing value within the same group is greater than
\code{step}.
}
\examples{
dat <- data.frame(
  Animal_ID = c(
    "A001", "A001", "A001",
    "A002", "A002", "A002"
  ),
  Visit = c(
    1, 2, 4,
    1, 2, 2
  )
)

group_sequence_check(
  dat,
  group = "Animal_ID",
  variable = "Visit"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_DataAuditRuleResult.R
\name{summary.DataAuditRuleResult}
\alias{summary.DataAuditRuleResult}
\title{Summarize a DataAudit Rule Result}
\usage{
\method{summary}{DataAuditRuleResult}(object, ...)
}
\arguments{
\item{object}{An object of class
\code{DataAuditRuleResult}.}

\item{...}{Additional arguments, currently unused.}
}
\value{
An object of class
\code{summary.DataAuditRuleResult}.
}
\description{
Produces a compact summary of custom validation-rule
results generated by \code{\link{apply_audit_rules}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unique_check.R
\name{unique_check}
\alias{unique_check}
\title{Check Unique Identifier}
\usage{
unique_check(data, variable)
}
\arguments{
\item{data}{A data.frame.}

\item{variable}{A single character string specifying the
variable to check.}
}
\value{
A data.frame of class \code{UniqueCheck}
containing row-level uniqueness diagnostics.
}
\description{
Evaluates whether a variable can serve as a unique
identifier by checking for missing and duplicated values.
}
\details{
Each observation is evaluated for missing and duplicated
identifier values.

A value is considered duplicated when it occurs more than
once among non-missing observations. All occurrences of a
repeated value are flagged.

Missing values are flagged separately and are not treated
as duplicated values.

The output contains:

\itemize{
\item \code{Row}: row number.
\item \code{Value}: identifier value.
\item \code{Missing}: whether the value is missing.
\item \code{Duplicate}: whether the non-missing value
occurs more than once.
\item \code{ValidUnique}: whether the value is both
non-missing and unique.
}
}
\examples{
dat <- data.frame(
  Animal_ID = c(101, 102, 102, 103, NA)
)

unique_check(
  dat,
  "Animal_ID"
)

}
